# LSB Test Execution Framework
# Journal & report module
#
# Copyright (C) 2006 Free Standards Group. All rights reserved.
#
# This program has been developed by ISP RAS for FSG with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2006 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Journal;

use strict;

use Common;
use Packages;
use IO::File;

# Export symbols
require Exporter;
@Journal::ISA = qw(Exporter);
push @Journal::EXPORT, qw(
    &parse_tet_journal &parse_tcl_journal &parse_python_journal 
    &parse_simple_journal &write_test_info &generate_report
    &pack_and_send_output
);


# HTML colors used in the report.
my $dark_gray = "rgb(64, 64, 64)";
my $gray = "rgb(200, 200, 200)";
my $neutral_gray = "rgb(230, 230, 230)";
my $bright_gray = "rgb(240, 240, 240)";
my $fluorescent_green = "rgb(64, 255, 64)";
my $fluorescent_yellow = "rgb(255, 255, 64)";
my $fluorescent_red = "rgb(255, 96, 96)";
my $pale_blue = "rgb(220, 220, 240)";
my $dark_blue = "rgb(0, 0, 128)";
my $dark_red = "rgb(128, 0, 0)";
my $red = "rgb(255, 0, 0)";
my $green = "rgb(0, 160, 0)";
my $blue = "rgb(0, 0, 255)";

# Summary statistics
my $stats_failures_reported = 0;
my $stats_warnings_reported = 0;
my $stats_fips_reported = 0;
my $stats_notices_reported = 0;
my $stats_tests_total = 0;
my $stats_tests_passed = 0;
my $stats_tests_skipped = 0;
my $stats_tests_quelled = 0;
my $tests_cancelled = 0;
my $overall_verdict = "";

# The problem database, hashed by the "test_name test_point" strings.
my %problem_database;

# Object representing a test being processed.
my $test_object;

# The final exit code, which is 0 if all tests have passed, or 1 if some
# failures were detected.
my $exit_code = 0;

# The report file.
my $report;

# The single-line summary file.
my $summary_line;

# This variable is set to 1 if part of the problem table was already printed
# (at least the header), false if not (this implies that no problems were
# found so far).
my $problem_table_printed = 0;

# A monotonously increasing, unique identifier of a message box.
my $msgbox_id = 1;

# Copy of the input line that was processed last.
my $line = "";

#----------------------------------------------------------------------------

# The journal file being processed.
my $journal_file;

# The title of the journal being processed.
my $journal_title = "";

# The HTML anchor used for marking this journal in the resulting report.
my $journal_anchor = "";

#----------------------------------------------------------------------------

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
    0, "Passed",
    1, "Failed",
    2, "Unresolved",
    3, "Unused",
    4, "Unsupported",
    5, "Untested",
    6, "Uninitiated",
    7, "Unreported",
    101, "Warning",
    102, "FIP",
    103, "Unimplemented",
    104, "Unapproved",
);

#----------------------------------------------------------------------------
# Prints time and date in a human-readable format
sub format_time {
    my $timestamp = shift;
    
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = 
        localtime ($timestamp);

    return sprintf ("%02d:%02d:%02d %02d-%02d-%04d\n", $hour, $min, $sec,
        $mday, ($mon + 1), ($year + 1900) );
}

#----------------------------------------------------------------------------

# Loads the problem database which is used for automatic identification
# of known test suite failures, warnings and false positives.
sub load_problem_database {
    my $file_name = shift;

    # Open the database file.
    open(FILE, $file_name) or fail("Could not open problem database $file_name");

    # Reset the line number counter.
    my $line_number = 0;

    # The problem record being read.
    my $problem;

    # Read the database line by line.
    my ($test_name, $test_point);
    while ($line = <FILE>) {

        # Increment the line number.
        $line_number++;

        # Skip empty lines and comment lines.
        next if ($line =~ /^\s*(\#.*)?$/);

        # The 'test' line starts a new test record.
        # It contains the name of the test and the test point number.
        if ($line =~ /^test\s*(\S*)\s*(\d*)/) {

            # Store the test name and point.
            $test_name = $1;
            $test_point = $2;

            # Clear the old record and prepare a clean one.
            $problem = {
                test_name => $test_name,
                test_point => $test_point,
                markers => [],
                resolution => "",
                comment => "",
                quell => 0,
                lsb_report_id => "",
                lsb_resolution_id => "",
                expected_severity => ""
            };
        }

        # The 'expected_severity' line indicates the severity
        # expected to be reported by the test suite.
        elsif ($line =~ /^expected_severity\s*(.*)/) {
            my $expected_severity = $1;

            # Check if the expected severity is one of the predefined values.
            unless (($expected_severity eq "passed")
                || ($expected_severity eq "failed")
                || ($expected_severity eq "unresolved")
                || ($expected_severity eq "unused")
                || ($expected_severity eq "unsupported")
                || ($expected_severity eq "untested")
                || ($expected_severity eq "uninitiated")
                || ($expected_severity eq "unreported")
                || ($expected_severity eq "warning")
                || ($expected_severity eq "fip")
                || ($expected_severity eq "unimplemented")
                || ($expected_severity eq "unapproved")
            ) {
                fail("problem database: $test_name $test_point: " .
                    "unrecognized expected_severity '$1'");
            }

            $problem->{'expected_severity'} = $expected_severity;

        }

        # The 'marker' line contains a marker which can be found
        # to identify the particular problem from the messages.
        # There may be (and typically are) more markers, each on its own
        # 'marker' line.
        elsif ($line =~ /^marker\s*(.*)/) {
            push @{$problem->{'markers'}}, $1;
        }

        # The 'resolution' line contains the resolution of the problem.
        elsif ($line =~ /^resolution\s*(.*)/) {
            $problem->{'resolution'} = $1;
        }

        # If the 'quell' line is present, the problem is considered harmless
        # regardless of severity. This is used for already solved bugs
        # and false positives.
        elsif ($line =~ /^quell\s*(.*)/) {
            $problem->{'quell'} = 1;
        }

        # The 'comment' line holds a comment for human reader
        # which is copied to the problem report. There can be more
        # comment lines.
        elsif ($line =~ /^comment\s(.*)/) {
            $problem->{'comment'} .= "$1\n";
        }

        # The 'lsb_report_id' contains the LSB problem report ID.
        elsif ($line =~ /^lsb_report_id\s(\S*)/) {
            $problem->{'lsb_report_id'} = $1;
        }

        # The 'lsb_resolution_id' contains the LSB resolution ID.
        elsif ($line =~ /^lsb_resolution_id\s(\S*)/) {
            $problem->{'lsb_resolution_id'} = $1;
        }

        # The 'end' line finishes one problem entry.
        elsif ($line =~ /^end/) {

            # Check if this problem report has all required fields.
            if ($problem->{'expected_severity'} eq "") {

                # Expected severity must be provided.
                fail("problem database: $test_name $test_point: " .
                    "no expected_severity field");
            }

            # The problem record database is indexed by the test name
            # and point number, joined together to a key.
            # If there are multiple records for the same test point, they are
            # distinguished by an additional sequential number.
            my $record_key = $test_name . " " . $test_point;

            # If there are already some existing records for this test point,
            # find the next unused sequential number.
            my $sequential_number = 0;
            while (my $existing_record = $problem_database{$record_key}) {
                $sequential_number++;
                $record_key = $test_name . " " . $test_point . "($sequential_number)";
            }

            # Add the problem record to the database.
            $problem_database{$record_key} = $problem;
        }

        # Anything else is a syntax error.
        else {
            fail("problem database: syntax error at line $line_number");
        }
    }
    close FILE;
}

#----------------------------------------------------------------------------

# Finds a problem record in the database.
#
# Arguments:
# $1 - test name
# $2 - test point
# $3 - problem severity
# $4 - test messages
sub find_problem_record
{
    my ($test_name, $test_point, $severity, $test_messages) = @_;

    # Some automatically generated problem reports have no test name or point;
    # these are not meant to be searched in the database but directly reported
    # to the user.
    unless (($test_name) && ($test_point > 0)) {
        # Do not bother searching, return nothing immediately.
        return undef;
    }

    # The basic record key is the name of the test and the test point number.
    my $record_key = $test_name . " " . $test_point;

    # Try finding the problem record.
    my $sequential_number = 0;
    while (my $problem_record = $problem_database{$record_key}) {

        # Check if the severity matches.
        my $expected_severity = $problem_record->{'expected_severity'};
        if (lc($severity) eq lc($expected_severity)) {

            # Look which distinguishing markers the problem should have.
            my $markers_ref = $problem_record->{'markers'};

            # If the problem record contains no markers, it matches only
            # if there were no messages.
            unless (@$markers_ref) {
                unless ($test_messages) {
                    return $problem_record;
                }
            }

            # Check if the markers are found in the test messages.
            my $markers_matched = 0;
            foreach (@$markers_ref) {
                if (index($test_messages, $_) >= 0) {
                    $markers_matched++;
                }
            }

            # All the markes must match.
            if ($markers_matched == @$markers_ref) {
                # Yup, this is the problem report we looked for.
                return $problem_record;
            }
        }

        # This report does not match, but there may be another one
        # for the same test point - increment the sequential number
        # and retry.
        $sequential_number++;

        # Add the sequential number to the record key.
        $record_key = $test_name . " " . $test_point . "($sequential_number)";
    }

    # Nothing was found.
    return undef;
}

#----------------------------------------------------------------------------

# Processes the result of a test point.
#
# Arguments:
# $1 - test number (0 if the tests are not numbered)
# $2 - test name
# $3 - test point (0 if the tests are not divided into test points)
# $4 - severity
# $5 - test messages
sub process_test_point_result {
    my ($test_number, $test_point, $test_name, $severity, $test_messages) = @_;

    # Sanity check: Test name must be valid.
    unless ($test_name) {
        debug_inform ("Empty test name");
        return;
    }

    # Compare severity strings in lowercase.
    my $lowercase_severity = lc($severity);

    # Increment the total test count.
    $test_object->{TESTS_TOTAL}->{$journal_title}++;
    $stats_tests_total++;

    # Passed and unused tests are not printed, only counted.
    if ($lowercase_severity eq "passed") {
        $test_object->{TESTS_PASSED}->{$journal_title}++;
        $stats_tests_passed++;
        return;
    }
    elsif (($lowercase_severity eq "unused") || ($lowercase_severity eq "untested") ||
        ($lowercase_severity eq "unsupported") || ($lowercase_severity eq "unimplemented")) {
        $test_object->{TESTS_SKIPPED}->{$journal_title}++;
        $stats_tests_skipped++;
        return;
    }

    my $quelled = 0;
    my $resolution = "";
    my $comment = "";
    my $lsb_report_id = "";
    my $lsb_resolution_id = "";

    # Compare the problem description with the database.
    my $problem_record = find_problem_record($test_name, $test_point, $severity, $test_messages);
    if ($problem_record) {

        # Load the data from the problem record.
        $quelled = $problem_record->{'quell'};
        $resolution = $problem_record->{'resolution'};
        $comment = $problem_record->{'comment'};
        $lsb_report_id = $problem_record->{'lsb_report_id'};
        $lsb_resolution_id = $problem_record->{'lsb_resolution_id'};

        # Add the "(quelled)" keyword to the severity if the problem is quelled.
        # This means it can be safely ignored.
        if ($quelled) {

            # Increment the quelled tests count statistic.
            $test_object->{TESTS_QUELLED}->{$journal_title}++;
            $stats_tests_quelled++;

            # Add the "(quelled)" keyword to the severity. This means
            # the report can be safely ignored.
            $severity .= " (quelled)";
        }
    }
    
    # Print the problem to the table.
    print_problem_table_line(
        $test_number,
        $test_point,
        $test_name,
        $severity,
        $resolution,
        $comment,
        $lsb_report_id,
        $lsb_resolution_id,
        $test_messages
    );
}
#----------------------------------------------------------------------------

# Prints a header of the problem table. It is called automatically
# from report_problem() when the first problem is reported.
# (If there are no problems during the whole test run, the table
# is not printed at all.)
sub print_problem_table_header() {

    my $help_id = "problem_summary_help" . int(10000 * rand());

    $test_object->report("<h4>Problem Summary</h4>\n\n");
    $test_object->report("<p>Click on lines in the table to see the details " .
        "about each problem.</p>\n\n");
    $test_object->report("<table width=\"100%\">\n");
    $test_object->report("<thead>\n");
    $test_object->report("<tr style=\"font-weight:bold; background-color:$gray;\">\n");
    $test_object->report("<td>Test Name</td><td>Severity</td><td>Resolution</td>\n");
    $test_object->report("</tr>\n");
    $test_object->report("</thead>\n");
    $test_object->report("<tbody>\n");
}

#----------------------------------------------------------------------------

# Prints a line of the problem table.
#
# Arguments:
#
# $1 - test number (0 if the tests are not numbered, or if this problem line
#      is not connected to any particular test).
# $2 - test point number (0 if there is no test point).
# $3 - test name (empty if this problem line is not connected to any
#      particular test).
# $4 - severity.
# $5 - resolution (empty if unknown).
# $6 - comment
# $7 - report ID in the LSB database (empty if not yet reported).
# $8 - resolution ID in the LSB database (empty if not yet resolved).
# $9 - messages printed by the testing program.
sub print_problem_table_line
{
    my $test_number = shift;
    my $test_point = shift;
    my $test_name = shift;
    my $severity = shift;
    my $resolution = shift;
    my $comment = shift;
    my $lsb_report_id = shift;
    my $lsb_resolution_id = shift;
    my $test_messages = shift;

    # Compare the severity string case-insensitively (we are only humans).
    my $lowercase_severity = lc($severity);

    # Choose the color for the specified message severity.
    my $color;
    if ($lowercase_severity =~ /\(quelled\)/) {
        # Anything what is quelled is harmless, so use neutral gray
        # for all quelled problems.
        $color = "gray";
    }
    elsif (($lowercase_severity eq "failed") || ($lowercase_severity eq "failure") ||
        ($lowercase_severity eq "unresolved") || ($lowercase_severity eq "unreported") ||
        ($lowercase_severity eq "uninitiated") || ($lowercase_severity eq "test suite error")
    ) {
        # Use fluorescent red for various failures.
        $color = "red";

        # Increment the number of failures reported.
        $test_object->{FAILURES_REPORTED}->{$journal_title}++;
        $stats_failures_reported++;

        # Even a single failure in a single test is enough for flipping
        # the exit code from 0 to 1.
        $exit_code = 1 << 8;
    }
    elsif ($lowercase_severity eq "warning") {
        # Use fluorescent yellow for warnings.
        $color = "yellow";

        # Increment the number of warnings reported.
        $test_object->{WARNINGS_REPORTED}->{$journal_title}++;
        $stats_warnings_reported++;
    }
    elsif ($lowercase_severity eq "fip") {
        # Use fluorescent green for FIPs.
        $color = "green";

        # Increment the number of FIPs reported.
        $test_object->{FIPS_REPORTED}->{$journal_title}++;
        $stats_fips_reported++;
    }
    elsif ($lowercase_severity eq "notice") {
        # Use neutral gray for harmless messages.
        $color = "gray";

        # Increment the number of notices reported.
        $test_object->{NOTICES_REPORTED}->{$journal_title}++;
        $stats_notices_reported++;
    }
    else {
        debug_inform ("Unrecognized severity code '$severity'");
    }

    # If the problem table header was not yet printed, print it now.
    unless ($problem_table_printed) {
        print_problem_table_header();
        $problem_table_printed = 1;
    }

    # Give the message box a unique ID.
    my $message_box_id = "msgbox" . $msgbox_id; $msgbox_id++;

    # Start the line (with attached script that opens the detail block
    # on clicking).
    $test_object->report("<tr class=\"ptl_$color\" title='Click here!' onclick=\"flip('$message_box_id');\">");

    # Print the test name and test point (if available).
    $test_object->report("<td>");
    if ($test_name) {
        $test_object->report("$test_name ");
    }
    else {
        $test_object->report("(none) ");
    }
    if ($test_point > 0) {
        $test_object->report("<span style=\"color:$dark_blue;\">$test_point</span>\n");
    }
    $test_object->report("</td>\n");

    # Print severity and resolution.
    $test_object->report("<td>$severity</td>\n");
    $test_object->report("<td>$resolution</td>\n");

    # Finish the line.
    $test_object->report("</tr>\n");

    # Print the detail block, which is initially hidden and unrolls when
    # the table row is clicked (via the javascript above).
    $test_object->report("<tr id=\"$message_box_id\" style=\"display:none;\">\n");
    $test_object->report("<td colspan=\"3\">");
    $test_object->report("<span>\n");
    $test_object->report("<span style=\"padding-left:32px;\">Messages from the test:</span><br />\n");
    $test_object->report("<span style=\"font-family:monospace; padding-left:64px; color:$dark_gray\">\n");
    if ($test_messages) {
        # Escape HTML entities
        $test_messages =~ s/&/&amp;/g;
        $test_messages =~ s/</&lt;/g;
        $test_messages =~ s/>/&gt;/g;
        $test_messages =~ s/[\n\r]+/<br \/>\n/g;
        $test_object->report($test_messages."\n</span><br />\n");
    }
    else {
        $test_object->report("(none)</span><br />\n");
    }
    if ($comment) {
        $test_object->report("<span style=\"padding-left:32px;\">Comment:</span><br />\n");
        $test_object->report("<span style=\"padding-left:64px; color:$dark_gray;\">\n");
        $test_object->report("$comment\n</span><br />\n");
    }
    if ($lsb_report_id or $lsb_resolution_id) {
        $test_object->report("<span style=\"padding-left:32px;\">LSB Report Status:</span><br />\n");
        $test_object->report("<span style=\"padding-left:64px; color:$dark_gray;\">");
        if ($lsb_report_id) {
            $test_object->report("Reported to the LSB Committee as ");
            $test_object->report("<a href=\"http://www.freestandards.org/cert/pr.php?view=$lsb_report_id\">");
            $test_object->report("PR $lsb_report_id</a>.<br />\n");
        }
        if ($lsb_resolution_id) {
            $test_object->report("Resolved (resolution ID is ");
            $test_object->report("<span style=\"color:$dark_blue;\">$lsb_resolution_id</span>).<br />\n");
        }
        $test_object->report("</span>");
    }
    $test_object->report("</span>");
    if ( $color eq "red" ) {
#        $test_object->report("<br />");

         $test_object->report("<div align=\"right\">"
            ."<input type=\"button\" value=\"Report bug in test\""
            ." onclick=\"javascript:alert('"
            ."The function will be implemented in the future.".
            "')\" title=\"Press this button if you beleive the problem is in the test, not in your system.\" /></div>"
            );
#        $test_object->report("<br />");
    }
    $test_object->report("</td>");
    $test_object->report("</tr>\n");
}

#----------------------------------------------------------------------------

# Prints the footer of the problem table. Does nothing if the problem table
# was not printed at all.
sub finish_problem_table() {

    # Complain if we have not seen any tests at all; this may mean the test suite
    # terminated somehow at the very start.
    if ($test_object->{TESTS_TOTAL}->{$journal_title} == 0) {
        print_problem_table_line(
            0,
            0,
            "(none)",
            "Failure",
            "No tests found in the journal",
            "",
            "",
            $line
        );
    }

    # Close the table or report that no problems were detected
    if ($problem_table_printed) {
        $test_object->report("</tbody>\n");
        $test_object->report("</table>\n");
    }
    else {
        $test_object->report("<h4>Problem Summary</h4>\n");
        $test_object->report("<p>No problems were detected.</p>\n");
    }
}

#----------------------------------------------------------------------------

# Starts a two-column table.
sub start_two_column_table {
    $test_object->report("<table>\n");
    $test_object->report("<tbody>\n");
}

#----------------------------------------------------------------------------

# Prints a single line of a two-column table.
#
# Arguments:
# $1 - The first column.
# $2 - The second column.
# $3 (optional) - The foreground color of the second column.
# $4 (optional) - The background color of the second column.
sub two_column_table_line {
    my $first_column = shift;
    my $second_column = shift;
    my $foreground_color = shift;
    my $background_color = shift;
    
    # Build the style specifier for the second column.
    my $second_column_style = "padding: 2px; margin: 2px;";
    
    # If the second column contains a number, print it right-aligned.
#    if ($second_column =~ /^[0-9\.]*$/) {
        $second_column_style .= "text-align: right; ";
#    }
    
    # Apply the foreground and background colors if specified.
    if ($foreground_color) {
        $second_column_style .= "color: $foreground_color; ";
    }
    if ($background_color) {
        $second_column_style .= "background-color: $background_color; ";
    }
    else {
        $second_column_style .= "background-color: $neutral_gray; ";
    }
    
    # Print the line with appropriate decorations.
    return "<tr style=\"padding: 4px; margin: 4px;\">".
    "<td style=\"background-color:$gray;\">$first_column</td>\n".
    "<td style=\"$second_column_style\">$second_column</td>\n".
    "</tr>\n";
}
#----------------------------------------------------------------------------

# Prints a two-column table line to the test report
sub print_two_column_table_line {
    $test_object->report(two_column_table_line(@_));
}
#----------------------------------------------------------------------------

# Finishes a two-column table.
sub finish_two_column_table {
    $test_object->report("</tbody>\n");
    $test_object->report("</table>\n");
}

#----------------------------------------------------------------------------

sub print_statistics_line {
    my ($title, $value, $warn_if_nonzero, $warning_color) = @_;

    # Use neutral gray background by default.
    my $background_color = $neutral_gray;

    # If a warning is requested, change the color
    # when the value is nonzero.
    if ($warn_if_nonzero && ($value > 0) ) {
        print_two_column_table_line($title, $value, $dark_blue,
            $warning_color);
    }
    else {
        print_two_column_table_line($title, $value, $dark_blue);
    }
}


#----------------------------------------------------------------------------

# Prints the statistics.
sub print_statistics() {

    $test_object->report("<h4>Statistics</h4>\n");

    # Calculate the time interval between the tests start and stop.
    my $test_duration = "unknown";
    if ( $test_object->{FINISH_TIME} && $test_object->{START_TIME} )
    {
        $test_duration = $test_object->{FINISH_TIME} - $test_object->{START_TIME};
        my $test_duration_sec = $test_duration % 60;
        $test_duration = int($test_duration / 60);
        my $test_duration_min = $test_duration % 60;
        my $test_duration_hour = int($test_duration / 60);
        
        if($test_duration_hour > 0) {
            $test_duration = "${test_duration_hour}h ${test_duration_min}m ${test_duration_sec}s";
        }
        else {
            $test_duration = "${test_duration_min}m:${test_duration_sec}s";
        }
    }
    
    # Print the statistics table.
    my $background_color;
    start_two_column_table();
    print_statistics_line("Test time", $test_duration, 0);
    print_statistics_line("Total tests", $test_object->{TESTS_TOTAL}->{$journal_title}, 0);
    print_statistics_line("Passed tests", $test_object->{TESTS_PASSED}->{$journal_title}, 0);
    if ($test_object->{TESTS_SKIPPED}->{$journal_title} > 0) {
        print_statistics_line("Tests not relevant/not applicable", $test_object->{TESTS_SKIPPED}->{$journal_title}, 0);
    }
    if ($test_object->{TESTS_QUELLED}->{$journal_title} > 0) {
        print_statistics_line("Quelled problems", $test_object->{TESTS_QUELLED}->{$journal_title}, 0);
    }
    if ($test_object->{FIPS_REPORTED}->{$journal_title} > 0) {
        print_statistics_line("FIPs", $test_object->{FIPS_REPORTED}->{$journal_title}, 1, $fluorescent_green);
    }
    if ($test_object->{WARNINGS_REPORTED}->{$journal_title} > 0) {
        print_statistics_line("Warnings", $test_object->{WARNINGS_REPORTED}->{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($test_object->{FAILURES_REPORTED}->{$journal_title} > 0) {
        print_statistics_line("Failures", $test_object->{FAILURES_REPORTED}->{$journal_title}, 1, $fluorescent_red);
    }

    # Give the final verdict.
    if ($test_object->{FAILURES_REPORTED}->{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Failure");
    }
    elsif ($test_object->{WARNINGS_REPORTED}->{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Warnings");
    }
    else {
        print_two_column_table_line("Verdict", "Success");
    }

    # Finish the table.
    finish_two_column_table();

    # Sanity check: If everything looks fine, check if the number of tests
    # matches the sum of skipped, passed, quelled and FIPed tests. This helps
    # preventing false negatives (which are most dangerous).
    if ($test_object->{FAILURES_REPORTED}->{$journal_title} == 0) {
        if ($test_object->{TESTS_SKIPPED}->{$journal_title} + $test_object->{TESTS_PASSED}->{$journal_title} + $test_object->{FIPS_REPORTED}->{$journal_title}
            + $test_object->{TESTS_QUELLED}->{$journal_title} != $test_object->{TESTS_TOTAL}->{$journal_title}) {

            # Oops.
            debug_inform ("No failures reported but some tests are missing");
        }
    }
}

#----------------------------------------------------------------------------

sub print_test_summary_line
{
    my ($title, $anchor) = @_;

    my $title_comment = sprintf ("%-25s", "<!-- $title -->");

    # Add the result into the single-line summary.
    $summary_line->print ("    $title_comment <td>");
    if ($test_object->{FAILURES_REPORTED}->{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"failure_icon.png\" alt=\"F\">");
        $summary_line->print ("</a>");
    }
    elsif ($test_object->{WARNINGS_REPORTED}->{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"warning_icon.png\" alt=\"W\">");
        $summary_line->print ("</a>");
    }
    elsif ($test_object->{TESTS_TOTAL}->{$journal_title} > 0) {
        $summary_line->print ("<a href=\"test_log.html#$anchor\">");
        $summary_line->print ("<img src=\"ok_icon.png\" alt=\"OK\">");
        $summary_line->print ("</a>");
    }
    else {
        $summary_line->print ("<!-- result unavailable -->");
    }
    $summary_line->print ("</td>\n");
}

#----------------------------------------------------------------------------

sub open_journal
{
    my $title = shift;
    my $anchor = shift;
    my $file_name = shift;
    
    $test_object = shift;
    
    $journal_title = $title;
    
    # If journal isn't closed then close it.
    if ( $journal_file ) { close_journal (); }

    # Store the journal title and anchor.
    $test_object->open_report($title, $anchor);

    $problem_table_printed = 0;
    
    debug_inform("Reading journal: $file_name");
    # Open the journal file.
    $journal_file = new IO::File;
    unless (open ($journal_file, $file_name)) {
        
        # The journal file could not be open; mention this in the log
        # (so that we know that something was missing) and return.
        
        $tests_cancelled++;
        
        $test_object->report("<p>$title not tested, or the test journal ($file_name) was not present.</p>\n");
        print_test_summary_line ($title, $anchor);
        $journal_file = undef;
        debug_inform("Failed to read journal: $file_name\n$title wasn't tested, or the test journal was not present.");
    }
}

#----------------------------------------------------------------------------

sub close_journal
{
    unless ($journal_file) { return; }
    
    # Finish the problem table.
    finish_problem_table();

    # Print the statistics.
    print_statistics();

    # Add the result into the single-line summary.
    print_test_summary_line ($journal_title, $journal_anchor);

    # Close the journal file.
    $journal_file->close ();

    # Reset state variables.
    $journal_file = undef;
    $journal_title = "";
    $journal_anchor = "";
}

#----------------------------------------------------------------------------

# The main TET journal parser function which loads the TET journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tet_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }
    unless ($test_object) { return; }
    
    # This flag is set to nonzero if we are inside a test point.
    my $inside_test = 0;
    my $inside_test_point = 0;

    # Read the journal line by line.
    my ($test_name, $test_number, $test_point, $test_messages);
    my $test_points_in_this_case;
    while ($line = <$journal_file>) {

        # Each journal line consists of three parts, separated by
        # a vertical bar:
        #
        # 1. A numeric line code, determining the meaning of the line.
        # 2. Additional data, format depends on line code.
        # 3. Additional text (message for the user, or some comment).
        unless ($line =~ /^(\d*)\|(.*)\|(.*)$/) {

            # Lines that do not conform to this format are reported
            # to the user and skipped.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Warning",
                "Unrecognized entry in test journal",
                "",
                "",
                "",
                $line
            );
            next;
        }
        my $line_code = $1;
        my $line_data = $2;
        my $line_text = $3;

        # Determine the type of the line and process it appropriately.
        if ($line_code == 0) {
            # Code 0 -- Test Case Controller Start.
            # The line structure is:
            # 0|version time date|who
            unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "",
                    "",
                    "",
                    $line);
                next;
            }

            # This line should be one of the first;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case Start entry appears in the middle of the journal",
                    "",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $test_engine_version = $1;
                my $test_start_time = $2;
                my $test_start_date = $3;
                my $test_command_line = $line_text;

                # Print the test engine version, time and date to the output.
                $test_object->report("<h4>Test suite information</h4>");
                start_two_column_table();
                print_two_column_table_line("Test Engine Version", $test_engine_version);
                print_two_column_table_line("Test Started At", "$test_start_time $test_start_date");
                print_two_column_table_line("Test Command Line", "$test_command_line");
                finish_two_column_table();
            }
        }
        elsif ($line_code == 5) {

            # Code 5 -- Local System Information.
            # The line structure is:
            # 5|sysname nodename release version_and_machine|text
            unless ($line =~ /^5\|(\S*)\s*(\S*)\s*(\S*)\s*(.*)\|/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "",
                    "",
                    "",
                    $line);
                next;
            }

            # The system information should be one of the first items;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "System info entry appears in the middle of the journal",
                    "",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $system_name = $1;
                my $node_name = $2;
                my $release = $3;
                my $version = $4;

                # Print the local system data.
                $test_object->report("<h4>Tested system information</h4>\n");
                start_two_column_table();
                print_two_column_table_line("System Name", $system_name);
                print_two_column_table_line("Node Name", $node_name);
                print_two_column_table_line("Release", $release);
                print_two_column_table_line("Version", $version);
                finish_two_column_table();
            }
        }
        elsif ($line_code == 10) {

            # Code 10 -- Test Case Start.
            # The line structure is:
            # 10|activitynum testname hh:mm:ss|list of invocable components
            unless ($line =~ /^\d*\|(\d*)\s*(\S*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "",
                    "",
                    "",
                    $line);
                next;
            }

            if ($test_name) {
                # The previous test case was not properly finished;
                # issue a warning.
                print_problem_table_line(
                    $1,
                    0,
                    $test_name,
                    "Warning",
                    "Test case starts before the previous one ($test_name) has finished",
                    "",
                    "",
                    "",
                    ""
                );
            }

            # Store the test name and number.
            $test_number = $1;
            $test_name = $2;

            # Reset the test point number and counter.
            $test_point = 0;
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 15) {
            # Code 15 -- Test Case Manager Start.
            # Not interesting to us.
        }
        elsif ($line_code == 20) {
            # Code 20 -- Local/Remote System Configuration Start.
            # Not interesting to us.
        }
        elsif ($line_code == 30) {
            # Code 30 -- Configuration Variable.
            # Not interesting to us.
        }
        elsif ($line_code == 40) {
            # Code 40 -- Configuration End.
            # Not interesting to us.
        }
        elsif ($line_code == 50) {
            # Code 50 -- Test Case Controller Message.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Test case controller reports a problem",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 70) {
            # Code 70 -- Scenario Information.
            # Not interesting to us.
        }
        elsif ($line_code == 80) {
            # Code 80 -- Test Case End.
            unless ($test_name) {
                # Test case end found when no test case was started;
                # issue a warning.
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case End entry found when no test case seems to be running",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Warn if the test case seems to be empty.
            if ($test_points_in_this_case == 0) {
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "Warning",
                    "Test case contains no test points",
                    "",
                    "",
                    "",
                    ""
                );
            }

            # Clear the test name, number, and messages.
            $test_name = "";
            $test_number = 0;
            $test_messages = "";
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 90) {
            # Code 90 -- User Abort.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Test Suite Error",
                "Test suite execution interrupted by the user",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 100) {
            # Code 100 -- Captured Output.
            # Not interesting to us.
        }
        elsif ($line_code == 110) {
            # Code 110 -- Build Start.
            # Not interesting to us.
        }
        elsif ($line_code == 130) {
            # Code 130 -- Build End.
            # Not interesting to us.
        }
        elsif ($line_code == 200) {
            # Code 200 -- Test Purpose Start.
            # The structure of the line is as follows:
            # 200|activitynum pointnum hh:mm:ss|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "",
                    "",
                    "",
                    $line);
                next;
            }

            # Check if the previous test point was properly finished.
            if ($test_point > 0) {
                if ($test_point == $2) {
                    # Some test suites have a strange habit of re-starting
                    # the same test point more than once.
                    print_problem_table_line(
                        $test_number,
                        $test_point,
                        $test_name,
                        "Notice",
                        "Test was restarted in the middle",
                        "",
                        "",
                        "",
                        ""
                    );
                }
                else {
                    # Notify about a disappeared test.
                    print_problem_table_line(
                        $test_number,
                        $2,
                        $test_name,
                        "Warning",
                        "Test point starts before the previous one ($test_point) has finished",
                        "",
                        "",
                        "",
                        ""
                    );
                }
            }

            # Clear the messages.
            $test_messages = "";

            # Store the test point number.
            $test_point = $2;

            # Increment the number of test points.
            $test_points_in_this_case++;
        }
        elsif ($line_code == 220) {
            # Code 220 -- Test Purpose Result.
            # The structure of the line is as follows:
            # 220|activitynum pointnum result time|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)\s*(\d*)/) {
                complain("Syntax error in journal");
                next;
            }

            # Print a warning if the test purpose seems to never been started.
            if ($test_point == 0) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test result entry found but the test point had not started",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Print a warning if the test purpose number is different from
            # what we expected.
            elsif ($test_point != $2) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test purpose number at the end ($2) is different " .
                    "from the starting one ($test_point)",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Get the result code.
            my $result_code = $3;

            # Find the meaning of the result code (severity).
            my $severity = $result_code_map{$result_code};
            unless ($severity) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    "Unrecognized test result code ($result_code)",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }
            else {
                # Process the result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    $severity,
                    $test_messages
                );
            }

            # Clear the test point number.
            $test_point = 0;
        }
        elsif ($line_code == 310) {
            # 310 -- Clean Start.
            # Not interesting to us.
        }
        elsif ($line_code == 320) {
            # 320 -- Clean End.
            # Not interesting to us.
        }
        elsif ($line_code == 400) {
            # 400 -- Invocable Component Start.
            # Not interesting to us.
        }
        elsif ($line_code == 410) {
            # 410 -- Invocable Component End.
            # Not interesting to us.
        }
        elsif ($line_code == 510) {
            # 510 -- Error reported by the underlying test suite.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Error reported by the test suite",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 520) {
            # 520 -- Message from the underlying test suite.

            # Coalesce consequent rows of spaces in the message
            # to a single space. This improves matching with markers.
            my $message = $line_text;
            $message =~ s/\s+/ /g;

            # Add the message to the accumulated messages.
            $test_messages .= "$message\n";
        }
        elsif ($line_code == 600) {
            # 600 -- Parallel Start.
            # Not interesting to us.
        }
        elsif ($line_code == 620) {
            # 620 -- Parallel End.
            # Not interesting to us.
        }
        elsif ($line_code == 630) {

            # 630 -- Implied Sequential Start.
            # Not interesting to us.
        }
        elsif ($line_code == 640) {
            # 640 -- Implied Sequential End.
            # Not interesting to us.
        }
        elsif ($line_code == 700) {
            # 700 -- Repeat Start.
            # Not interesting to us.
        }
        elsif ($line_code == 720) {
            # 720 -- Repeat End.
            # Not interesting to us.
        }
        elsif ($line_code == 730) {
            # 730 -- Timed Loop Start.
            # Not interesting to us.
        }
        elsif ($line_code == 740) {
            # 740 -- Timed Loop End.
            # Not interesting to us.
        }
        elsif ($line_code == 750) {
            # 750 -- Random Start.
            # Not interesting to us.
        }
        elsif ($line_code == 760) {
            # 760 -- Random End.
            # Not interesting to us.
        }
        elsif ($line_code == 800) {
            # 800 -- Remote Start.
            # Not interesting to us.
        }
        elsif ($line_code == 820) {
            # 820 -- Remote End.
            # Not interesting to us.
        }
        elsif ($line_code == 830) {
            # 830 -- Distributed Start.
            # Not interesting to us.
        }
        elsif ($line_code == 840) {
            # 840 -- Distributed End.
            # Not interesting to us.
        }
        elsif ($line_code == 900) {
            # 900 -- Test Case Controller End.
            # Not interesting to us.
        }
        else {
            # Other lines are notified to the user and otherwise ignored.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Notice",
                "Unrecognized journal entry (code $line_code)",
                "",
                "",
                $line
            );
        }
    }

    # Finish the journal processing and close it.
    close_journal ();
}

#----------------------------------------------------------------------------

# The Tcl-style journal parser function which loads the journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tcl_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $test_number = 1;
    my ($test_name, $test_messages);
    while ($line = <$journal_file>) {

        # A test record message ends with a line with a rough format
        # '==== testname result'
        if ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*FAILED/) {

            unless ($test_name) {

                # Oops, we have missed something...
                print_problem_table_line(
                    $test_number,
                    0,
                    $1,
                    "Failure",
                    "End of test record found but no start was seen",
                    "",
                    "",
                    "",
                    $line
                );
            }

            # Process the result.
            process_test_point_result(
                $test_number,
                0,
                $test_name,
                "Failed",
                $test_messages
            );

            # Clear the previous test messages.
            $test_messages = "";

            # Maintain an artificial sequence of test numbers.
            $test_number++;

            # Finish the test record.
            $test_name = "";
        }

        # A failure report starts with a line with a rough format
        # '==== testname errormessage result'.
        elsif ($line =~ /^=+\s*([a-zA-Z0-9_\-\.]+)\s*(.+)\s*FAILED/) {

            # Load the test name.
            $test_name = $1;

            # Add the error message to the test messages.
            $test_messages .= "$2\n";
        }

        # A line that contains only the name of the test (which ends
        # with '.test', marks a passed test.
        elsif ($line =~ /^([a-zA-Z0-9_\-]+)\.test/) {
            process_test_point_result(
                $test_number,
                0,
                $1,
                "Passed",
                ""
            );

            # Maintain an artificial sequence of test numbers.
            $test_number++;
        }

        # Anything other?
        else {
            if ($test_name) {

                # We are inside a test point result record, so it will be
                # a test message.
                $test_messages .= "$line\n";
            }
            else {

                # Anything other is better ignored (there is too much noise
                # for considering unrecognized lines to be errors).
            }
        }
    }

    close_journal ();
}

#----------------------------------------------------------------------------

# The Python-style journal parser function which loads the journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_python_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages);
    while ($line = <$journal_file>) {

        # Are we reading a test point record?
        if ($inside_test_point) {

            # Any line that resembles a start of another test record
            # stops this test record.
            if ($line =~ /^test_/) {
                $inside_test_point = 0;

                # Process this result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    $test_messages
                );

                # Clear the previous test messages.
                $test_messages = "";
            }
            else {

                # Just append this line to the test record messages
                # and continue by the next line.
                $test_messages .= "$line\n";
                next;
            }
        }

        # Any other line is supposed to end the test point record.
        $inside_test_point = 0;

        # A line that contains only the name of the test (a single
        # identifier), marks a passed test.
        if ($line =~ /^([a-zA-Z0-9_]+)\s*$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Passed",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'testname skipped -- ...' marks a test
        # that was skipped, and the end of the line contains
        # the description why.
        elsif ($line =~ /^([a-zA-Z0-9_]+)\s*skipped\s*--\s*(.*)$/) {
            process_test_point_result(
                $test_number,
                $test_point,
                $1,
                "Unsupported",
                ""
            );

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;
        }

        # A line in form 'test testname failed -- ...' marks a failed test.
        elsif ($line =~ /^test\s*([a-zA-Z0-9_]+)\s*(failed|crashed)\s*--\s*(.*)$/) {

            # Get the test name.
            $test_name = $1;
            $test_point = 1;

            # We are reading the test failure report.
            $inside_test_point = 1;

            # Get the first line of the messages.
            $test_messages = $3;
        }

        # Other lines are silently ignored.
    }

    close_journal ();
}

#----------------------------------------------------------------------------

sub parse_simple_journal
{
    # Open the journal and prepare processing.
    open_journal (@_);
    
    unless ($journal_file) { return; }

    # Read the journal line by line.
    my $severity = "";
    my $inside_test_point = 0;
    my $test_number = 1;
    my ($test_name, $test_point, $test_messages, $test_result);
    while ($line = <$journal_file>) {

        # Each line in form 'test name: result' should contain
        # the result of one test.
        if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)/) {

            # Tests are not naturally numbered so figure out a monotonous sequence.
            $test_number++;

            # Load the name and result.
            $test_name = $1;
            $test_result = $2;

            # An optional detail message can follow.
            if ($line =~ /^\s*([^:]*)\s*:\s*(\w*)\s*:\s*(.*)$/) {
                $test_messages = $3;
            }

            # Check for failures.
            if ($test_result eq "failed") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "Failure",
                    $test_messages
                );
            }
            elsif ($test_result eq "OK") {
                process_test_point_result(
                    $test_number,
                    0,
                    $test_name,
                    "Passed",
                    $test_messages
                );
            }
            else {

                # Anything other is an error in the journal.
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "Failed",
                    "Syntax error in journal",
                    "",
                    "",
                    "",
                    $line
                );
            }
        }
        else {

            # Anything other is an error in the journal.
            print_problem_table_line(
                $test_number,
                0,
                $test_name,
                "Failed",
                "Syntax error in journal",
                "",
                "",
                "",
                $line
            );
        }
    }

    close_journal ();
}
#----------------------------------------------------------------------------

# Writes the information about the system and autotest to the output
# directory.
sub write_test_info
{
    # Write the host name, machine type, platform and OS version, start time.
#    write_string_as_file ( "$results_dir/host_info/host_machine",
#                           "$host_machine\n" );
#
#    write_string_as_file ( "$results_dir/host_info/host_name", 
#                           "$host_name\n" );
#
#    write_string_as_file ( "$results_dir/host_info/host_platform",
#                           "$host_platform\n" );
#
#    write_string_as_file ( "$results_dir/host_info/host_os",
#                           "$host_os\n" );
#
#    write_string_as_file ( "$results_dir/host_info/host_kernel",
#                           "$host_kernel\n" );

    unless(-d "$results_dir/host_info") {
        cmd("mkdir -p $results_dir/host_info");
    }

    write_string_as_file ( "$results_dir/host_info/etc_issue",
                           "$host_etc_issue\n" );

#    write_string_as_file ( "$results_dir/autotest_info/start_time",
#                           format_time($start_time) );

    # Write versions of the installed test packages.
    write_string_as_file ( "$results_dir/host_info/package_versions"
                           , installed_packages() );
}
#----------------------------------------------------------------------------

sub compile_report {
    # Print Title
    $report->print ("<h1>LSB automatic test results for $host_os on $host_platform</h1>\n");
    $report->print ("<h3>Test Info</h3>\n");
    
    # Print general information
    $report->print (
    "<table>\n"
    .two_column_table_line("Test Started At", format_time($start_time))
    .two_column_table_line("Test Finished At", format_time($finish_time))
    .two_column_table_line("Host Name", $host_name)
    .two_column_table_line("Host OS", $host_os)
    .two_column_table_line("Host Machine", $host_machine)
    .two_column_table_line("Host Platform", $host_platform)
    .two_column_table_line("Host Kernel", $host_kernel)
    .two_column_table_line("/etc/issue", "<pre>$host_etc_issue</pre>")
    ."</table>\n"
    );
    
    # Print an index
    $report->print("<h3>Tests executed</h3>");
    $report->print("<table>");
    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        foreach $journal_title ( keys %{$test_obj->report} ) {
            $report->print("<tr>");
            $report->print("<td style=\"background-color:$gray\"><b><a href=\"#".$test_obj->{JOURNAL_ANCHOR}->{$journal_title}
                    ."\">".$journal_title."</a></b></td>");
            my $bg_color = $neutral_gray;
            my $text = "";
            if ($test_obj->{FAILURES_REPORTED}->{$journal_title} > 0) {
                $bg_color = $fluorescent_red;
                $text = "Failures: ".$test_obj->{FAILURES_REPORTED}->{$journal_title};
            }
            elsif ($test_obj->{WARNINGS_REPORTED}->{$journal_title} > 0) {
                $bg_color = $fluorescent_yellow;
                $text = "Warnings: ".$test_obj->{WARNINGS_REPORTED}->{$journal_title};
            }
            elsif ($test_obj->{TESTS_PASSED}->{$journal_title} == 0) {
                $bg_color = $fluorescent_red;
                $text = "Not Tested";
            }
            else {
                $bg_color = $fluorescent_green;
                $text = "Success";
            }

            $report->print("<td style=\"background-color:$bg_color\">$text");
            $report->print("</td><td style=\"background-color:$neutral_gray\">");
            
            $report->print("<span style='color:black'>Passed: ".
                    $test_obj->{TESTS_PASSED}->{$journal_title}."</span>" );

            $report->print("</td></tr>");
        }
    }    
    $report->print("</table>");
    
    # Test journals
    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        foreach $journal_title( keys %{$test_obj->report} ) {
            my $test_report = $test_obj->report->{$journal_title};
            # Print the title.
            $report->print("<hr />\n");
            $report->print("<h2><a name=\"".$test_obj->{JOURNAL_ANCHOR}->{$journal_title}
                    ."\"></a><font color=\"#336699\">".$journal_title."</font></h2>\n");
            
            $report->print ( $test_report );
        }
    }
    
    # Print the host info to the one-line summary.
    $summary_line->print ("<table>\n <tr>\n");
    $summary_line->print ("    <!-- OS version -->       <td class=\"title\">$host_os</td>\n");
    $summary_line->print ("    <!-- Architecture -->     <td class=\"title\">$host_machine</td>\n");
    $summary_line->print (" </tr>\n</table>\n");
}

# Writes the final information about the whole autotest run.
sub write_final_info
{
    # Write the finishing time.
    unless(-d "$results_dir/autotest_info") {
        cmd("mkdir -p $results_dir/autotest_info");
    }
#    write_string_as_file( "$results_dir/autotest_info/finish_time",
#                          format_time($finish_time)  );
}

#----------------------------------------------------------------------------

sub generate_INFO_file {
    my $INFO_file_name = shift;
    
    my $text = ""
        .$stats_tests_total." Tests Total\n"
        .$stats_tests_passed." Tests Passed\n"
        .$stats_failures_reported." Failures\n"
        .$stats_tests_quelled." Tests Quelled\n"
        .$stats_tests_skipped." Tests Skipped\n"
        ;
    
    if ( $tests_cancelled > 0 ) {
        $text .= "There are cancelled tests\n";
    }
    
    $text .= "Verdict: $overall_verdict\n";
    
    $text .= "\n".$test_comments;
    
    write_string_as_file ($INFO_file_name, $text);
}
#----------------------------------------------------------------------------

sub pack_and_send_output
{
    #Go to the output directory
    chdir($results_dir);
    
    # If the output file is not specified, generate a name using
    # the host machine name and the current time.
    unless ($output_file_name) {

        # Get the starting time and break it up to components.
        my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ($start_time);

        # Generate the name for the output directory and file using
        # the host machine type, host name, and the test start time.
        $output_file_name = sprintf(
            "lsb-autotest-%s-%s-%04d-%02d-%02d-%02dh-%02dm.tgz",
            $host_platform, $host_name,
            $year + 1900, $mon + 1, $mday, $hour, $min  );
    }
    
    if ( $output_file_name !~ /\./ ) {
        $output_file_name .= ".tgz";
    }
    
    if ( !$omit_packing || $send_email ) {
        # Create an output archive, and remove the temporary output directory.
        inform ("Creating the output archive");
        my $files = capture("ls");
        cmd ("tar czf $output_file_name *");
        unless ($omit_packing) {
            $files =~ s/[\n\r]+/ /g;
            cmd("rm -rf $files");
        }
    }
    
    unless ($omit_INFO ) {
        # Write INFO file
        my $INFO_file_name = $output_file_name;
        $INFO_file_name =~ s/[^\/]*$/INFO/;
        generate_INFO_file ( $INFO_file_name );
    }
    
    if ( $send_email ) {
        # Send the results by e-mail if requested.
        debug_inform("Sending an e-mail");
        
        # Use sendmail utility
        &sendmail ( { 
            TO      => $tester_email,
            SUBJECT => "LSB autotest results for $host_name",
            TEXT    => "Please see the attached archive for the test results.",
            ATTACHMENT => "$output_file_name",
                    } );
        
        if ( $omit_packing ) {
            # Result archive isn't necessary by now. 
            # It was created just to be sended by e-mail.
            cmd ("rm -f $output_file_name");
        }
    }
    
    if ( $omit_packing ) {
        debug_inform ( "The output is ready: $results_dir/" );
    } else {
        debug_inform ( "The output archive is ready: $results_dir/$output_file_name" );
    }
}

sub Init {
    # Load the problem database (it should be in the same directory
    # as the main program, so simply get the program name and replace
    # the file part by 'problem_db').
    
    my $database_file_name = "$autotest_dir/problem_db";
    load_problem_database($database_file_name);
    
    # Open the report files.
    $report = new IO::File;
    unless ($report->open ("> $results_dir/test_log.html")) {
        fail ("Could not open the report file for writing");
    }
    $report->print ("<style type=\"text/css\">
.ptl_gray { background: #E6E6E6;}
.ptl_gray:hover { background: #D6D6F0;}
.ptl_green { background: #CDFFCD;}
.ptl_green:hover { background: #A0FFA0;}
.ptl_yellow { background: #FFFFCD;}
.ptl_yellow:hover { background: #FFFF90;}
.ptl_red { background: #FFCDCD;}
.ptl_red:hover { background: #FFA0A0;}
.ptl_gray,.ptl_green,.ptl_yellow,.ptl_red { cursor: pointer; }
</style>
    ");
    $report->print ("<html>\n<body>\n");
    $report->print ("
<script language=\"javascript\" type=\"text/javascript\">
// <![CDATA[
function flip (message_box_id) {
    msgbox = document.getElementById(message_box_id);
    if (msgbox.style.display != 'none') {
        msgbox.style.display = 'none';
    }
    else {
        msgbox.style.display = '';
    }
}
// ]]>
</script>
            ");
        
    # Open the one-line-summary file.
    $summary_line = new IO::File;
    unless ($summary_line->open ("> $results_dir/one_line_summary.html")) {
        fail ("Could not open the one-line summary file for writing"
             ." ($results_dir/one_line_summary.html)" );
    }
    $summary_line->print ("<html>\n<body>\n");
}

#----------------------------------------------------------------------------

sub Finish {
    # Write the final information.
    write_final_info();

    # Close the report files.
    $report->print ("</body>\n</html>\n");
    $report->close ();
    
    $summary_line->print ("</body>\n</html>\n");
    $summary_line->close ();

    return $exit_code;
}

#----------------------------------------------------------------------------

sub generate_certification_report {
    # Open the report files.
    my $c_report = new IO::File;
    unless ($c_report->open ("> $results_dir/cert.html")) {
        inform ("Could not open the report file for writing");
        return;
    }
    
    $c_report->print ("<html>\n<body>\n");
    $c_report->print ("<center>\n");

    if ( $overall_verdict eq "Success" ) {
        $c_report->print ("<h1 style=\"width: 70%; background-color: #00c000;"
                ." color: white; font-size: 150%;\">Congratulations!<br />\n"
                ."You have passed the LSB&nbsp;3.1 certification tests!</h1>\n"
        );
    }
    else
    {
        $c_report->print ("<h1 style=\"width: 70%; background-color: #e00000;"
                ." color: white; font-size: 150%;\">Sorry, the certification"
                ." tests have failed.<br />The tests for LSB&nbsp;3.1"
                ." compliance have reported some problems.</h1>"
        );

    }
    
    $c_report->print ("</center>\n");
    $c_report->print ("<h2>Send Results to FSG</h2>\n");
    
    $c_report->print ("<p align=\"justify\">\n");
    if ( $overall_verdict eq "Success" ) {
        $c_report->print ("If you wish, you may send your results to FSG to get"
                ." certified. Please, check your data in the form below and"
                ." press the \"Upload journals\" button.<br />\n"
            
        );
    }
    else
    {
        $c_report->print ("However, if you wish, you may send your results"
                ." to FSG. Please, check your data in the form below and press"
                ." the \"Upload journals\" button.<br />\n"
        );

    }
    $c_report->print ("All the fields below are mandatory, except for the comments.\n");
    
    $c_report->print ("</p>\n");
    $host_os =~ s/"/\\"/g;
    
    my $test_comments_html = $test_comments;
    $test_comments_html =~ s/&/&amp;/g;
    $test_comments_html =~ s/</&lt;/g;
    $test_comments_html =~ s/>/&gt;/g;
    $test_comments_html =~ s/[\n\r]+/<br \/>\n/g;
    
    $c_report->print ("
<form method=\"post\" action=\"tests_send_data.pl\">
<table border=\"0\" cellpadding=\"2\" cellspacing=\"2\" width=\"500px\">
  <tr>
    <td width=\"170px\">FSG Login:</td>
    <td width=\"330px\"><input type=\"text\" name=\"login\" value=\"\" style=\"width: 150px;\" />&nbsp;&nbsp;&nbsp;<a href=\"https://freestandards.org/index.php?title=Special:Userlogin&type=signup\">Not&nbsp;registered?</a></td>
  </tr>
  <tr>
    <td>Password:</td>
    <td><input type=\"password\" name=\"password\" value=\"\" style=\"width: 150px;\" /></td>
  </tr>
  <tr>
    <td>Product:</td>
    <td><input type=\"text\" name=\"product\" value=\"$host_os\" style=\"width: 150px;\" /></td>
  </tr>
  <tr>
    <td>Architecture:</td>
    <td><b>$host_platform</b></td>
  </tr>
  <tr>
    <td>LSB Version:</td>
    <td><b>$lsb_version</b></td>
  </tr>
  <tr>
    <td colspan=\"2\">
      Additional comments:
    </td>
  </tr>
  <tr>
    <td colspan=\"2\">
      <textarea name=\"comments\" rows=\"5\" cols=\"54\">$test_comments_html</textarea>
    </td>
  </tr>
  <tr>
    <td colspan=\"2\">
      <input type=\"button\" value=\"Upload journals\" style=\"width: 150px;\" onclick=\"javascript:alert('Sorry, this function is not implemented yet.\\nThis is only the user interface demo.'); return false;\" />
    </td>
  </tr>
</table>
</form>
<br />
");
    my $results_ref = $results_dir;
    $results_ref =~ m/([^\\\/]+)$/;
    $results_ref = $1;
    if ( $overall_verdict eq "Success" ) {
        $c_report->print ("
<p align=\"justify\">
After you have sent the results to the FSG, you need to proceed with the following steps:
<ol>
  <li>Step 1: ...</li>
  <li>Step 2: ...</li>
  <li>Step 3: ...</li>
</ol>
<br />
</p>
<h2>View detailed report</h2>
<p align=\"justify\">
Here you can see the <a href=\"tests_results.pl?details=$results_ref\"><b>detailed report</b></a> about the tests run.<br />
It contains all the information about what particular tests have been run, how many warnings they reported (if any) and so on.
</p>
");
    }
    else
    {
        $c_report->print ("
<h2>View detailed report</h2>
<p align=\"justify\">
Here you can see the <a href=\"tests_results.pl?details=$results_ref\"><b>detailed report</b></a> about the tests run.<br />
It contains all the information about what particular tests have been run, how many warnings they reported (if any) and so on.
</p>
");
    }
    
    # Close the report files.
    $c_report->print ("</body>\n</html>\n");
    $c_report->close ();
}

#----------------------------------------------------------------------------

# Generates a HTML report and detects test failures.
# Returns the return code of the reporting utility, which is 0 if all tests
# have passed, 1 if some tests have failed.
sub generate_report
{
    inform ("Generating HTML report");
    
    #Go to the output directory
    chdir($results_dir);
    
    system("for name in `ls *.tar.gz`; do "
            ."tar -xzf \$name --overwrite; "
            ."done;"
           );
    cmd("rm -rf *.tar.gz");
    
    # Write the test information.
    write_test_info ();
    
    # Open a report file and print a header
    Init();
    
    # Parse each test journal
    foreach my $test_name (@tests_to_run) 
    {
        $tests{$test_name}->journal ();
    }
    
    # Give the final verdict.
    if ( ($stats_failures_reported > 0) || ($tests_cancelled > 0) ) {
        $overall_verdict = "Failed";
    }
    else {
        $overall_verdict = "Success";
    }
    
    compile_report ();
    
    if ( $certification_mode ) {
        generate_certification_report ();
    }
    
    my $exit_code = Finish();
    if ($exit_code == 0) {
        # Report was generated, no failures were detected.
        return 0;
    }
    elsif (($exit_code >> 8) == 1) {
        # Report was generated, some test failures were detected.
        return 1;
    }
    else {
        fail ("Error generating report (exit code $?)");
    }
}

1; # return value   