#!/usr/bin/perl -w 
# LSB Test Execution Framework
# Main executable script
#
# Copyright (C) 2006 Free Standards Group. All rights reserved.
#
# This program has been developed by ISP RAS for FSG with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2006 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;

# Is debugging mode (for developering purpose)
our $debugging = 0;

# verbose level: -v[0-2]
#   0 - minimal. Only error messages are put into stderr; 
#   1 - default; 
#   2 - extended debug information.
our $verbose_lvl = 1;

# The directory where the script itself is placed.
our $script_directory;

# Options specified on the command line.
my $forced_platform = "";        # --platform
my $forced_triplet = "";         # --triplet
my $forced_package_manager = ""; # --pm
our $tester_email = "";          # --mail-to
our $send_email = 0;
our $output_file_name = "";      # --output
our $certification_mode = 0;
    
our $keep_packages = 0;          # --keep-packages
    $keep_packages = 1 if ($debugging);
my $cleanup = 0;                 # --cleanup
our $download_packages = 0;      # --download
our $try_beta_packages = 0;      # --beta
my $check_only = 0;              # --check-only
my $not_run = 0;                 # --not-run
my $print_help = 0;              # --help
my $print_list = 0;              # --list
our $list_refresh = 0;           # --list-refresh
our $omit_packing = 0;           # --omit-packing
our $omit_INFO = 0;              # --omit-INFO

our $short_scenarios = 0;        # --short
    $short_scenarios = 1 if ($debugging);

# Local directory where packages are located.
our $package_dir = "";           # -d, --package-dir
    
# Properties of the host machine.
our $host_machine = "";
our $host_platform = "";
our $host_name = "";
our $host_os = "";
our $host_kernel = "";
our $host_triplet = "";
our $host_etc_issue = "";

# Components release version
our $lsb_version = "";

# Array of tests to run; they will be run in the specified order. 
our @tests_to_run;

# Test -> download mode
our %test_mode;

# Test names -> test objects.
our %tests;

# How the wget tool should be called.
# -r, --recursive         specify recursive download
# -nd, --no-directories   don't create directories
# -np, --no-parents       don't ascend to the parent directory.
# -w, --wait= SECONDS     wait SECONDS between retrievals
# --retr-symlinks         when recursing, get linked-to files, (not dir)
our $wget = "wget --timestamping -r -nd -np -w 3 --retr-symlinks --timeout=120 ";

# The time when the test was started.
our $start_time;
our $finish_time;

# Information about launch.
our $tester_name = "";
our $organization = "";
our $test_comments = "";

# Directory where reports will be stored
# path should be absolute or relative to /tmp/lsb-autotest-temp
our $results_dir = "";

# The directory where autotest is placed
our $autotest_dir = "./../autotest";

# A command to run autotest. Will be generated dynamically.
our $autotest_cmd = "";
#----------------------------------------------------------------------------

# Ensures that the specified path is absolute, prepending the current
# working directory to relative paths.
sub make_path_absolute
{
    my $input_path = shift;

    # Determine the current working directory.
    my $working_directory = `pwd`;
    chomp ($working_directory);

    # If the path is absolute (starts with a slash), leave it as it is.
    my $output_path;
    if ($input_path =~ /^\//) {
        $output_path = $input_path;
    }

    # If the path is relative (does not start with a slash), make it
    # absolute by prepending the current working directory.
    else {
        $output_path .= $working_directory . "/" . $input_path;
    }

    return $output_path;
}
#----------------------------------------------------------------------------

# Determines the directory the script is executed from,
# and stores it to the global 'script_directory' variable.
sub find_script_directory
{
    $script_directory = $0;
    $script_directory =~ s/[^\/]*$//;
    $script_directory = make_path_absolute ($script_directory);
    $autotest_dir = "$script_directory/../autotest";
}
#----------------------------------------------------------------------------

sub BEGIN {
    $| = 1; # Immediate flush of all output

    # LSB 3.1 / 16.4. Path For System Administration Utilities
    # Certain utilities used for system administration (and other privileged
    # commands) may be stored in /sbin, /usr/sbin, and /usr/local/sbin. 
    # Applications requiring to use commands identified as system 
    # administration utilities should add these directories to their PATH.
    $ENV{PATH} .= ":/sbin:/usr/sbin:/usr/local/sbin";

    # Append the script directory to the @INC to attach our own modules
    find_script_directory();
    push @INC, $script_directory;
}

use Common;
use Journal;
use Packages;
#=============================================================================
# Subroutines of the main script
#=============================================================================

sub search_for_test_modules
{
    my $ref = shift; # relative path
    my $dir = $script_directory."Tests/".$ref; 
    
    my @found_tests;
    
    # Open directory
    opendir DIR, $dir or die "Can't open directory $dir: $!\n";
    
    # Read directory for files and dirs
    my @files = readdir(DIR);
    
    foreach my $file(@files)
    {
        next if $file =~ /^\./; # skip hidden files, . and ..
        
        if (-f "$dir/$file")   # -f checks it is a file
        {
            next if( $file !~ /_test\.pm$/ );    # Search only for *_test.pm files.
            
            push @found_tests, $ref.$`;
        }
        elsif (-d "$dir/$file")  # -d checks it is a directory
        {
            push @found_tests, &search_for_test_modules($ref.$file."/"); # Do recursive search
        }
    }
    closedir DIR;
    return @found_tests;
}

#----------------------------------------------------------------------------

# Prints a short help text.
sub print_help
{
    my @test_names;
    
    foreach(values %tests) {
        push @test_names, $_->name;
    }
    
    print("
    Usage: lsb-autotest [OPTION] TESTNAME...
    Runs one or more of LSB tests and produces a HTML report.

    WARNING! PLEASE DO NOT USE THIS SCRIPT ON ANY MACHINE THAT HOLDS
    IMPORTANT DATA, OR PROVIDES IMPORTANT SERVICES! The tests
    sometimes put a heavy stress on the tested computer,
    and in rare unfortunate cases they can crash the system or damage
    the stored data. They also are a security risk as they pass along
    the root password.

    Tests  available:
        @test_names

    \'all\' means all available tests.
    \'appbat\' means app-battery tests.
    \'runtime-tests\' means runtime tests.
        
    Additional options:

    -B, --beta               Use beta packages if available.
    --cleanup                Just uninstall test packages and exit.
    -d, --package-dir        Directory for the test packages.
    -r, --results-dir        Directory for results.
    -D, --download           Automatically download needed test packages.
    -h, --help               Print this help and exit.
    -k, --keep-packages      Do not uninstall installed test packages.
    -m, --mail-to            Send the results by e-mail to given address.
    -o, --output             Output file ('.tar.gz' will be appended).
    -p, --platform           Assume specified host platform.
    -t, --triplet            Assume specified host triplet.
    -pm                      Point out the proper package manager (RPM|dpkg)
    -v[0-2]                  Verbose level. 0 means only errors output,
                             1 - default output,
                             2 - extra output (for debug purposes).
    --check-only             Do some checks without installing and running 
                             any tests.
    --not-run                Install tests but do not run them.
    -f, --profile            Get options set from the profile.
    -op, --omit-packaging    Do not pack results.
    -oI, --omit-INFO         Do not create INFO file.

    Please report any bugs and improvements to fsgteam\@ispras.ru.
");
}
#----------------------------------------------------------------------------

# Load the command line options.
sub read_command_line
{
    my $package_dir_expected = 0;
    my $forced_platform_expected = 0;
    my $forced_triplet_expected = 0;
    my $forced_package_manager_expected = 0;
    my $tester_email_expected = 0;
    my $output_file_expected = 0;
    my $profile_expected = 0;
    my $results_dir_expected = 0;
    my $profile = "";

    foreach my $arg (@ARGV) {
        # Handle arguments for options which have one.

        if ($package_dir_expected) {
            $package_dir = $arg;
            $package_dir_expected = 0;
        }
        elsif ($forced_platform_expected) {
            $forced_platform = $arg;
            $forced_platform_expected = 0;
        }
        elsif ($forced_triplet_expected) {
            $forced_triplet = $arg;
            $forced_triplet_expected = 0;
        }
        elsif ($forced_package_manager_expected) {
            $forced_package_manager = $arg;
            $forced_package_manager_expected = 0;
        }
        elsif ($tester_email_expected) {
            $tester_email = $arg;
            $tester_email_expected = 0;
            $send_email = 1;
        }
        elsif ($output_file_expected) {
            $output_file_name = $arg;
            $output_file_expected = 0;
        }
        elsif ($profile_expected) {
            $profile = $arg;
            $profile_expected = 0;
            
            read_profile ($profile);
        }

        # Anything other is considered to be a command line option.

        else {
            if (($arg eq "-B") || ($arg eq "--beta")) {
                $try_beta_packages = 1;
            }
            elsif ($arg eq "--cleanup") {
                $cleanup = 1;
            }
            elsif (($arg eq "-d") || ($arg eq "--package-dir")) {
                $package_dir_expected = 1;
            }
            elsif ($arg =~ "--package-dir=(.*)") {
                $package_dir = $1;
            }
            elsif (($arg eq "-D") || ($arg eq "--download")) {
                $download_packages = 1;
            }
            elsif (($arg eq "-h") || ($arg eq "--help")) {
                $print_help = 1;
            }
            elsif (($arg eq "-k") || ($arg eq "--keep-packages") ) {
                $keep_packages = 1;
            }
            elsif (($arg eq "-m") || ($arg eq "--mail-to") || ($arg eq "--mailto")) {
                $tester_email_expected = 1;
            }
            elsif (($arg =~ "--mail-to=(.*)") || ($arg =~ "--mailto=(.*)")) {
                $tester_email = $1;
                $send_email = 1;
            }
            elsif (($arg eq "-o") || ($arg eq "--output")) {
                $output_file_expected = 1;
            }
            elsif ($arg =~ "--output=(.*)") {
                $output_file_name = $1;
            }
            elsif (($arg eq "-p") || ($arg eq "--platform")) {
                $forced_platform_expected = 1;
            }
            elsif ($arg =~ "--platform=(.*)") {
                $forced_platform = $1;
            }
            elsif (($arg eq "-t") || ($arg eq "--triplet")) {
                $forced_triplet_expected = 1;
            }
            elsif ($arg =~ "--triplet=(.*)") {
                $forced_triplet = $1;
            }
            elsif ($arg eq "--pm") {
                $forced_package_manager_expected = 1;
            }
            elsif ($arg =~ "--pm=(.*)") {
                $forced_package_manager = $1;
            }
            elsif ($arg =~ "-v([0-2])") {
                $verbose_lvl = $1;
            }
            elsif ($arg =~ "--check-only") {
                $check_only = 1;
            }
            elsif ($arg =~ "--not-run") {
                $not_run = 1;
            }
            elsif ($arg eq "--list") {
                $print_list = 1;
            }
            elsif ($arg eq "--list-refresh") {
                $print_list = 1;
                $list_refresh = 1;
                $download_packages = 1;
            }
            elsif ($arg eq "--list-refresh-only") {
                $list_refresh = 1;
                $download_packages = 1;
            }
            elsif (($arg eq "-f") || ($arg eq "--profile")) {
                $profile_expected = 1;
            }
            elsif ($arg =~ "--profile=(.*)") {
                $profile = $1;
                
                read_profile ($profile);
            }
            elsif (($arg eq "-r") || ($arg eq "--results-dir")) {
                $results_dir_expected = 1;
            }
            elsif ($arg =~ "--results-dir=(.*)") {
                $results_dir = $1;
            }
            elsif (($arg eq "-op") || ($arg eq "--omit-packing")) {
                $omit_packing = 1;
            }
            elsif (($arg eq "-oI") || ($arg eq "--omit-INFO")) {
                $omit_INFO = 1;
            }
            elsif ($arg =~ "--lsb=(.*)") {
                $lsb_version = $1;
            }
            elsif ($arg =~ "--short") {
                $short_scenarios = 1;
            }

            # 'all' means running all available tests (some tests may be
            # available only in beta).
            elsif ($arg eq "all") {
                # Use all released tests.
                @tests_to_run = keys %tests;
            }

            # 'appbat' means running the app battery tests.
            elsif ($arg eq "appbat") {
                push (@tests_to_run, "apache", "expect", "groff",
                    "lynx", "python", "rsync", "tcl", "xpdf");
            }
            elsif ($arg eq "runtime-tests") {
                push (@tests_to_run, "cmdchk", "libchk", "libstdc++", "runtime" );
            }
            elsif ($arg eq "libstdcpp") {
                push (@tests_to_run, "libstdc++");
            }
            # These parameters determine which tests to run.
            elsif ( $tests{lc($arg)} )  { push (@tests_to_run, "$arg"); }

            else {

                # Invalid argument.
                fail_nomail ("Unrecognized argument $arg (try '--help').");
            }
        }
    }

    # Check for missing argument at the end.

    if (
        $package_dir_expected or
        $forced_platform_expected or
        $forced_triplet_expected or
        $tester_email_expected or
        $forced_package_manager_expected or
        $output_file_expected or
        $profile_expected or
        $results_dir_expected
    ) {
        fail_nomail ("Missing commandline argument (try '--help')");
    }
    
    if ( $certification_mode ) {
        $short_scenarios = 0;
            $short_scenarios = 1 if ($debugging);
        $try_beta_packages = 0;
        $download_packages = 1;
        
        # Use all released tests.
        @tests_to_run = keys %tests;
            @tests_to_run = ("cmdchk") if ($debugging);
        
        %test_mode = ();
    }
}
#----------------------------------------------------------------------------

# Does initial environment checks before anything else is done.
# Terminates the script if some of the checks fails.
sub do_initial_checks
{
    # Complain if we are not run as root.
    if ($> != 0) {
        fail_nomail ("This test must be run as root.");
    }
    
    # Check for basic utilities we will need.
    check_for_tool ("lsb_release");
    #check_for_tool ("gcc");
    check_for_tool ("passwd");
    check_for_tool ("uname");
    {
        # Check whether the uname utility knows about the '-m' option.
        if (system ("uname --help | grep -- ' -m\\W' >/dev/null") != 0
            && !$forced_platform ) {
            fail_nomail ("It seems that the uname utility does not support " .
                "-m option;".
                "\nYou should explain the host platform manually using '-p' option.");
        }
    }
    
    # wget is an optional tool for downloading packages.
    if($download_packages) { 
        # Look if the shell knows this tool.
        unless( does_shell_know ("wget") ) {
            fail_nomail ("Could not find a tool: wget.\nIf you haven't".
                " this tool, use -d option and do not use -D options.");
        }
    }
    
    # RPM is required for any package manager
    check_for_tool ("rpm");
    if ( $host_package_manager eq "dpkg") {
        # Check for alien and dpkg
        check_for_tool("alien");
        check_for_tool("dpkg");
    }
    
    if ( $send_email ) {
        check_for_tool("/usr/sbin/sendmail");
    }
    
    # Tests custom checks.
    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        debug_inform($test_obj->name." check...");
        $test_obj->check();
    }
}

#----------------------------------------------------------------------------

# This function creates objects, which work with test suites.
sub create_test_objects {
    
    my @test_files = &search_for_test_modules("");

    foreach my $test_name (@test_files) {
        $test_name = "\u${test_name}_test";
        my $test_module_name = "Tests::$test_name";
        $test_module_name =~ s/[\/\\]/::/g;
        $test_name = &extract_filename($test_name);
        my $test_obj;
        
        # Link up the module and create object.
        eval "use $test_module_name; \$test_obj = $test_name->new();";
        if ($@) { # Any errors?
            complain($@);
            exit(2);
        }
        
        # Check the test availability
        if ( $test_obj->is_available ) {
            # Store created objects to a hash
            $tests{$test_obj->name} = $test_obj;
        }
    }
}
#----------------------------------------------------------------------------

# Removes previously installed test packages.
sub purge_packages {

    # Collect all package names to a set, then purge it
    my %packages_to_purge;
    my @packages_to_purge;

    foreach my $test_obj (values %tests) 
    {
        foreach my $package_name ( split(" ",$test_obj->packages))
        { 
            if(!$packages_to_purge{$package_name})
            {
                $packages_to_purge{$package_name} = 1;
                push @packages_to_purge, $package_name;
            }
        
        }
        
        #$test_obj->purge();
    }
    
    # Deinstall in reverse order
    @packages_to_purge = reverse @packages_to_purge;
    
    my $num_to_uninstall;
    
    # If removing fails, e. g. due to dependencies, then retry
    # until number of packages stops to decrease.
    do {
        $num_to_uninstall = @packages_to_purge;
        
        for (my $i = 0; $i<$num_to_uninstall; $i++)
        {
            my $package_to_purge = shift @packages_to_purge;

            inform ("Purging package $package_to_purge...");
            
            # Uninstalls the specified package if it is installed.
            remove_package($package_to_purge);
            
            # Check if the package was really uninstalled
            if ( !package_info($package_to_purge) )
            {
                inform ("Ok.");
            }
            else
            {
                inform ("Fail.");
                
                # Retry to uninstall the packages, which haven't uninstalled properly
                push @packages_to_purge, $package_to_purge;
            }
        }
        # Retry to uninstall the packages, which haven't uninstalled properly
    } while (@packages_to_purge < $num_to_uninstall);
}
#----------------------------------------------------------------------------

# Reads information about the host machine.
sub read_host_info
{
    # Check for basic utilities we will need here.
    check_for_tool ("lsb_release");
    check_for_tool ("uname");
    
    # The machine type and architecture is discovered via uname.
    $host_machine = `uname -m`; chomp ($host_machine);
    $host_name = `uname -n`; chomp ($host_name);
    
    unless ($forced_platform) {
        # Discover the host platform.
        
        # If uname utility supports -i options then use it
        if (system ("uname --help | grep -- ' -i\\W' >/dev/null") == 0)
        {
            $host_platform = `uname -i`; chomp ($host_platform);
            $host_platform = guess_platform($host_platform);
        }
        
        if ( !$host_platform || $host_platform =~ "unknown" )
        {
            # Try to find out the host platform using the host_machine.
            $host_platform = guess_platform($host_machine);
        }
        
        if ( !$host_platform || $host_platform =~ "unknown" )
        { fail_nomail("Could not recognize the host platform.\n".
            "You should explain the host platform manually ".
            "using '-p' option."); 
        }
    }

    $host_kernel = `uname -r -v`; chomp ($host_kernel);

    # The OS version is determined via the lsb_release tool
    # (this should be the canonical, distro independent way).
    $host_os = `lsb_release -s -d`;
    $host_os =~ s/\"(.*)\"/$1/;    # remove quotes
    chomp($host_os);               # remove trailing whitespace
    
    if ( !$lsb_version ) {
        # Trying to detect supported LSB version
        my $lsb_release_v = `lsb_release -s -v`;
        if    ( $lsb_release_v =~ /4\.0/ ) { $lsb_version = "4.0"; }
        elsif ( $lsb_release_v =~ /3\.2/ ) { $lsb_version = "3.2"; }
        elsif ( $lsb_release_v =~ /3\.1/ ) { $lsb_version = "3.1"; }
        elsif ( $lsb_release_v =~ /3\.0/ ) { $lsb_version = "3.0"; }
        elsif ( $lsb_release_v =~ /2\.0/ ) { $lsb_version = "2.0"; }
        else { $lsb_version = "3.1"; }
    }
    # Guess the host description triplet.
    $host_triplet = "$host_machine\-gnu-linux";

    # Read the /etc/issue file, if present. Sometimes it is the only place
    # where an exact OS version can be found.
    $host_etc_issue = `cat /etc/issue`;

    # If the '--platform' or '--triplet' option is in effect,
    # change the autodetected platform and/or triplet to the requested value.
    if ($forced_platform) {
        inform ("Platform manually set to: $forced_platform");
        $host_platform = $forced_platform;
    }
    if ($forced_triplet) {
        inform ("Host triplet manually set to: $forced_triplet");
        $host_triplet = $forced_triplet;
    }
    
    $host_package_manager = guess_package_manager();

    if ($forced_package_manager) {
        inform ("Package manager manually set to: $forced_package_manager");
        $host_package_manager = $forced_package_manager;
        
        if( $host_package_manager !~ /(RPM|dpkg)/ ) {
            $host_package_manager = "dpkg";
        }
    }
        
    if(!$host_package_manager || $host_package_manager eq "unknown") {
        # Unknown distro
        fail_nomail("read_host_info: couldn't recognize package manager"
          ." for this distro ($host_os). Use --pm=(RPM|dpkg) key to define it manually.");
    }
    
    inform("Host package manager: $host_package_manager");
}
#----------------------------------------------------------------------------

# Tries to set the package source to the specified directory, provided
# that no package source was set before, and the directory exists.
sub try_package_directory
{
    my $directory_name = shift;

    # Do nothing if the package source is already set.
    if ($package_dir) {
        return;
    }

    # Try if the directory exists, if it is, use it.
    if (-d $directory_name) {
        $package_dir = $directory_name;
        inform ("Using package directory: $directory_name");
    }
}

#----------------------------------------------------------------------------

# Finds where the packages are (or should be) stored.
sub prepare_package_directory
{
    if ($package_dir) {

        # Ensure that the package dir path is absolute, so that
        # we do not depend on the working directory.
        $package_dir = make_path_absolute($package_dir);
    }
    else {

        # If no package source is set, look for some default locations.
        try_package_directory ("/var/opt/lsb-tef/lsb-packages");
        #try_package_directory ("$script_directory/packages");
        #try_package_directory ("$script_directory/../lsb-packages");
        #try_package_directory ("/tmp/lsb-packages");
        
        # Not found?
        unless ($package_dir) {

            # Create directory for packages in a default place
            $package_dir = "/var/opt/lsb-tef/lsb-packages";
            inform ("Using package directory: $package_dir");
            unless (-d $package_dir) { cmd ("mkdir -p \"$package_dir\""); }

            # Otherwise there is no way to get the packages, so stop.
            unless (-d $package_dir) {
                fail_nomail ("Can't create packages directory"
                    ." (Have you enough rights for it?)."
                    ."\nUse '--package-dir' option to specify another"
                    ." packages directory" );
            }
        }
    }
}
#----------------------------------------------------------------------------

# Does all test-dependent cleanups
sub tests_cleanup 
{
    foreach my $test_name (@tests_to_run) 
    {
        if($tests{$test_name})
            { $tests{$test_name}->cleanup (); }
    }
}
#----------------------------------------------------------------------------

# Does all necessary cleanups at the end of the autotest run.
# This must be called at the program termination, otherwise we are leaving
# the system in unstable and insecure state.
sub cleanup
{
    debug_inform("-- Cleanup:");
    
    tests_cleanup();
}
#----------------------------------------------------------------------------

sub read_profile {
    my $profile = shift;
    
    open (FILE, $profile) or fail("Could not open profile $profile");
    
    while (my $line = <FILE> )
    {
        chomp($line);
        
        if( $line =~ /^\s*NAME:\s+(.*)$/ ) {
            $tester_name = $1;
        }
        elsif( $line =~ /^\s*ORGANIZATION:\s+(.*)$/ ) {
            $organization = $1;
        }
        elsif( $line =~ /^\s*EMAIL:\s+(.*)$/ ) {
            $tester_email = $1;
        }
        elsif( $line =~ /^\s*SEND_EMAIL:\s+([0-1])\s*$/ ) {
            $send_email = $1;
        }
        elsif( $line =~ /^\s*VERBOSE_LEVEL:\s+([0-2])\s*$/ ) {
            $verbose_lvl = $1;
        }
        elsif( $line =~ /^\s*KEEP_PACKAGES:\s+([0-1])\s*$/ ) {
            $keep_packages = $1;
        }
        elsif( $line =~ /^\s*ARCHITECTURE:\s+(.*)$/ ) {
            $forced_platform = $1;
        }
        elsif( $line =~ /^\s*USE_INTERNET:\s+([0-1])\s*$/ ) {
            $download_packages = $1;
        }        
        elsif( $line =~ /^\s*COMMENTS:\s+(.*)$/ ) {
            $test_comments .= "$1\n";
        }
        elsif( $line =~ /^\s*SHORT_SCENARIOS:\s+(.*)$/ ) {
            $short_scenarios = $1;
        }
        elsif( $line =~ /^\s*LSB_VERSION:\s+(.*)$/ ) {
            $lsb_version = $1;
        }
        elsif( $line =~ /^\s*CERTIFY:\s+(.*)$/ ) {
            $certification_mode = $1;
        }
        elsif ( $line =~ /^\s*(\S*)\s+([pab])\s*$/ ) {
            my $test_name = $1;
            my $mode = $2;
            
            $test_name =~ s/libstdcpp/libstdc++/;
            
            push @tests_to_run, $test_name;
            
            $test_mode{$test_name} = $mode;
        }
        else {
            #fail_nomail( "Syntax error in profile $profile;\nLine: $.: $line" );
        }
    }
    
    close (FILE);
}

#============================================================================
# The main script
#============================================================================

# Find our own directory. 
find_script_directory ();

# Create test objects according to @tests_to_run list.
create_test_objects ();

# Read the command line.
read_command_line ();

# if '--help' switch is used, print help and exit.
if($print_help)
{
    &print_help();
    exit(0);
}

# Set Wget verbose level
if ($verbose_lvl == 0) 
    { $wget .= " -q"; }

# Setup Carp module.
if ($verbose_lvl >= 2)
    { $SIG{__DIE__} = sub { require Carp; Carp::confess(@_) } };

# Read information about the host machine.
read_host_info (); # cleanup should know package manager

inform ("Script directory is $script_directory");

# Create symlink on the directory with packages
prepare_package_directory ();

if($print_list || $list_refresh) {
    &list_files ("b");
    if ($print_list) {
        &print_files_list ();
    }
    exit(0);
}

inform ("Host name: $host_name");
inform ("Host machine: $host_machine");
inform ("Host platform: $host_platform");
inform ("Host kernel: $host_kernel");
inform ("Host OS: $host_os");
inform ("Host triplet: $host_triplet");

if (@tests_to_run) {
    inform ("Tests to be run: @tests_to_run");
} else {
    fail_nomail ("No tests were chosen (use 'all' to execute all tests).");
}

# Do some initial environment checking.
do_initial_checks ();

# If --check-only option was provided, then do not continue execution.
if($check_only) { exit(0); }

# If the '--cleanup' switch is used, just purge all packages and temporary
# files, then exit.
if ($cleanup) {

    inform ("Purging test packages");
    purge_packages ();

    inform ("Purging temporary files");
    cmd ("rm -rf /tmp/lsb-autotest-*");

    inform ("Finished");
    exit (0);
}

# Check for autotest and make it if necessary
#if (-f "$autotest_dir/Makefile") {
#    inform("Making autotest");
#    run_command("cd $autotest_dir; make");
#}

unless (   (-x "$autotest_dir/lsb-autotest") 
        && (-x "$autotest_dir/lsb-autotest-reporter") ) {
    fail("Can't find lsb-autotest and lsb-autotest-reporter in $autotest_dir");
}


# Look for available packages
inform("Looking for available packages");
inquire_for_files_list ();

# Get the time when the tests were started.
$start_time = time ();

# Prepare a dir for results
unless( $results_dir ) {
    $results_dir = "/var/opt/lsb-tef/results";
}

my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ($start_time);
my $last_result_dir = sprintf( "%s-%s-%04d-%02d-%02d-%02dh-%02dm-%02ds",
  $host_platform, $host_name, $year + 1900, $mon + 1, $mday, $hour, $min, $sec );
$results_dir .= "/$last_result_dir";

unless (-d $results_dir) {
    run_command( "mkdir -p \"$results_dir\"" ); 
}

debug_inform ("Result directory is $results_dir");

unless ( $not_run ) {

    foreach my $test_name (@tests_to_run) 
    {
        my $test_obj = $tests{$test_name};
        
        inform("Prepare $test_name test");
        my $result = prepare_test ($test_obj);
        
        if ($result != 0) {
            inform ("error: Failed to prepare test $test_name. Test will be skipped");
            next;
        }
        
        # Remember start time
        $test_obj->{START_TIME} = time ();

        # Compile command line for the autotest.
        my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday)
            = localtime ( $test_obj->{START_TIME} );
        
        $output_file_name = sprintf(
            "%s-%s-%s-%04d-%02d-%02d-%02dh-%02dm",
            $host_platform, $host_name, $test_name, 
            $year + 1900, $mon + 1, $mday, $hour, $min
        );

        $autotest_cmd = 
            "$autotest_dir/lsb-autotest -v --trust-packages --keep-packages --beta --no-download".
            " -p \"$host_platform\" -o \"$results_dir/$output_file_name\"";

        # Run the test
        inform("Run $test_name test");
        $result = $tests{$test_name}->run ();
        
        # Remember finish time
        $test_obj->{FINISH_TIME} = time ();

        debug_inform("$test_name has returned code $result");
        if( $result == 111 ) {
            inform("Stop tests execution");
            last;
        }
                
        unless ($keep_packages) {

            # Uninstall test packages.
            # This ensures that all config files will be fresh on the next
            # start and also that all packages really belong to the selected 
            # platform (on multi-platform machines).
            inform ("Purging old packages");
            purge_packages ();
        }
    }
}

# Get the time when all the tests were finished.
$finish_time = time();

# New reporter. Disabled yet.
#cmd ("if cd $results_dir; then $autotest_dir/lsb-autotest-reporter .;fi");
###

# Generate a HTML report and detect test failures.
# Pack the output and optionally send it by e-mail.
my $test_failures_detected = generate_report ();

# Pack the output and optionally send it by e-mail.
pack_and_send_output();

inform ("Finished");
# Finish and return 0 if no tests failed, or 1 if there were some failures.
#exit ($test_failures_detected);
# Finish and always return 0
exit (0);

sub END
{
    my $errcode = $?;
    
    # Remove all temporary changes
    cleanup();
    
    $? = $errcode;
}