/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "file_utils.h"
#include "inform.h"
#include "string_utils.h"
#include "run_command.h"
#include <stdio.h>
#include <poll.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

/* Should be defined in stdlib.h but somehow it is not. */
char * mkdtemp(char * template);

/*-------------------------------------------------------------------------*/

void write_string_as_file(const char * filename, const char * text, ...)
{
    FILE * file = NULL;
    va_list args;

    assert(filename);
    assert(text);

    /* Open the file for writing. */
    file = fopen(filename, "w");
    if (file == NULL) {
        fail("Could not open file '%s' for writing: %m", filename);
    }

    /* Write the string with printf-like formatting. */
    va_start(args, text);
    if (vfprintf(file, text, args) < 0) {
        fail("Error writing to file: %s: %m", filename);
    }
    va_end(args);

    /* Close the file. */
    fclose(file);
}

/*-------------------------------------------------------------------------*/

void append_string_to_file(const char * filename, const char * text, ...)
{
    FILE * file = NULL;
    va_list args;

    assert(filename);
    assert(text);

    /* Open the file for appending. */
    file = fopen(filename, "a");
    if (file == NULL) {
        fail("Could not open file '%s' for writing: %m", filename);
    }

    /* Write the string with printf-like formatting. */
    va_start(args, text);
    if (vfprintf(file, text, args) < 0) {
        fail("Error writing to file: %s: %m", filename);
    }
    va_end(args);

    /* Close the file. */
    fclose(file);
}

/*-------------------------------------------------------------------------*/

char * read_first_line_from_file(const char * filename)
{
    FILE * file = NULL;
    char * result = NULL;
    int result_size = 0;

    assert(filename);

    /* Open the file for reading. */
    file = fopen(filename, "r");
    if (file == NULL) {

        /* Could not open the file. */
        return NULL;
    }

    /* Read the first line. */
    if (read_line(file, &result, &result_size) == 0) {
        return NULL;
    }

    /* Done. Close the file and return the resulting string. */
    fclose(file);
    return result;
}

/*-------------------------------------------------------------------------*/

int file_exists(const char * filename)
{
    struct stat stat_buffer;

    assert(filename);

    /* Attempt to stat the file. */
    if (stat(filename, &stat_buffer) == 0) {

        /* The file exists. */
        return 1;
    }

    if (errno == ENOENT) {

        /* The file was not found. */
        return 0;
    }

    /* An error has occurred. */
    fail("Could not stat file '%s': %m", filename);
    return 0;   /* never happens */
}

/*-------------------------------------------------------------------------*/

int is_directory(const char * filename)
{
    struct stat stat_buffer;

    /* Look at the file. */
    if (stat(filename, &stat_buffer) == 0) {

        /* Exists, check if it is a directory. */
        return (S_ISDIR(stat_buffer.st_mode));
    }

    /* Does not exist, or is not a directory, or an error has occurred. */
    return 0;
}

/*-------------------------------------------------------------------------*/

int is_regular_file(const char * filename)
{
    struct stat stat_buffer;

    /* Look at the file. */
    if (stat(filename, &stat_buffer) == 0) {

        /* Exists, check if it is a directory. */
        return (S_ISREG(stat_buffer.st_mode));
    }

    /* Does not exist, or is not a directory, or an error has occurred. */
    return 0;
}

/*-------------------------------------------------------------------------*/

int has_input_data(int fd)
{
    struct pollfd fds[1];
    int result;

    /* Test if the file descriptor is ready for reading. */
    fds[0].fd = fd;
    fds[0].events = POLLIN;
    result = poll(fds, 1, 0);

    /* Check the results and return. */
    if (result == 0) { return 0; }
    else if (result > 0) { return 1; }
    else {
        fail("I/O error when polling on file descriptor: %m");
        return 0;  /* never happens */
    }
}

/*-------------------------------------------------------------------------*/

int wait_for_input(int fd, int timeout_sec)
{
    struct pollfd fds[1];
    int result;

    /* Test if the file descriptor is ready for reading. */
    fds[0].fd = fd;
    fds[0].events = POLLIN;
    result = poll(fds, 1, timeout_sec * 1000);

    /* Check the results and return. */
    if (result == 0) {
        return 0;
    }
    else if (result > 0) {
        return 1;
    }

    fail("I/O error when polling on file descriptor: %m");
    return 0;      /* never happens */
}

/*-------------------------------------------------------------------------*/

void remove_dir_recursive(const char * name)
{
    assert(name);

    /* Remove the directory if it exists. */
    if (file_exists(name)) {
        run_command("rm -rf %s", name);
    }
}

/*-------------------------------------------------------------------------*/

void create_dir(const char * name, ...)
{
    va_list args;
    char buffer[COMMAND_BUFFER_SIZE];

    assert(name);

    /* Print the formatted name to the buffer. */
    va_start(args, name);
    vsnprintf(buffer, COMMAND_BUFFER_SIZE, name, args);
    va_end(args);

    /* Create the directory. */
    if (mkdir(buffer, 0755) != 0) {
        fail("Could not create directory: %s: %m", name);
    }
}

/*-------------------------------------------------------------------------*/

char * create_temp_dir(const char * name_prefix)
{
    char * name = format_malloc("%sXXXXXX", name_prefix);
    if (mkdtemp(name) == NULL) {
        fail("Could not create temporary directory: %m", name);
    }
    return name;
}

/*-------------------------------------------------------------------------*/

char * make_filename_absolute(const char * filename)
{
    char * current_dir;
    char * result;

    assert(filename);

    if (string_starts_with(filename, "/")) {

        /* The filename is already absolute. */
        return strdup(filename);
    }

    /* Prepend the current directory to the filename. */
    current_dir = getcwd(NULL, 0);
    result = format_malloc("%s/%s", current_dir, filename);
    free(current_dir);

    return result;
}

/*-------------------------------------------------------------------------*/

FILE * open_file_for_reading(const char * filename)
{
    FILE * f = fopen(filename, "r");
    if (f == NULL) {
        fail("Could not open file \"%s\" for reading: %m", filename);
    }
    return f;
}

/*-------------------------------------------------------------------------*/

FILE * open_file_for_writing(const char * filename)
{
    FILE * f = fopen(filename, "w");
    if (f == NULL) {
        fail("Could not open file \"%s\" for writing: %m", filename);
    }
    return f;
}

/*-------------------------------------------------------------------------*/

int read_line(FILE * f, char ** buffer, int * buffer_size)
{
    int chars_read = 0;
    int c;

    assert(f);
    assert(buffer);
    assert(buffer_size);

    /* Fail right away if the end of file was already reached,
     * of if an I/O error has occurred.
     */
    if (feof(f) || ferror(f)) {
        free(*buffer);
        *buffer = NULL;
        *buffer_size = 0;
        return 0;
    }

    if (*buffer_size == 0) {

        /* Allocate the initial buffer (128 could be enough). */
        *buffer_size = 128;
        *buffer = (char *) malloc(*buffer_size);
        assert(*buffer);
    }

    /* Read characters until a newline or EOF is found. */
    for (;;) {
        c = getc(f);

        /* Stop on a newline or on an EOF. */
        if ((c < 0) || (c == '\n')) { break; }

        /* If the buffer is full, enlarge it. */
        assert(*buffer_size >= chars_read + 1);
        if (*buffer_size == chars_read + 1) {
            *buffer_size *= 2;
            *buffer = (char *) realloc(*buffer, *buffer_size);
            assert(*buffer);
        }

        /* Append the character to the buffer. */
        (*buffer)[chars_read] = (char) c;
        chars_read++;
    }

    /* Add the terminating zero. */
    assert(*buffer_size >= chars_read + 1);
    (*buffer)[chars_read] = '\0';

    /* Success. */
    return 1;
}

/*-------------------------------------------------------------------------*/

char * dirname(const char * path)
{
    const char * last_slash;

    assert(path);

    last_slash = strrchr(path, '/');
    if (last_slash == NULL) {

        /* The path has no directory elements (it is a simple file name). */
        return string_duplicate(".");
    }
    return string_duplicate_max(path, last_slash - path);
}

/*-------------------------------------------------------------------------*/

char * basename(const char * path)
{
    const char * last_slash;
    const char * filename_part;
    const char * extension_dot;

    assert(path);

    /* Find where the filename starts. */
    last_slash = strrchr(path, '/');
    if (last_slash) {
        filename_part = last_slash + 1;
    }
    else {
        filename_part = path;
    }

    /* Find the (first) extension dot in the filename. */
    extension_dot = strchr(filename_part, '.');

    /* Return the filename up to the first dot. */
    if (extension_dot) {
        return string_duplicate_max(
            filename_part,
            extension_dot - filename_part
        );
    }

    /* There is no extension dot, return the whole filename. */
    return string_duplicate(filename_part);
}

/*-------------------------------------------------------------------------*/
