/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#ifndef FILE_UTILS_H
#define FILE_UTILS_H

#include <stdio.h>

/*-------------------------------------------------------------------------*/

/**
 * Writes a printf-style formatted string to a newly created file.
 * If the file exists, its old contents are removed.
 * A newline is added automatically.
 */
void write_string_as_file(const char * filename, const char * text, ...);

/**
 * Appends a printf-style formatted string to the specified file,
 * preserving the old contents. A newline is added automatically.
 */
void append_string_to_file(const char * filename, const char * text, ...);

/**
 * Reads the first line from the specified file and returns it
 * as a string.
 *
 * Returns the newly allocated string on success, NULL on failure.
 */
char * read_first_line_from_file(const char * filename);

/**
 * Returns nonzero if the specified file exists, zero otherwise.
 */
int file_exists(const char * filename);

/**
 * Returns nonzero if the file exists and is a directory,
 * otherwise returns zero.
 */
int is_directory(const char * filename);

/**
 * Returns nonzero if the file exists and is a regular file,
 * otherwise returns zero.
 */
int is_regular_file(const char * filename);

/**
 * Checks whether there are readable data on the given file descriptor.
 */
int has_input_data(int fd);

/**
 * Waits for input for the specified number of seconds.
 * Returns nonzero if the input was received, zero on timeout.
 */
int wait_for_input(int fd, int timeout_sec);

/**
 * Removes the directory with all its contents.
 */
void remove_dir_recursive(const char * name);

/**
 * Creates a directory with name formatted in printf-like way.
 */
void create_dir(const char * name, ...);

/**
 * Creates a temporary directory with the specified name prefix
 * (which may contain a path to create the directory in)
 * and returns its name as a newly allocated string.
 */
char * create_temp_dir(const char * name_prefix);

/**
 * Produces an absolute filename from the specified filename
 * by prepending the current working directory
 * if the filename is not already absolute.
 * Returns a newly allocated string.
 */
char * make_filename_absolute(const char * filename);

/**
 * Opens the file for reading. Terminates the program if the opening fails.
 */
FILE * open_file_for_reading(const char * filename);

/**
 * Opens the file for writing. Terminates the program if the opening fails.
 */
FILE * open_file_for_writing(const char * filename);

/**
 * Reads a single line into the specified buffer, resizing the buffer
 * as needed and adding a terminating zero automatically.
 * If no buffer is given (buffer_size is 0), a new buffer is allocated
 * via malloc().
 * If a valid buffer is provided, it must be allocated via malloc() as well.
 * If the buffer is resized during the call, the buffer_size variable
 * is updated to reflect the new size.
 * If the end of file is reached, or if an error occurs, the buffer
 * is automatically freed, buffer_size is set to 0 and the call returns 0.
 * Otherwise, the call returns nonzero.
 */
int read_line(FILE * f, char ** buffer, int * buffer_size);

/**
 * Returns the directory part of the specified path (or "." if the path
 * does not contain any directory part) as a newly allocated string.
 */
char * dirname(const char * path);

/**
 * Returns the base name part of the specified path (without the directory
 * and file extension) as a newly allocated string.
 */
char * basename(const char * path);

/*-------------------------------------------------------------------------*/

#endif
