/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "html_output.h"
#include "string_utils.h"
#include <stdarg.h>

/*=========================================================================*/
/* Module-private functions                                                */
/*=========================================================================*/

/*
 * Prints a HTML prologue and stylesheet.
 */
static void html_print_prologue(html_file * file)
{
    assert(file);

    html_println(file, "<html>");
    html_println(file, "<head>");
    html_println(file, "<style>");
    html_println(file, "table { width: 100%; border-collapse: collapse; }");
    html_println(file, "th { border: solid 2px #ddd; font-weight: bold; }");
    html_println(file, "td { border: solid 2px #ddd; vertical-align: top; }");
    html_println(file, "</style>");
    html_println(file, "</head>");
    html_println(file, "<body>");
}

/*-------------------------------------------------------------------------*/

/*
 * Prints a HTML epilogue.
 */
static void html_print_epilogue(html_file * file)
{
    assert(file);

    html_println(file, "</body>");
    html_println(file, "</html>");
}

/*=========================================================================*/
/* Public functions                                                        */
/*=========================================================================*/

html_file * html_open(const char * file_name)
{
    FILE * f;
    html_file * file;

    /* Open the file for writing. */
    assert(file_name);
    f = fopen(file_name, "w");
    if (f == NULL) {

        /* Could not open the file. */
        return NULL;
    }

    /* Allocate memory for the structure. */
    file = (html_file *) malloc(sizeof(html_file));
    if (file == NULL) {

        /* Could not allocate memory. */
        fclose(f);
        return NULL;
    }

    /* Attach to the file. */
    file->f = f;

    /* Print the HTML prologue. */
    html_print_prologue(file);

    /* Return the HTML file structure. */
    return file;
}

/*-------------------------------------------------------------------------*/

void html_close(html_file * file)
{
    assert(file);

    /* Print the HTML epilogue. */
    html_print_epilogue(file);

    /* Close the file. */
    fclose(file->f);
    file->f = NULL;

    /* Free the file structure. */
    free(file);
}

/*-------------------------------------------------------------------------*/

void html_println(html_file * file, const char * format, ...)
{
    assert(file);
    assert(file->f);
    assert(format);

    va_list args;
    va_start(args, format);
    vfprintf(file->f, format, args);
    va_end(args);

    fputc('\n', file->f);
}

/*-------------------------------------------------------------------------*/

void html_print_head(html_file * file, unsigned int level, const char * text)
{
    assert(text);
    assert(level > 0);
    html_println(file, "<h%d>%s</h%d>", level, text, level);
}

/*-------------------------------------------------------------------------*/

void html_start_paragraph(html_file * file)
{
    html_println(file, "<p>");
}

/*-------------------------------------------------------------------------*/

void html_end_paragraph(html_file * file)
{
    html_println(file, "</p>");
}

/*-------------------------------------------------------------------------*/

void html_start_table(html_file * file)
{
    html_println(file, "<table>");
}

/*-------------------------------------------------------------------------*/

void html_end_table(html_file * file)
{
    html_println(file, "</table>");
}

/*-------------------------------------------------------------------------*/

void html_start_table_header(html_file * file)
{
    html_println(file, "<thead>");
}

/*-------------------------------------------------------------------------*/

void html_end_table_header(html_file * file)
{
    html_println(file, "</thead>");
}

/*-------------------------------------------------------------------------*/

void html_print_table_header_field(html_file * file, const char * text)
{
    html_println(file, "<th>%s</th>", text);
}

/*-------------------------------------------------------------------------*/

void html_start_table_body(html_file * file)
{
    html_println(file, "<tbody>");
}

/*-------------------------------------------------------------------------*/

void html_end_table_body(html_file * file)
{
    html_println(file, "</tbody>");
}

/*-------------------------------------------------------------------------*/

void html_start_table_line(html_file * file)
{
    html_println(file, "<tr>");
}

/*-------------------------------------------------------------------------*/

void html_end_table_line(html_file * file)
{
    html_println(file, "</tr>");
}

/*-------------------------------------------------------------------------*/

void html_next_table_line(html_file * file)
{
    html_println(file, "</tr><tr>");
}

/*-------------------------------------------------------------------------*/

void html_print_table_field(html_file * file, const char * text)
{
    html_println(file, "<td>%s</td>", text);
}

/*-------------------------------------------------------------------------*/

void html_print_multicolumn_field(html_file * file, const char * text,
    unsigned int columns)
{
    html_println(file, "<td colspan=\"%u\">%s</td>", columns, text);
}

/*-------------------------------------------------------------------------*/
