/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "inform.h"
#include "host_info.h"
#include "command_line.h"
#include <time.h>
#include <stdarg.h>
#include <assert.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

/* Callback function to be called on fail(). */
static fail_callback_t fail_callback = NULL;

/*-------------------------------------------------------------------------*/

/* Exit code */
int exit_code = 127;

/*-------------------------------------------------------------------------*/

/*
 * Prints the current time (without a newline so a message can be attached).
 */
void print_time(FILE * f)
{
    time_t numeric_time;
    struct tm * current_time;

    time(&numeric_time);
    current_time = localtime(&numeric_time);
    fprintf(
        f,
        "[%02d:%02d:%02d] ",
        current_time->tm_hour,
        current_time->tm_min,
        current_time->tm_sec
    );
}

/*-------------------------------------------------------------------------*/

void inform(const char * message, ...)
{
    va_list args;

    /* Print the current time before the message. */
    print_time(stdout);

    /* Print the formatted message. */
    assert(message);
    va_start(args, message);
    vfprintf(stdout, message, args);
    va_end(args);

    /* Add a newline. */
    putc('\n', stdout);
}

/*-------------------------------------------------------------------------*/

void warn(const char * message, ...)
{
    va_list args;

    /* Print the current time. */
    print_time(stderr);

    /* Print the formatted message. */
    fputs("warning: ", stderr);
    assert(message);
    va_start(args, message);
    vfprintf(stderr, message, args);
    va_end(args);

    /* Add a newline. */
    putc('\n', stderr);
}

/*-------------------------------------------------------------------------*/

void set_fail_callback(fail_callback_t callback)
{
    fail_callback = callback;
}

/*-------------------------------------------------------------------------*/

void fail(const char * message, ...)
{
    va_list args;
    char buffer[FAIL_MESSAGE_BUFFER_SIZE];

    /* Ensure that any previous output is printed. */
    fflush(NULL);

    /* Prepare the formatted message. */
    assert(message);
    va_start(args, message);
    vsnprintf(buffer, FAIL_MESSAGE_BUFFER_SIZE, message, args);
    va_end(args);

    /* Print the error message with printf-style formatting. */
    print_time(stderr);
    fprintf(stderr, "error: %s\n", buffer);

    /* Call the fail callback if set. */
    if (fail_callback) {
        fail_callback(buffer);
    }

    /* Exit the program immediately. */
    exit(exit_code);
}

/*-------------------------------------------------------------------------*/
