/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "lsb-autotest-probe.h"
#include "tet_journal.h"
#include "tet_journal_indexed.h"
#include "inform.h"
#include "useful_macros.h"
#include "file_utils.h"
#include <stdlib.h>
#include <assert.h>

/*-------------------------------------------------------------------------*/

#define EXIT_TEST_PASSED        0x00
#define EXIT_TEST_SKIPPED       0x01
#define EXIT_TEST_FIPPED        0x02
#define EXIT_TEST_WARNING       0x03
#define EXIT_TEST_FAILED        0x10

/*-------------------------------------------------------------------------*/

/* Name of the journal to analyze. */
const char * journal_file_name;

/* Name of the test case to query. */
const char * test_case_name;

/* Number of the test point to query. */
int test_point_number;

/* The test journal for analysis. */
TetJournal * journal;

/* The test case of interest. */
TetTestCase * test_case;

/* The test point of interest. */
TetTestPoint * test_point;

/*-------------------------------------------------------------------------*/

static void print_help(void)
{
    printf(
        "Usage: \n\n"
        "lsb-autotest-probe FILE [TESTCASE [TESTPOINT]]\n"
        "\tAnalyzes a file containing test results acquired by lsb-autotest\n"
        "\tand prints the results.\n\n"
        "\tIf a test case is specified, only results from that test case\n"
        "\tare printed. If a test point number is specified as well,\n"
        "\tonly a single result is printed.\n\n"
        "\tThe input file can be either a TET-style journal (*.journal)\n"
        "\tor a pre-parsed indexed results file (*.idx/*.res).\n"
        "\tSearching in an indexed results file is much faster so it is\n"
        "\tadvisable for larger queries. The lsb-autotest-reporter tool\n"
        "\tcan be used for obtaining such a file from a TET-style journal.\n"
        "\n"
    );
}

/*-------------------------------------------------------------------------*/

/**
 * Updates the exit code, if needed, to reflect the specified test result.
 * This is most useful when a set of tests is analyzed as a group;
 * in this case, as the exit code is only one, it reflects the most
 * serious of the results.
 * The order of decreasing seriousness is:
 * failure > warning > fip > skip > pass.
 * Returns the new exit code.
 */
static int update_exit_code_by_test_result(int exit_code, int test_result)
{
    int new_exit_code = exit_code;

    if (tet_result_is_failure(test_result)) {
        new_exit_code = EXIT_TEST_FAILED;
    }
    else if (test_result == TET_RESULT_WARNING) {
        if (
            (exit_code == EXIT_TEST_PASSED) ||
            (exit_code == EXIT_TEST_SKIPPED) ||
            (exit_code == EXIT_TEST_FIPPED)
        ) {
            new_exit_code = EXIT_TEST_WARNING;
        }
    }
    else if (test_result == TET_RESULT_FIP) {
        if (
            (exit_code == EXIT_TEST_PASSED) ||
            (exit_code == EXIT_TEST_SKIPPED)
        ) {
            new_exit_code = EXIT_TEST_FIPPED;
        }
    }
    else if (tet_result_is_skip(test_result)) {
        if (exit_code == EXIT_TEST_PASSED) {
            new_exit_code = EXIT_TEST_SKIPPED;
        }
    }

    return new_exit_code;
}

/*-------------------------------------------------------------------------*/

static int print_all_results(void)
{
    int exit_code = EXIT_TEST_PASSED;

    assert(journal);

    /* Iterate over test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* Iterate over test points in the test case. */
        FOREACH(test_point, test_case->first_test_point) {

            /* Print the result. */
            printf(
                "%s %d %s\n",
                test_case->name,
                test_point->number,
                tet_result_to_string(test_point->result)
            );

            /* Update the exit code to reflect the test result. */
            exit_code = update_exit_code_by_test_result(
                exit_code,
                test_point->result
            );
        }
    }

    return exit_code;
}

/*-------------------------------------------------------------------------*/

static int print_test_case_results(void)
{
    int exit_code = EXIT_TEST_PASSED;

    assert(journal);
    assert(test_case);

    /* Iterate over test points in the test case. */
    FOREACH(test_point, test_case->first_test_point) {

        /* Print the result. */
        printf(
            "%s %d %s\n",
            test_case->name,
            test_point->number,
            tet_result_to_string(test_point->result)
        );

        /* Update the exit code to reflect the test result. */
        exit_code = update_exit_code_by_test_result(
            exit_code,
            test_point->result
        );
    }

    return exit_code;
}

/*-------------------------------------------------------------------------*/

static int probe_index_file(int argc, char ** argv)
{
    FILE * index_file;
    FILE * results_file;
    const char * index_file_name;
    char * results_file_name;
    char * index_file_dirname;
    char * index_file_basename;
    unsigned long offset;
    unsigned long result_count;
    int exit_code = EXIT_TEST_PASSED;
    int i;
    int test_point_result;

    /* At least 2 arguments must be specified. */
    if (argc == 2) {
        fail(
            "At least the test case must be specified when "
            "using an indexed result file."
        );
    }

    /* The first argument is the name of the index file. */
    index_file_name = argv[1];
    index_file = open_file_for_reading(index_file_name);

    /* The appropriate results file has the same name except
     * that it has the '.res' extension.
     */
    index_file_dirname = dirname(index_file_name);
    index_file_basename = basename(index_file_name);
    results_file_name = format_malloc(
        "%s/%s.res",
        index_file_dirname,
        index_file_basename
    );
    free(index_file_basename);
    free(index_file_dirname);
    results_file = open_file_for_reading(results_file_name);

    /* The second argument is the test case name. */
    test_case_name = argv[2];

    /* Find the test case in the index file. */
    if (
        !idx_find_test_case(
            index_file,
            test_case_name,
            &offset,
            &result_count
        )
    ) {
        fail("Test case not found: %s", test_case_name);
    }

    /* Find the list of results in the results file. */
    if (!res_seek_to_test_point_result(results_file, offset, 1)) {
        fail("Error seeking in the results file");
    }

    /* If only the test case name is given, print all results
     * from this test case.
     */
    if (argc == 3) {

        /* Scan all results for the test case. */
        for (i = 0; i < result_count; i++) {
            test_point_result = res_read_test_point_result(results_file);
            printf(
                "%s %d %s\n",
                test_case_name,
                i + 1,              /* test point numbers start with 1 */
                tet_result_to_string(test_point_result)
            );

            /* Update the exit code to reflect the test result. */
            exit_code = update_exit_code_by_test_result(
                exit_code,
                test_point_result
            );
        }
        return exit_code;
    }

    /* The third argument is the test point number. */
    test_point_number = atoi(argv[3]);
    if (test_point_number <= 0) {
        fail("Invalid test point number");
    }

    /* Read the result from the results and index file. */
    test_point_result = tet_journal_read_result_from_indexable_file(
        results_file_name,
        index_file_name,
        test_case_name,
        test_point_number
    );
    if (test_point_result < 0) {
        fail("Could not find test point");
    }

    /* Print the result. */
    printf(
        "%s %d %s\n",
        test_case_name,
        test_point_number,
        tet_result_to_string(test_point_result)
    );

    /* Update the exit code to reflect the test result. */
    exit_code = update_exit_code_by_test_result(
        exit_code,
        test_point_result
    );

    /* Clean up. */
    free(results_file_name);

    /* Done. */
    return exit_code;
}

/*-------------------------------------------------------------------------*/

static int probe_journal_file(int argc, char ** argv)
{
    /* Load the journal (or try to). */
    journal = tet_journal_load(journal_file_name);
    if (journal == NULL) {
        fail("Could not load journal: %s", journal_file_name);
    }

    /* If this the only argument, print all results (lots of output!) */
    if (argc == 2) {
        return print_all_results();
    }

    /* The second argument is the test case name. */
    test_case_name = argv[2];

    /* Look up the test case. */
    test_case = tet_journal_find_test_case(journal, test_case_name);
    if (test_case == NULL) {
        fail("Test case not found: %s", test_case_name);
    }

    /* If only these two are specified, print all results
     * of this test case.
     */
    if (argc == 3) {
        return print_test_case_results();
    }

    /* The third argument is the test point number. */
    test_point_number = atoi(argv[3]);
    if (test_point_number <= 0) {
        fail("Invalid test point number");
    }

    /* Find the test point. */
    test_point = tet_journal_get_test_point(test_case, test_point_number);
    if (test_point == NULL) {
        fail(
            "Test point not found: %s %u (only %u - %u are available)",
            test_case_name,
            test_point_number,
            1,
            test_case->max_test_point_number
        );
    }

    /* Print the result. */
    printf(
        "%s %d: %s\n",
        test_case->name,
        test_point->number,
        tet_result_to_string(test_point->result)
    );

    /* Clean up. */
    tet_journal_delete(journal);
    return 0;
}

/*-------------------------------------------------------------------------*/

int main(int argc, char ** argv)
{
    /* At least one argument is mandatory. */
    if (argc < 2) {
        print_help();
        return 0;
    }

    /* The first argument is the journal or index file. */
    journal_file_name = argv[1];

    /* Is it an index file or a journal file? */
    if (string_ends_with(journal_file_name, ".idx")) {
        return probe_index_file(argc, argv);
    }
    else {
        return probe_journal_file(argc, argv);
    }
}

/*-------------------------------------------------------------------------*/
