/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "lsb-autotest.h"
#include "string_utils.h"
#include "subshell.h"
#include "host_info.h"
#include "inform.h"
#include "run_command.h"
#include "useful_macros.h"
#include "command_line.h"
#include "backdoor.h"
#include "file_utils.h"
#include "package.h"
#include "user_utils.h"
#include "tests.h"
#include "output_file.h"
#include "time_utils.h"
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

/*-------------------------------------------------------------------------*/

char * package_dir = NULL;
char * output_file = NULL;
char * output_dir = NULL;
int download_enabled = 1;
int keep_packages = 0;
int try_beta_packages = 0;
char * tester_email = NULL;
struct tm startup_time;
char * our_dir;
pid_t main_process_pid;
int verbose = 0;
int trust_packages = 0;

/*-------------------------------------------------------------------------*/

static void print_help()
{
    printf("\nUsage: lsb-autotest [OPTIONS] [TESTS]\n");
    printf("Runs one or more LSB tests and produces a HTML report.\n\n");
    printf("WARNING! PLEASE DO NOT USE THIS SCRIPT ON ANY MACHINE THAT\n");
    printf("HOLDS IMPORTANT DATA, OR PROVIDES IMPORTANT SERVICES!\n");
    printf("The tests sometimes put a heavy stress on the computer,\n");
    printf("and also pose a security risk as they run as root.\n\n");
    printf("Available tests:\n");
    printf("    runtime libchk cmdchk libstdc++ vsw4 xts5 desktop\n");
    printf("    expect tcl python\n\n");
    printf("If no tests are explicitly specified, all tests are run.\n\n");
    printf("Options:\n");
    printf("--beta                 ");
    printf("Use beta packages instead of stable versions.\n");
    printf("--cleanup              ");
    printf("Just uninstall test packages and exit.\n");
    printf("-D, --download         ");
    printf("Download test packages as needed. This is the default.\n");
    printf("--help                 ");
    printf("Print this help text and exit.\n");
    printf("--keep-packages        ");
    printf("Do not reinstall packages at the start.\n");
    printf("--no-download          ");
    printf("Do not download packages, fail if a package is unavailable.\n");
    printf("-m, --mailto           ");
    printf("Send results by e-mail to specified address.\n");
    printf("-o, --output           ");
    printf("Write results to the specified file.\n");
    printf("--package-dir          ");
    printf("Use the specified directory for packages.\n");
    printf("-p, --platform         ");
    printf("Assume the specified platform instead of auto-detection.\n");
    printf("\n");
}

/*-------------------------------------------------------------------------*/

static int process_argument(
    const char * arg,
    const char * next_arg,
    void * user_data
)
{
    /* Enable or disable various tests. */
    if (string_equals(arg, "all")) {

        /* Enable all tests. */
        tests_to_run |= ALL_TESTS_MASK;
        return 1;
    }
    else if (string_one_of(arg, "cmdchk", "command", NULL)) {
        tests_to_run |= CMDCHK_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "desktop")) {
        tests_to_run |= DESKTOP_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "expect")) {
        tests_to_run |= EXPECT_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "libchk")) {
        tests_to_run |= LIBCHK_TEST_BIT;
        return 1;
    }
    else if (string_one_of(arg, "libstdc++", "libstdcpp", NULL)) {
        tests_to_run |= LIBSTDCPP_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "python")) {
        tests_to_run |= PYTHON_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "runtime")) {
        tests_to_run |= RUNTIME_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "tcl")) {
        tests_to_run |= TCL_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "vsw4")) {
        tests_to_run |= VSW4_TEST_BIT;
        return 1;
    }
    else if (string_equals(arg, "xts5")) {
        tests_to_run |= XTS5_TEST_BIT;
        return 1;
    }

    /* Various options. */
    else if (string_equals(arg, "--beta")) {
        try_beta_packages = 1;
        return 1;
    }
    else if (string_one_of(arg, "-D", "--download", NULL)) {
        download_enabled = 1;
        return 1;
    }
    else if (string_equals(arg, "--cleanup")) {

        /* Just do cleanup and exit. */
        purge_packages();
        inform("Cleanup complete.");
        exit(0);
    }
    else if (string_equals(arg, "--help")) {

        /* Just print help and exit. */
        print_help();
        exit(0);
    }
    else if (string_equals(arg, "--no-download")) {
        download_enabled = 0;
        return 1;
    }
    else if (string_equals(arg, "--keep-packages")) {
        keep_packages = 1;
        return 1;
    }
    else if (string_one_of(arg, "-m", "--mailto", NULL)) {
        if (next_arg == NULL) {
            fail("Missing argument after '%s'", arg);
        }
        string_replace_with_copy(&tester_email, next_arg);
        return 2;
    }
    else if (string_starts_with(arg, "--mailto=")) {
        string_replace_with_copy(&tester_email, arg + 9);
        return 1;
    }
    else if (string_one_of(arg, "-o", "--output", NULL)) {
        if (next_arg == NULL) {
            fail("Missing argument after '%s'", arg);
        }
        string_replace(&output_file, format_malloc("%s.tar.gz", next_arg));
        return 2;
    }
    else if (string_starts_with(arg, "--output=")) {
        string_replace(&output_file, format_malloc("%s.tar.gz", arg + 9));
        return 1;
    }
    else if (string_starts_with(arg, "--package_dir")) {
        if (next_arg == NULL) {
            fail("Missing argument after '%s'", arg);
        }
        string_replace_with_copy(&package_dir, next_arg);
        return 2;
    }
    else if (string_starts_with(arg, "--package_dir=")) {
        string_replace_with_copy(&package_dir, arg + 14);
        return 1;
    }
    else if (string_one_of(arg, "-p", "--platform", NULL)) {
        if (next_arg == NULL) {
            fail("Missing argument after '%s'", arg);
        }
        string_replace_with_copy(&host_platform, next_arg);
        string_replace(&host_machine_triplet, format_malloc("%s-linux-gnu", host_platform));
        return 2;
    }
    else if (string_starts_with(arg, "--platform=")) {
        string_replace_with_copy(&host_platform, arg + 11);
        string_replace(&host_machine_triplet, format_malloc("%s-linux-gnu", host_platform));
        return 2;
    }
    else if (string_equals(arg, "-v")) {
        verbose = 1;
        return 1;
    }
    else if (string_equals(arg, "--trust-packages")) {
        trust_packages = 1;
        return 1;
    }

    fail("Unrecognized command line option '%s' (try '--help')", arg);
    return 0;   /* not reached */
}

/*-------------------------------------------------------------------------*/

static void finish_arguments()
{
    char * filename_time;

    /* If no tests were explicitly selected, we will run everything. */
    if (tests_to_run == 0) {
        tests_to_run = ALL_TESTS_MASK;
    }

    /*
     * FIXME: The xts5 test is still beta, so disable it unless the '--beta'
     * option is used.
     */
    if (!try_beta_packages) {
        if (tests_to_run & XTS5_TEST_BIT) {
            tests_to_run &= ~XTS5_TEST_BIT;
            inform(
                "Warning: The XTS5 test is still beta, disabling it for now."
                " (Use '--beta' to try it anyway.)"
            );
        }
    }

    /* If the output file name was not specified, provide a default. */
    if (output_file == NULL) {

        /*
         * Build the output file name from the startup time, host name,
         * and host platform.
         */
        assert(host_platform);
        assert(hostname);
        filename_time = format_time_for_filename(&startup_time);
        output_file = format_malloc(
            "/tmp/lsb-autotest-%s-%s-%s.tar.gz",
            host_platform,
            hostname,
            filename_time
        );
        free(filename_time);
    }

    /* If the output file name is explicitly specified, ensure that it is
     * an absolute file name.
     */
    else {
        string_replace(&output_file, make_filename_absolute(output_file));
    }

    /* If the package directory was not specified, provide a default. */
    if (package_dir == NULL) {
        package_dir = format_malloc("/tmp/lsb-packages");
    }

    /* If the temporary output directory was not specified,
     * provide a default.
     */
    if (output_dir == NULL) {
        output_dir = format_malloc("/tmp/lsb-autotest-result");
    }

    /* Show which tests were selected. */
    print_selected_tests();

    /* Print the options. */
    inform("Output file is %s", output_file);
    if (tester_email) {
        inform("Results will be sent by e-mail to %s", tester_email);
    }
    inform("Packages will be searched in %s", package_dir);
}

/*-------------------------------------------------------------------------*/

/* Checks whether we are run as root, calls fail() otherwise. */
static void check_for_root()
{
    if (geteuid() != 0) {
        fail("This program must be run as root.");
    }
}

/*-------------------------------------------------------------------------*/

/* Checks whether the local 'passwd' utility knows about the '--stdin'
 * option; calls fail() if it does not.
 */
static void check_for_smart_passwd()
{
    if (try_command("passwd --help | grep -- '--stdin' >/dev/null") != 0) {
        fail(
            "It seems that the passwd utility does not support "
            "the --stdin option; please upgrade."
        );
    }
}

/*-------------------------------------------------------------------------*/

/* Checks whether basic Unix tools are available; calls fail()
 * if some are missing.
 */
static void check_for_basic_tools()
{
    int i;
    static const char * required_tools[] = {
        "chmod", "cp", "grep", "ls", "lsb_release", "mv", "passwd", "rm",
        "rpm", "tail", "tar", "useradd", "userdel", "wget", NULL
    };

    for (i = 0; required_tools[i]; i++) {
        check_for_tool(required_tools[i]);
    }
}

/*-------------------------------------------------------------------------*/

/*
 * Writes startup information to the output directory.
 */
static void write_startup_info()
{
    char * output_file_name;

    output_file_name = format_malloc("%s/host_info/host_machine", output_dir);
    write_string_as_file(
        output_file_name,
        "%s",
        host_platform
    );
    free(output_file_name);

    output_file_name = format_malloc("%s/host_info/host_name", output_dir);
    write_string_as_file(
        output_file_name,
        "%s",
        hostname
    );
    free(output_file_name);

    output_file_name = format_malloc("%s/host_info/host_platform", output_dir);
    write_string_as_file(
        output_file_name,
        "%s",
        host_platform
    );
    free(output_file_name);

    output_file_name = format_malloc("%s/host_info/host_os", output_dir);
    write_string_as_file(
        output_file_name,
        "%s %s",
        host_distro_name,
        host_distro_release
    );
    free(output_file_name);

    output_file_name = format_malloc("%s/host_info/host_kernel", output_dir);
    write_string_as_file(
        output_file_name,
        "%s %s %s",
        host_os_name,
        host_os_release,
        host_os_version
    );
    free(output_file_name);

    output_file_name = format_malloc("%s/autotest_info/start_time", output_dir);
    write_string_as_file(
        output_file_name,
        "%02d:%02d:%02d %02d-%02d-%04d",
        startup_time.tm_hour,
        startup_time.tm_min,
        startup_time.tm_sec,
        startup_time.tm_mday,
        startup_time.tm_mon + 1,
        startup_time.tm_year + 1900
    );
    free(output_file_name);
}

/*-------------------------------------------------------------------------*/

/*
 * Finds the directory where the autotest executable resides;
 * this is the directory where helpers and data files should be.
 */
static void find_our_directory()
{
    char buffer[COMMAND_OUTPUT_BUFFER_SIZE];
    int path_length;
    char * last_slash;

    /* Get the path to the running executable by reading /proc/self/exe. */
    path_length = readlink(
        "/proc/self/exe",
        buffer,
        COMMAND_OUTPUT_BUFFER_SIZE - 1
    );
    if (path_length < 0) {
        fail("Could not read /proc/self/exe: %m");
    }

    /* Put the terminating zero to the buffer. */
    buffer[path_length] = '\0';

    /* Remove the portion after the last slash to obtain the directory. */
    last_slash = strrchr(buffer, '/');
    if (last_slash == NULL) {
        fail("Ill-formed executable path (no '/'): %s", buffer);
    }
    *last_slash = '\0';

    inform("Installed in directory: %s", buffer);

    /* Store the directory. */
    our_dir = string_duplicate(buffer);
}

/*-------------------------------------------------------------------------*/

/*
 * Handler for the SIGINT (Ctrl+C) signal. It ensures everything is cleaned up
 * and terminates the program.
 */
static void sigint_handler(int signo)
{
    exit_code = 111;
    fail("SIGINT signal received, terminating.");
}

/*-------------------------------------------------------------------------*/

/*
 * Handler for the SIGSEGV signal.
 */
static void sigsegv_handler(int signo)
{
    exit_code = 111;
    fail("SIGSEGV signal received, terminating.");
}

/*-------------------------------------------------------------------------*/

/*
 * Handler for the SIGALRM signal. It does nothing at all, it is here
 * just to allow us to use alarm() for stopping long reads.
 */
static void sigalrm_handler(int signo)
{
}

/*-------------------------------------------------------------------------*/

/*
 * Installs signal handlers for SIGALRM, SIGINT and SIGSEGV.
 */
static void install_signal_handlers()
{
    struct sigaction act;

    act.sa_handler = sigalrm_handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, NULL);

    act.sa_handler = sigint_handler;
    sigaction(SIGINT, &act, NULL);

    act.sa_handler = sigsegv_handler;
    sigaction(SIGSEGV, &act, NULL);
}

/*-------------------------------------------------------------------------*/

static void complain_by_mail(const char * message)
{
    assert(hostname);
    try_command(
        "echo 'The autotest run has failed with the following error: "
        "%s' | mail -s 'LSB autotest failed at %s' '%s'",
        message,
        tester_email,
        hostname
    );
}

/*-------------------------------------------------------------------------*/

static void cleanup()
{
    /* All cleaning is done by the main process only; forked helper processes
     * can simply quit right now.
     */
    if (getpid() != main_process_pid) {
        return;
    }

    /* Remove the backdoor if it is installed. */
    remove_backdoor();

    /*
     * NOTE: The temporary output directory is NOT removed.
     * This is because cleanup() is also called when the autotest run fails,
     * and when this happens, we want the output directory to be preserved
     * in order to see what happened.
     */
}

/*-------------------------------------------------------------------------*/

int main(int argc, char ** argv)
{
    /* Print a short header. */
    inform("This is lsb-autotest 1.0-WIP, built %s %s.", __DATE__, __TIME__);
    inform("Use 'lsb-autotest --help' for help.");

    /* Set up signal handlers to handle crashes and interrupts
     * as safely as possible.
     */
    install_signal_handlers();

    /* Remember the startup time. */
    get_current_time(&startup_time);

    /* Remember the PID of this process (it will do all the cleaning
     * at the end).
     */
    main_process_pid = getpid();

    /* Determine initial information about the program and system. */
    find_our_directory();
    read_host_info();
    guess_package_manager();
    print_host_info();

    /* Process the command line. */
    process_command_line(
        argc,
        argv,
        process_argument,
        finish_arguments,
        NULL
    );

    /* If the tester has provided his/her e-mail address, ensure
     * that he/she will be notified if an error occurs.
     */
    if (tester_email) {
        set_fail_callback(complain_by_mail);
    }

    /* Do the initial checks. */
    check_for_root();
    check_for_basic_tools();
    // check_for_smart_passwd();

    /* Build URLs where the test packages for the host platform
     * can be found.
     */
    build_package_urls();

    /* Create an output directory and write the information gathered
     * so far.
     */
    create_output_dir();
    write_startup_info();

    /* Switch to the output directory. */
    EXPECT_ZERO(chdir(output_dir), "Could not switch to output directory");

    /* Ensure that cleanup() will be called on termination, even if
     * the program fails or crashes (this is very important as we are
     * changing system configuration files so they must be restored
     * to proper state).
     */
    atexit(cleanup);

    /* Remove all previously installed test packages to ensure a clean start
     * (the "--keep-packages" option suppresses this behavior).
     */
    if (!keep_packages) {
        purge_packages();
    }

    /*
     * LSB 3.1 / 16.4. Path For System Administration Utilities
     * Certain utilities used for system administration (and other privileged
     * commands) may be stored in /sbin, /usr/sbin, and /usr/local/sbin.
     * Applications requiring to use commands identified as system
     * administration utilities should add these directories to their PATH.
     */
    char * env_path;
    env_path = getenv ("PATH");
    if (env_path!=NULL) {
        env_path = format_malloc("PATH=%s%s",env_path,":/sbin:/usr/sbin:/usr/local/sbin");
        putenv(env_path);
    } else {
        putenv("PATH=/sbin:/usr/sbin:/usr/local/sbin");
    }

    /* Run tests that were selected on the command line. */
    run_selected_tests();

    /* Create the output archive file. */
    create_output_file();

    /* Done. */
    cleanup();
    return 0;
}

/*-------------------------------------------------------------------------*/
