/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#define _XOPEN_SOURCE 600
#define _POSIX_C_SOURCE 200112L

#include "package.h"
#include "lsb-autotest.h"
#include "string_utils.h"
#include "run_command.h"
#include "inform.h"
#include "command_line.h"
#include "host_info.h"
#include "file_utils.h"
#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

/*-------------------------------------------------------------------------*/
int package_manager = 0;

const char * wget = "wget --timestamping -r -nd -np -w 3 --retr-symlinks --timeout=120";

const char * download_url = "ftp://ftp.freestandards.org/pub/lsb";

char * runtest_url = NULL;
char * appbat_url = NULL;
char * testdata_url = NULL;
char * beta_runtest_url = NULL;
char * beta_appbat_url = NULL;

char * package_platform = NULL;
char * appbat_package_platform = NULL;
char * package_platform_dir = NULL;

/*-------------------------------------------------------------------------*/

void build_package_urls()
{
    /*
     * Test packages are sorted by platform on the download site,
     * but unfortunately the platform naming is a bit inconsistent.
     */
    assert(host_platform);
    if (string_equals(host_platform, "i686")) {
        package_platform_dir = string_duplicate("ia32");
        package_platform = string_duplicate("i486");
        appbat_package_platform = string_duplicate("i486");
    }
    else if (string_equals(host_platform, "x86_64")) {
        package_platform_dir = string_duplicate("amd64");
        package_platform = string_duplicate("x86_64");
        appbat_package_platform = string_duplicate("x86_64");
    }
    else if (string_equals(host_platform, "ppc")) {
        package_platform_dir = string_duplicate("ppc32");
        package_platform = string_duplicate("ppc");
        appbat_package_platform = string_duplicate("ppc");
    }
    else {
        package_platform_dir = string_duplicate(host_platform);
        package_platform = string_duplicate(host_platform);
        appbat_package_platform = string_duplicate(host_platform);
    }

    /* Build the complete URLs. */
    assert(download_url);
    runtest_url = format_malloc(
        "%s/test_suites/released/binary/runtime",
        download_url
    );
    beta_runtest_url = format_malloc(
        "%s/test_suites/beta/binary/runtime",
        download_url
    );
    appbat_url = format_malloc(
        "%s/app-battery/released-3.0.0/%s",
        download_url,
        package_platform_dir
    );
    beta_appbat_url = format_malloc(
        "%s/app-battery/beta/%s",
        download_url,
        package_platform_dir
    );
    testdata_url = format_malloc(
        "%s/app-battery/tests",
        download_url
    );

    /* Inform about the URLs being used. */
    inform("Package source URLs:");
    inform("Runtime: %s", runtest_url);
    inform("Beta runtime: %s", beta_runtest_url);
    inform("Appbat: %s", appbat_url);
    inform("Beta appbat: %s", beta_appbat_url);
}

/*-------------------------------------------------------------------------*/

char * find_package_file(const char * name)
{
    assert(name);

    inform("Looking for local RPM file for package: %s", name);

    /* Look for the file with appropriate name. If there are
     * multiple versions, pick the newest one.
     */
    assert(package_dir);
    assert(package_platform);
    assert(appbat_package_platform);
    char * result = string_duplicate(
        capture(
            "ls -1 -v %s/%s*.%s.rpm %s/%s*.%s.rpm "
            "2>/dev/null | grep %s-[0-9] | tail -n1 2>/dev/null",
            package_dir,
            name,
            package_platform,
            package_dir,
            name,
            appbat_package_platform,
            name
        )
    );

    /* If the result is empty, discard it right now and return plain NULL. */
    if (string_empty(result)) {
        free(result);
        return NULL;
    }

    /* Otherwise return the file name. */
    return result;
}

/*-------------------------------------------------------------------------*/

void install_rpm_file(const char * rpm_filename)
{
    /* Check that 'rpm_filename' file exists */
    if ( !file_exists(rpm_filename) )
        fail ("File doesn't exist: %s", rpm_filename);
    
    if ( package_manager == PACKAGE_MANAGER_RPM )
    {
        run_command("rpm -i %s", rpm_filename);
    }
    else
    if ( package_manager == PACKAGE_MANAGER_DPKG )
    {
        run_command("alien -c -i -k %s", rpm_filename);
    }
    else
    {
        /* Can't install packages when don't know package manager */
        fail("Could not identify a package manager");
    }
}

/*-------------------------------------------------------------------------*/

void install_package(const char * package_name, const char * package_repository)
{
    char * package_file = NULL;
    const char * beta_url = NULL;
    const char * url = NULL;
    const char * platform = NULL;

    assert(package_name);
    assert(package_repository);

    /* Look if the package already exists in the local package directory. */
    package_file = find_package_file(package_name);
    if (package_file == NULL) {

        /* The package is not here, we need to download it. */

        /* Fail right now if downloading is disabled. */
        if (!download_enabled) {
            fail(
                "RPM file for package '%s' not found locally "
                "(try '--download' for downloading it)",
                package_name
            );
        }

        /* Create the package directory if it does not exist yet. */
        if (mkdir(package_dir, 0755) != 0) {

            /* EEXIST (already exists) is not an error, anything other is. */
            if (errno != EEXIST) {
                fail(
                    "Could not create package directory: %s: %m",
                    package_dir
                );
            }
        }

        /* Switch to the package directory. */
        if (chdir(package_dir) != 0) {
            fail("Could not switch to package directory: %s: %m", package_dir);
        }

        /* Select the URL for downloading according to the repository
         * the package belongs to.
         */
        if (string_equals(package_repository, "appbat")) {
            url = appbat_url;
            beta_url = beta_appbat_url;
            platform = appbat_package_platform;
        }
        else if (string_equals(package_repository, "runtime")) {
            url = runtest_url;
            beta_url = beta_runtest_url;
            platform = package_platform;
        }
        else {
            fail("Unrecognized package repository: %s", package_repository);
        }

        inform("Downloading: %s (%s)", package_name, package_repository);

        /* If the '--beta' option is used, try to download
         * the beta version first.
         */
        if (try_beta_packages) {
            run_command(
                "%s \"%s/%s*.%s.rpm\"",
                wget,
                beta_url,
                package_name,
                platform
            );
        }

        /* If the '--beta' option is NOT used, or if we were not able
         * to get the beta package, get the stable version.
         */
        if (!try_beta_packages || (!find_package_file(package_name))) {
            run_command(
                "%s \"%s/%s*.%s.rpm\"",
                wget,
                url,
                package_name,
                platform
            );
        }

        /* Look if we have downloaded something usable. */
        package_file = find_package_file(package_name);
        if (package_file == NULL) {
            fail("Could not download package: %s", package_name);
        }
    }

    /* Install it. */
    inform("Installing RPM: %s", package_file);
    install_rpm_file(package_file);

    /* Check whether the package was installed successfully. (The exit value
     * of the 'rpm -i' command is not always reliable.)
     */
    if (!is_package_installed(package_name)) {
        fail("Package %s was not correctly installed", package_name);
    }

    /* Clean up. */
    free(package_file);
}

/*-------------------------------------------------------------------------*/

int is_package_installed(const char * package_name)
{
    if ( package_manager == PACKAGE_MANAGER_RPM )
    {
        return !try_command("rpm -q %s >/dev/null 2>&1", package_name);
    }
    else
    if ( package_manager == PACKAGE_MANAGER_DPKG )
    {
        char * dpkg_capture = string_duplicate( 
            capture("dpkg -s %s  2>/dev/null "
                    "| grep '^\\(Version\\|Status\\):'", package_name )
            );
        
        if ( string_empty(dpkg_capture) ) 
        {
            /*
             * Package in qmtest_libstdcpp-*.rpm can be named 
             * like qmtest-libstdcpp. So, replace all '_' with '-'.
             */
            char * tmp = string_duplicate(package_name);
            string_tr(tmp, '_', '-');
                            
            /* Try again with the changed name */
            dpkg_capture = string_duplicate( 
                capture("dpkg -s %s  2>/dev/null "
                        "| grep '^\\(Version\\|Status\\):'", tmp)
                );
        }
        
        if ( strstr(dpkg_capture, "install ok") != NULL ) 
        {
            return 1;
        }
        
        return 0;
    }
    else
    {
        /* Can't check package when don't know package manager */
        fail("Could not identify a package manager");
    }

    return 0;   /* never happens */
}

/*-------------------------------------------------------------------------*/

void check_package(const char * package_name, const char * package_repository)
{
    if (trust_packages) return;
    
    inform("Looking for package: %s", package_name);
    if (!is_package_installed(package_name)) {
        install_package(package_name, package_repository);
    }
}

/*-------------------------------------------------------------------------*/

void remove_package(const char * package_name)
{
    if (is_package_installed(package_name)) {
        inform("Uninstalling package: %s", package_name);

        if ( package_manager == PACKAGE_MANAGER_RPM )
        {
            run_command("rpm -e %s", package_name);
        }
        else
        if ( package_manager == PACKAGE_MANAGER_DPKG )
        {
            run_command("dpkg -P %s", package_name); 
            
            if( is_package_installed(package_name) ) 
            {
                /*
                 * Package in qmtest_libstdcpp-*.rpm can be named 
                 * like qmtest-libstdcpp. So, replace all '_' with '-'.
                 */
                char * tmp = string_duplicate(package_name);
                string_tr(tmp, '_', '-');
                
                /* Try again with the changed name */
                run_command("dpkg -P %s", tmp); 
            }
        }
        else
        {
            /* Can't install packages when don't know package manager */
            fail("Could not identify a package manager");
        }
    }
}

/*-------------------------------------------------------------------------*/

void purge_packages()
{
    inform("Removing previously installed test packages");

    /*
     * Remove the packages one by one. Order is sometimes important
     * as there are dependencies.
     */
    remove_package("lsb-apache");
    remove_package("lsb-celestia");
    remove_package("lsb-expect");
    remove_package("lsb-groff");
    remove_package("lsb-lynx");
    remove_package("lsb-rsync");
    remove_package("lsb-samba");
    remove_package("lsb-tcl");
    remove_package("lsb-xpdf");
    remove_package("lsb-xpaint");
    remove_package("qmtest_libstdcpp");
    remove_package("qm");
    remove_package("lsb-python");
    remove_package("lsb-cmdchk");
    remove_package("lsb-libchk");
    remove_package("lsb-runtime-test");
    remove_package("lsb-test-vsw4");
    remove_package("lsb-test-xts5");
    remove_package("lsb-test-desktop");
    remove_package("lsb-tet3-lite");
}

/*-------------------------------------------------------------------------*/

void check_data_file(const char * file_name)
{
    char * file_path;

    file_path = format_malloc("%s/%s", package_dir, file_name);
    if ( file_exists(file_path) ) {

        /* The file is already there, so finish. */
        free(file_path);
        return;
    }
    free(file_path);

    /* Download the file from the web. */
    run_command(
        "%s \"%s/%s\"",
        wget,
        testdata_url,
        file_name
    );
}

/*-------------------------------------------------------------------------*/

void guess_package_manager() 
{
    /* Native package manager for Suse and Redhat is RPM */
    if (   strcasestr(host_distro_name, "suse")
        || strcasestr(host_distro_name, "red hat")
        || strcasestr(host_distro_name, "redhat")
        || strcasestr(host_distro_name, "asianux")
        || strcasestr(host_distro_name, "mandriva")
        )
    {
        package_manager = PACKAGE_MANAGER_RPM;
    }
    else 
    /* For Debian and Ubuntu use 'dpkg' and 'alien' */
    if (   strcasestr(host_distro_name, "debian")
        || strcasestr(host_distro_name, "ubuntu")
        )
    {
        package_manager = PACKAGE_MANAGER_DPKG;
    }
    else 
    {
        /* If 'dpkg' and 'alien' utils are available then use them */
        if (   (try_command("which dpkg >/dev/null 2>&1") == 0) 
            && (try_command("which alien >/dev/null 2>&1") == 0)
            )
        {
            package_manager = PACKAGE_MANAGER_DPKG;
        }
    }
    
    if ( !trust_packages ) {
        /* Can't install packages when don't know package manager */
        if ( package_manager == PACKAGE_MANAGER_UNKNOWN ) {
            if ( try_command("which dpkg >/dev/null 2>&1") == 0 ) {
                package_manager = PACKAGE_MANAGER_DPKG;
            } else {
                package_manager = PACKAGE_MANAGER_RPM;
            }
        }
    }
}

/*-------------------------------------------------------------------------*/