/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "command_line.h"
#include "file_utils.h"
#include "inform.h"
#include "reporter.h"
#include "run_command.h"
#include "string_utils.h"
#include "string_array.h"
#include "tet_journal.h"
#include "tet_journal_compat.h"
#include "tet_journal_html.h"
#include "tet_journal_indexed.h"
#include "html_output.h"
#include <assert.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

/* File to write test summaries to. */
static html_file * summary_file = NULL;

/*-------------------------------------------------------------------------*/

/* List of journal files to be processed. */
static StringArray * journals;

/*-------------------------------------------------------------------------*/

/**
 * Processes an archive of journals.
 */
static void process_journal_archive(const char * archive_file_name)
{
    TetJournal * journal = NULL;
    sstring * archive_base_name = sstring_create();
    sstring * current_file = sstring_create();
    sstring * journal_summary = sstring_create();
    sstring * journal_detail_file_base = sstring_create();
    sstring * journal_html_file_name = sstring_create();
    sstring * field = sstring_create();
    sstring * command = sstring_create();
    char * temp_dir = NULL;
    char * host_os = NULL;
    char * host_platform = NULL;
    char * host_kernel = NULL;
    char * test_time = NULL;
    sstring * tar_params = sstring_create();

    /* Get the base name of the archive, used for naming result files. */
    sstring_set(archive_base_name, archive_file_name);
    sstring_filename_to_basename(archive_base_name);

    /* Choose a decompression method. */
    if (string_ends_with(archive_file_name, ".tar.bz2")) {
        sstring_set(tar_params, "-jxvf");
    }
    else if (
        string_ends_with(archive_file_name, ".tar.gz") ||
        string_ends_with(archive_file_name, ".tgz")
    ) {
        sstring_set(tar_params, "-zxvf");
    }
    else {
        sstring_set(tar_params, "-xvf");    /* Assume a plain tar file. */
    }

    /* Unpack the archive to a temporary directory. */
    temp_dir = create_temp_dir("/var/tmp/lsb-autotest-reporter-");
    sstring_printf(
        command,
        "tar %s %s -C %s >/dev/null",
        CSTRING(tar_params),
        archive_file_name,
        temp_dir
    );
    inform("Processing archive: %s", archive_file_name);
    if (try_command(CSTRING(command)) != 0) {
        warn("Error unpacking archive, skipping.");
        goto finish_archive;
    }

    /* Read the host OS info. */
    sstring_printf(current_file, "%s/host_info/host_os", temp_dir);
    host_os = read_first_line_from_file(CSTRING(current_file));
    if (host_os == NULL) {

        /* No host info, the archive is probably not in our format. */
        warn("Unrecognized archive format, skipping.");
        goto finish_archive;
    }

    /* Read the host platform. */
    sstring_printf(current_file, "%s/host_info/host_platform", temp_dir);
    host_platform = read_first_line_from_file(CSTRING(current_file));

    /* Read the host kernel version. */
    sstring_printf(current_file, "%s/host_info/host_kernel");
    host_kernel = read_first_line_from_file(CSTRING(current_file));

    /* Read the date of the test run. */
    sstring_printf(current_file, "%s/autotest_info/start_time");
    test_time = read_first_line_from_file(CSTRING(current_file));

    /* Print the name of the archive. */
    html_start_table_line(summary_file);
    sstring_printf(field, "%s, %s, %s<br>"
        "<span style=\"font-size: xx-small;\">%s</span></td>\n",
        host_os, host_platform, test_time, CSTRING(archive_base_name));
    html_print_table_field(summary_file, CSTRING(field));

    /* Process journals stored in the archive; for each journal,
     * a field in the summary table is generated, and three files:
     * a detailed HTML report, and two indexed journal files
     * (.res and .idx).
     */

#define HANDLE_JOURNAL(name) { \
    sstring_printf(current_file, "%s/test_journals/" name, \
        temp_dir); \
    journal = load_any_journal(CSTRING(current_file)); \
    if (journal) { \
        \
        /* Build a file name for the result file. */ \
        sstring_printf(journal_detail_file_base, "%s-%s", \
            CSTRING(archive_base_name), name); \
        sstring_filename_to_basename(journal_detail_file_base); \
        sstring_filename_to_identifier(journal_detail_file_base); \
        \
        /* Write a detailed HTML report for this journal. */ \
        sstring_printf(journal_html_file_name, "%s.html", \
            CSTRING(journal_detail_file_base)); \
        tet_journal_to_html_file(CSTRING(journal_html_file_name), journal); \
        \
        /* Write the journal result to the summary table. */ \
        tet_journal_summarize_html(journal, journal_summary); \
        sstring_printf(field, "%s (<a href=\"%s\">see</a>)\n", \
            CSTRING(journal_summary), \
            CSTRING(journal_html_file_name)); \
        html_print_table_field(summary_file, CSTRING(field)); \
        tet_journal_delete(journal); \
    } \
    else { \
        html_print_table_field(summary_file, "-"); \
    } \
}

    HANDLE_JOURNAL("command_check.journal");
    HANDLE_JOURNAL("library_check.journal");
    HANDLE_JOURNAL("runtime_test.journal");
    HANDLE_JOURNAL("libstdcpp_test.journal");
    HANDLE_JOURNAL("x11_test.journal");
    HANDLE_JOURNAL("gtkvts_test.journal");
    HANDLE_JOURNAL("qt3_test.journal");
    HANDLE_JOURNAL("libxml_test.journal");
    HANDLE_JOURNAL("libpng_test.journal");
    HANDLE_JOURNAL("fontconfig_test.journal");
    HANDLE_JOURNAL("python.pythonjournal");
    HANDLE_JOURNAL("tcl.tcljournal");
    HANDLE_JOURNAL("expect.tcljournal");

#undef HANDLE_JOURNAL

    html_end_table_line(summary_file);

finish_archive:

    /* Remove the temporary directory. */
    remove_dir_recursive(temp_dir);
    free(temp_dir);

    /* Clean up. */
    sstring_delete(archive_base_name);
    sstring_delete(current_file);
    sstring_delete(journal_summary);
    sstring_delete(journal_detail_file_base);
    sstring_delete(journal_html_file_name);
    sstring_delete(field);
    sstring_delete(command);
    sstring_delete(tar_params);
}

/*-------------------------------------------------------------------------*/

static void process_archives_recursively(const char * dir_name)
{
    DIR * dir;
    struct dirent * dir_entry;
    const char * file_name;
    sstring * file_path = sstring_create();

    inform("Scanning directory: %s", dir_name);

    /* Open the directory. */
    dir = opendir(dir_name);
    if (dir == NULL) {
        fail("Could not open directory: %s", dir_name);
    }

    /* Scan all files in the directory, and process every regular file
     * with name ending with 'journal'.
     */
    for (;;) {
        dir_entry = readdir(dir);
        if (dir_entry == NULL) {

            /* End of the list. */
            break;
        }
        file_name = dir_entry->d_name;

        /* Skip the stupid '.' and '..' entries. */
        if (string_equals(file_name, ".") || string_equals(file_name, "..")) {
            continue;
        }

        /* Look at every file to see what it is. */
        sstring_printf(file_path, "%s/%s", dir_name, file_name);
        if (is_directory(CSTRING(file_path))) {

            /* Recurse into subdirectory. */
            process_archives_recursively(CSTRING(file_path));
        }
        else if (is_regular_file(CSTRING(file_path))) {
            if ((string_ends_with(file_name, ".tar.bz2")) ||
                (string_ends_with(file_name, ".tar.gz"))) {

                /* Unpack the archive and process it. */
                process_journal_archive(CSTRING(file_path));
            }
        }
    }

    /* Clean up. */
    closedir(dir);
    sstring_delete(file_path);
}

/*-------------------------------------------------------------------------*/

static void print_help(void)
{
    printf(
        "Usage: \n"
        "\nlsb-autotest-reporter [OPTIONS] FILE ...\n"
        "\tProcesses one or more archives from lsb-autotest testing\n"
        "\tand produces a HTML summary.\n"
        "\tIf a directory is specified, it is recursively searched\n"
        "\tfor journal archives.\n"
        "\nOptions:\n"
        "\t--help\n"
        "\t\tPrints this help text.\n"
        "\t--prefix=...\n"
        "\t\tAssume that HTML journals will be placed in the specified\n"
        "\t\tdirectory and make links from the main HTML page\n"
        "\t\tappropriately. Useful when the summaries are on a single\n"
        "\t\tpage while journals are in their own directory.\n"
        "\n"
    );
}

/*-------------------------------------------------------------------------*/

/*
 * Processes a command-line argument.
 */
static int process_argument(
    const char * arg,
    const char * next_arg,
    void * user_data
)
{
    if (string_equals(arg, "--help")) {
        print_help();
        exit(0);
    }

    /* Anything other is considered to be a journal file name. */
    else {
        string_array_append(journals, arg);
        return 1;
    }
}

/*-------------------------------------------------------------------------*/

int main(int argc, char ** argv)
{
    int i;
    const char * file_name;

    /* If no command line arguments are given, print help and exit. */
    if (argc == 1) {
        print_help();
        return 1;
    }

    journals = string_array_create();

    /* Load the command line. */
    process_command_line(argc, argv, process_argument, NULL, NULL);

    /* Handle incomplete arguments. */
    if (journals->length == 0) {
        fail("No input journals were specified");
    }

    /* Open the HTML summary file. */
    if (summary_file == NULL) {
        summary_file = html_open("test_summary.html");
    }

    /* Print the header of the summary result table. */
    html_start_table(summary_file);
    html_start_table_header(summary_file);
    html_print_table_header_field(summary_file, "Test Run Info");
    html_print_table_header_field(summary_file, "Commands");
    html_print_table_header_field(summary_file, "Libraries");
    html_print_table_header_field(summary_file, "Runtime");
    html_print_table_header_field(summary_file, "Libstdc++");
    html_print_table_header_field(summary_file, "X11");
    html_print_table_header_field(summary_file, "Gtk");
    html_print_table_header_field(summary_file, "Qt3");
    html_print_table_header_field(summary_file, "Libxml");
    html_print_table_header_field(summary_file, "Libpng");
    html_print_table_header_field(summary_file, "Fontconfig");
    html_print_table_header_field(summary_file, "Python");
    html_print_table_header_field(summary_file, "Tcl");
    html_print_table_header_field(summary_file, "Expect");
    html_end_table_header(summary_file);
    html_start_table_body(summary_file);

    /* Process specified journal files. */
    for (i = 0; i < journals->length; i++) {
        file_name = string_array_get(journals, i);
        process_archives_recursively(file_name);
    }

    /* Finish the summary table. */
    html_end_table_body(summary_file);
    html_end_table(summary_file);

    /* Finish the summary file. */
    html_close(summary_file);

    /* Clean up. */
    string_array_delete(journals);

    /* Done. */
    return 0;
}

/*-------------------------------------------------------------------------*/
