/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "run_command.h"
#include "inform.h"
#include "string_utils.h"
#include "useful_macros.h"
#include "user_utils.h"
#include <assert.h>
#include <ctype.h>
#include <grp.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

/*-------------------------------------------------------------------------*/

char command_output_buffer[COMMAND_OUTPUT_BUFFER_SIZE];

/*-------------------------------------------------------------------------*/

/*
 * Runs the given command in the shell and returns its exit code.
 * Terminates the program if the shell cannot be invoked, or if the command
 * crashes.
 */
static int shell(const char * command)
{
    int return_value;

    /* Run the shell command. */
    return_value = system(command);

    /* If the exit code is less than 0, the system() call has failed. */
    if (return_value < 0) {
        fail("Could not start shell command: %s: %m", command);
    }

    /* Check for abnormal terminations. */
    if (WIFSIGNALED(return_value)) {
        fail(
            "Shell command terminated by signal %d: %s",
            WTERMSIG(return_value),
            command
        );
    }
    if (!WIFEXITED(return_value)) {
        fail(
            "Shell command terminated abnormally: %s",
            command
        );
    }

    /* Return the exit code from the command. */
    return WEXITSTATUS(return_value);
}

/*-------------------------------------------------------------------------*/

void run_command(const char * command, ...)
{
    int result, exit_code;
    char command_buffer[COMMAND_BUFFER_SIZE];
    va_list args;

    assert(command);

    /* Format the command. */
    va_start(args, command);
    result = vsnprintf(command_buffer, COMMAND_BUFFER_SIZE, command, args);
    va_end(args);
    if ((result < 0) || (result >= COMMAND_BUFFER_SIZE)) {
        fail("Command string too long");
    }

    /* Run the command and return the result. */
    exit_code = shell(command_buffer);

    /* If the exit code is nonzero, the invoked command reported an error. */
    if (exit_code != 0) {
        fail(
            "Command failed: %s: exit code = %d",
            command_buffer,
            exit_code
        );
    }
}

/*-------------------------------------------------------------------------*/

void run_command_as_user(const char * user_name, const char * command, ...)
{
    int result, exit_code, subshell_pid, subshell_status;
    char command_buffer[COMMAND_BUFFER_SIZE];
    va_list args;
    uid_t user_id;
    gid_t user_gid;

    assert(command);

    /* Format the command. */
    va_start(args, command);
    result = vsnprintf(command_buffer, COMMAND_BUFFER_SIZE, command, args);
    va_end(args);
    if ((result < 0) || (result >= COMMAND_BUFFER_SIZE)) {
        fail("Command string too long");
    }

    /* Spawn a subprocess. */
    subshell_pid = fork();
    if (subshell_pid < 0) {
        fail("Could not spawn a subprocess: %m");
    }

    if (subshell_pid == 0) {

        /* We are the subprocess. */

        /* Determine the user ID and primary group ID. */
        user_id = get_user_id(user_name);
        user_gid = get_user_primary_group_id(user_name);

        /* Become the testing user. */
        EXPECT_ZERO(
            setgid(user_gid),
            "Could not set process group ID"
        );
        EXPECT_ZERO(
            initgroups(user_name, user_gid),
            "Could not init process groups"
        );
        EXPECT_ZERO(
            setuid(user_id),
            "Could not set process user ID"
        );

        /* Run the command and return the result. */
        exit_code = shell(command_buffer);

        /* Exit and pass the exit code to the parent process. */
        exit(exit_code);
    }

    /* Wait for the subprocess to finish. */
    if (waitpid(subshell_pid, &subshell_status, 0) < 0) {
        fail("Error waiting for a subprocess: %m");
    }

    /* Check for abnormal termination and/or errors. */
    if (subshell_status != 0) {
        fail("Command failed: %s", command_buffer);
    }
}

/*-------------------------------------------------------------------------*/

int try_command(const char * command, ...)
{
    int result, exit_code;
    char command_buffer[COMMAND_BUFFER_SIZE];
    va_list args;

    assert(command);

    /* Format the command. */
    va_start(args, command);
    result = vsnprintf(command_buffer, COMMAND_BUFFER_SIZE, command, args);
    va_end(args);
    if ((result < 0) || (result >= COMMAND_BUFFER_SIZE)) {
        fail("Command string too long");
    }

    /* Run the command and return the result. */
    exit_code = shell(command_buffer);
    return exit_code;
}

/*-------------------------------------------------------------------------*/

char * capture(const char * command, ...)
{
    FILE * command_pipe;
    int bytes_got;
    int result;
    char command_buffer[COMMAND_BUFFER_SIZE];
    va_list args;

    assert(command);

    /* Format the command. */
    va_start(args, command);
    result = vsnprintf(command_buffer, COMMAND_BUFFER_SIZE, command, args);
    va_end(args);
    if ((result < 0) || (result >= COMMAND_BUFFER_SIZE)) {
        fail("Command string too long");
    }

    /* Open a shell with pipe and run the command. */
    command_pipe = popen(command_buffer, "r");
    if (command_pipe == NULL) {
        fail("Could not start shell command: %s: %m", command);
    }

    /* Transfer data from the pipe to the buffer until the pipe is closed
     * on the other side, or until the buffer is full.
     */
    bytes_got = fread(
        command_output_buffer,
        1,
        COMMAND_OUTPUT_BUFFER_SIZE - 1,
        command_pipe
    );

    /* Check for I/O errors. */
    if ((bytes_got < 0) || (ferror(command_pipe))) {
        fail("Error reading from pipe: %m");
    }

    /* Remove trailing whitespace. */
    while ((bytes_got > 0) && (isspace(command_output_buffer[bytes_got - 1]))) {
        bytes_got--;
    }

    /* Write the terminating zero byte. */
    command_output_buffer[bytes_got] = '\0';

    /* Close the pipe. */
    pclose(command_pipe);

    /* Return a pointer to the buffer. */
    return command_output_buffer;
}

/*-------------------------------------------------------------------------*/

/*
 * Checks whether the specified tool is known to the shell, terminating
 * the program immediately with an error message if the tool is not found.
 */
void check_for_tool(const char * name)
{
    if (try_command("which '%s' >/dev/null 2>&1", name) != 0) {
        fail("Could not find required tool: %s", name);
    }
}

/*-------------------------------------------------------------------------*/
