/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "string_array.h"
#include "string_utils.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

/*-------------------------------------------------------------------------*/

StringArray * string_array_create()
{
    /* Allocate the StringArray structure. */
    StringArray * array = (StringArray *) malloc(sizeof(StringArray));
    assert(array);

    /* Initialize it. */
    array->strings = NULL;
    array->length = 0;
    array->allocated = 0;

    /* And return a pointer to it. */
    return array;
}

/*-------------------------------------------------------------------------*/

void string_array_delete(StringArray * array)
{
    assert(array);

    /* Deallocate all strings and the slots array. */
    string_array_clear(array);

    /* Free the array structure itself. */
    free(array);
}

/*-------------------------------------------------------------------------*/

void string_array_clear(StringArray * array)
{
    int i;

    assert(array);

    /* Free all stored strings. */
    for (i = 0; i < array->length; i++) {
        free(array->strings[i]);
    }

    /* Free the slot storage. */
    free(array->strings);

    /* Reset the array structure. */
    array->length = 0;
    array->allocated = 0;
    array->strings = NULL;
}

/*-------------------------------------------------------------------------*/

/*
 * Reallocates the array to larger number of slots.
 */
static void string_array_reallocate(StringArray * array,
    unsigned int to_be_allocated)
{
    char ** new_strings;
    unsigned int i;

    assert(array);

    /* Do not allow to shrink the array in this way. */
    assert(array->allocated >= array->length);
    assert(to_be_allocated > array->allocated);

    /* Reallocate. */
    new_strings = (char **) realloc(
        array->strings,
        to_be_allocated * sizeof(char **)
    );
    assert(new_strings);
    array->strings = new_strings;

    /* Clear the newly allocated area. */
    for (i = array->allocated; i < to_be_allocated; i++) {
        array->strings[i] = NULL;
    }

    /* Remember the new allocated size. */
    array->allocated = to_be_allocated;
}

/*-------------------------------------------------------------------------*/

void string_array_prepare_for_length(StringArray * array,
    unsigned int desired_length)
{
    unsigned int to_be_allocated;

    assert(array);

    if (desired_length == 0) { return; }

    if (array->allocated < desired_length) {

        /* Round the number of slots to the nearest higher multiple of 16. */
        to_be_allocated = ((15 + desired_length) / 16) * 16;
        assert(to_be_allocated >= desired_length);

        /* Reallocate the array. */
        string_array_reallocate(array, to_be_allocated);
    }
}

/*-------------------------------------------------------------------------*/

void string_array_put(
    StringArray * array,
    unsigned int index,
    const char * source
)
{
    assert(array);

    /* Ensure that we have enough space. */
    string_array_prepare_for_length(array, index + 1);
    assert(array->strings);

    /* Free the previous content of the slot, if any. */
    free(array->strings[index]);

    /* Store the new string in the slot. */
    assert(index < array->allocated);
    array->strings[index] = string_duplicate(source);

    /* Enlarge the used area if the newly accessed slot is outside. */
    if (index >= array->length) { array->length = index + 1; }
}

/*-------------------------------------------------------------------------*/

const char * string_array_get(StringArray * array, unsigned int index)
{
    assert(array);

    assert(index < array->length);
    return array->strings[index];
}

/*-------------------------------------------------------------------------*/

void string_array_append(StringArray * array, const char * source)
{
    assert(array);

    string_array_put(array, array->length, source);
}

/*-------------------------------------------------------------------------*/

unsigned int string_array_count_lengths(StringArray * array)
{
    int i;
    unsigned int result = 0;

    assert(array);

    for (i = 0; i < array->length; i++) {
        if (array->strings[i]) {
            result += strlen(array->strings[i]);
        }
    }
    return result;
}

/*-------------------------------------------------------------------------*/

unsigned int string_array_count_valid_strings(StringArray * array)
{
    int i;
    unsigned int result = 0;

    assert(array);

    for (i = 0; i < array->length; i++) {
        if (array->strings[i]) {
            result++;
        }
    }
    return result;
}

/*-------------------------------------------------------------------------*/

char * string_array_implode(StringArray * array, const char * separator)
{
    int i;
    unsigned int result_length;
    char * result;
    unsigned int current_result_length = 0;
    unsigned int separator_length = 0;
    unsigned int string_count = 0;
    unsigned int current_string_length;
    const char * current_string;
    int first_string;

    assert(array);

    if (separator) {
        separator_length = strlen(separator);
    }

    /* How many strings we will concatenate. */
    string_count = string_array_count_valid_strings(array);
    if (string_count == 0) {

        /* Empty result is a special case; return an empty string. */
        return string_duplicate("");
    }

    /* Calculate the length of the result string, including the separators. */
    result_length = string_array_count_lengths(array);
    if (separator) {
        result_length += separator_length * (string_count - 1);
    }

    /* Allocate memory for the resulting string, plus one for the terminating
     * zero.
     */
    result = (char *) malloc(result_length + 1);
    assert(result);

    /* Append all valid strings to the result. */
    first_string = 1;
    for (i = 0; i < array->length; i++) {
        current_string = array->strings[i];
        if (current_string) {

            /* Append the separator first (except for the first string). */
            if ((!first_string != 0) && (separator)) {
                memcpy(
                    result + current_result_length,
                    separator,
                    separator_length
                );
                current_result_length += separator_length;
            }

            /* Append the string. */
            current_string_length = strlen(current_string);
            if (current_string_length > 0) {
                memcpy(
                    result + current_result_length,
                    current_string,
                    current_string_length
                );
                current_result_length += current_string_length;
            }

            first_string = 0;
        }
    }

    assert(current_result_length == result_length);

    /* Place the terminating zero. */
    result[result_length] = '\0';

    return result;
}

/*-------------------------------------------------------------------------*/
