/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#ifndef STRING_ARRAY_H
#define STRING_ARRAY_H

/*-------------------------------------------------------------------------*/

/**
 * A resizable string array.
 */
typedef struct {

    /**
     * Slots containing pointers to strings (NULL if the array is empty).
     */
    char ** strings;

    /**
     * Length of the array (number of slots currently in use).
     */
    unsigned int length;

    /**
     * Number of slots allocated (including the unused ones).
     */
    unsigned int allocated;
}
StringArray;

/*-------------------------------------------------------------------------*/

/**
 * Creates a new empty string array.
 */
StringArray * string_array_create();

/**
 * Deletes the string array.
 */
void string_array_delete(StringArray * array);

/**
 * Clears the array, setting its length to 0 and deallocating its storage.
 */
void string_array_clear(StringArray * array);

/**
 * Ensures that the array can accomodate at least the specified
 * number of strings, reallocating its storage if needed.
 */
void string_array_prepare_for_length(StringArray * array,
    unsigned int desired_length);

/**
 * Stores a copy of the string into the array.
 * If the given indexed slot already contains a string, the old string
 * is removed. If the source pointer is NULL, the old string is removed
 * and no new string is stored.
 * If the slot index lies outside the current array length, the array
 * is grown appropriately; new unused slots are filled with NULLs.
 */
void string_array_put(
    StringArray * array,
    unsigned int index,
    const char * source
);

/**
 * Returns a pointer to the string stored in the array.
 */
const char * string_array_get(StringArray * array, unsigned int index);

/**
 * Stores the string to the end of the array, enlarging it by one.
 */
void string_array_append(StringArray * array, const char * source);

/**
 * Sums the length of all strings in the array together (not including
 * the terminating zero characters).
 * NULLs in the array are ignored.
 */
unsigned int string_array_count_lengths(StringArray * array);

/**
 * Counts the non-NULL elements in the array.
 */
unsigned int string_array_count_valid_strings(StringArray * array);

/**
 * Concatenates all non-NULL strings in the array, using
 * the specified separator, and returns the result
 * as a newly allocated string.
 */
char * string_array_implode(StringArray * array, const char * separator);

/*-------------------------------------------------------------------------*/

#endif
