/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "string_utils.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

/*-------------------------------------------------------------------------*/

char * format_malloc(const char * format, ...)
{
    va_list args;
    char * buffer = NULL;
    size_t buffer_size = 256;   /* Could be enough for most strings. */
    size_t bytes_written;

    assert(format);

    va_start(args, format);

    for (;;) {

        /* (Re)allocate the buffer. */
        buffer = (char *) realloc(buffer, buffer_size);
        assert(buffer);

        /* Attempt to print the result into the buffer. */
        bytes_written = vsnprintf(buffer, buffer_size, format, args);
        if (bytes_written >= buffer_size) {

            /* In new glibc, the vsnprintf() call gives us the exact number
             * of bytes that would be needed; allocate this much and retry.
             */
            buffer_size = bytes_written + 1;
        }
        else if (bytes_written < 0) {

            /* In older glibc, the call just fails returning negative number;
             * we must iterate, so double the buffer size and retry.
             */
            buffer_size *= 2;
        }
        else {

            /* Success. */
            break;
        }
    }

    va_end(args);
    return buffer;
}

/*-------------------------------------------------------------------------*/

char * string_duplicate(const char * str)
{
    assert(str);
    return string_duplicate_max(str, strlen(str));
}

/*-------------------------------------------------------------------------*/

char * string_duplicate_max(const char * str, unsigned int length)
{
    char * result;

    assert(str);

    /* Produce an empty string (not NULL) if the requested length is 0. */
    if (length == 0) {
        result = (char *) malloc(1);
        result[0] = '\0';
        return result;
    }

    /* Allocate memory for the copy (plus one byte for the trailing zero). */
    result = (char *) malloc(length + 1);
    assert(result);

    /* Copy the source and put the trailing zero. */
    string_copy_max(result, str, length + 1);

    return result;
}

/*-------------------------------------------------------------------------*/

char * string_duplicate_up_to(const char * str, const char * terminator)
{
    char * terminator_position;

    assert(str);
    assert(terminator);

    /* If the terminator is found, duplicate the string only up to it. */
    terminator_position = strstr(str, terminator);
    if (terminator_position) {
        return string_duplicate_max(str, terminator_position - str);
    }

    /* Otherwise duplicate the whole string. */
    return string_duplicate(str);
}

/*-------------------------------------------------------------------------*/

int string_equals(const char * str, const char * pattern)
{
    assert(str);
    assert(pattern);

    return (strcmp(str, pattern) == 0);
}

/*-------------------------------------------------------------------------*/

int string_starts_with(const char * str, const char * pattern)
{
    assert(str);
    assert(pattern);

    /* Compare the initial portion of the string with the pattern. */
    return (strncmp(str, pattern, strlen(pattern)) == 0);
}

/*-------------------------------------------------------------------------*/

int string_ends_with(const char * str, const char * pattern)
{
    size_t str_length, pattern_length;

    assert(str);
    assert(pattern);

    /* Determine the lengths of both the string and the pattern. */
    str_length = strlen(str);
    pattern_length = strlen(pattern);
    if (str_length < pattern_length) {

        /* The string is shorter than the pattern. */
        return 0;
    }

    /* Compare the trailing portion of the string with the pattern. */
    return (
        strcmp(
            str + str_length - pattern_length,
            pattern
        ) == 0
    );
}

/*-------------------------------------------------------------------------*/

int string_empty(const char * str)
{
    assert(str);

    return (str[0] == '\0');
}

/*-------------------------------------------------------------------------*/

void string_replace(char ** old_string, char * new_string)
{
    assert(old_string);

    free(*old_string);
    *old_string = new_string;
}

/*-------------------------------------------------------------------------*/

void string_replace_with_copy(char ** old_string, const char * new_string)
{
    assert(old_string);

    free(*old_string);
    *old_string = string_duplicate(new_string);
}

/*-------------------------------------------------------------------------*/

int string_one_of(const char * str, ...)
{
    va_list options;
    const char * option;
    int match_found = 0;

    /* Scan the options and compare each one with the string. */
    va_start(options, str);
    for (;;) {
        option = va_arg(options, const char *);
        if (option == NULL) {

            /* End of the list. */
            break;
        }
        if (strcmp(str, option) == 0) {

            /* Match. */
            match_found = 1;
            break;
        }
    }
    va_end(options);

    /* Return the result. */
    return match_found;
}

/*-------------------------------------------------------------------------*/

char * string_to_safe_identifier(const char * source)
{
    unsigned int source_length, i;
    char * result;

    assert(source);

    /* Allocate a copy of the source string and get its length. */
    source_length = strlen(source);
    result = string_duplicate(source);

    /* Replace all non-alphanumeric, non-underscore characters
     * by underscores.
     */
    for (i = 0; i < source_length; i++) {
        if (isalnum(result[i]) || (result[i] == '_')) {
            continue;
        }
        result[i] = '_';
    }

    /* Return the result. */
    return result;
}

/*-------------------------------------------------------------------------*/

void string_copy_max(char * target, const char * source,
    unsigned int target_size)
{
    assert(target_size > 1);
    assert(source);
    assert(target);

    /* Copy characters up to the target size, minus 1 for the zero. */
    strncpy(target, source, target_size - 1);

    /* Write the terminating zero. */
    target[target_size - 1] = '\0';
}

/*-------------------------------------------------------------------------*/

char * string_append_and_realloc(char * target, const char * source)
{
    size_t target_length, source_length;
    char * new_target;

    assert(target);
    assert(source);

    // Measure both strings and allocate a buffer to fit them both.
    source_length = strlen(source);
    target_length = strlen(target);
    new_target = (char *) malloc(target_length + source_length + 1);

    // Append both strings into the new buffer.
    memcpy(new_target, target, target_length);
    memcpy(new_target + target_length, source, source_length);

    // Append the terminating zero.
    new_target[target_length + source_length] = '\0';

    // Free the old target buffer and return the new one.
    free(target);
    return new_target;
}

/*-------------------------------------------------------------------------*/

sstring * sstring_create(void)
{
    sstring * sstr;

    /* Allocate the memory for the structure. */
    sstr = (sstring *) malloc(sizeof(sstring));
    assert(sstr);

    /* Initialize the structure to defined empty state. */
    sstr->buffer = NULL;
    sstr->length = 0;

    /* Return a pointer to it. */
    return sstr;
}

/*-------------------------------------------------------------------------*/

sstring * sstring_create_from_cstring(const char * source)
{
    sstring * sstr;

    /* Allocate the memory for the structure. */
    sstr = (sstring *) malloc(sizeof(sstring));
    assert(sstr);

    /* Allocate buffer and copy the source including the terminating zero. */
    assert(source);
    sstr->length = strlen(source);
    sstr->buffer = (char *) malloc(sstr->length + 1);
    memcpy(sstr->buffer, source, sstr->length + 1);

    /* Return a pointer to it. */
    return sstr;
}

/*-------------------------------------------------------------------------*/

void sstring_delete(sstring * sstr)
{
    assert(sstr);

    /* First clear the old contents, then delete the whole structure. */
    sstring_clear(sstr);
    free(sstr);
}

/*-------------------------------------------------------------------------*/

void sstring_clear(sstring * sstr)
{
    assert(sstr);

    /* Free the buffer. */
    free(sstr->buffer);

    /* Reset the structured string state to initial empty state. */
    sstr->buffer = NULL;
    sstr->length = 0;
}

/*-------------------------------------------------------------------------*/

void sstring_set(sstring * sstr, const char * source)
{
    assert(source);
    sstring_set_exact(sstr, source, strlen(source));
}

/*-------------------------------------------------------------------------*/

void sstring_set_exact(sstring * sstr, const char * source,
    unsigned int source_length)
{
    char * new_buffer;

    /* Zero length is the same as sstring_clear(). */
    if (source_length == 0) {
        sstring_clear(sstr);
        return;
    }

    /* The source must not be NULL if source_length is nonzero. */
    assert(source);

    /* Allocate a new buffer of appropriate size. */
    new_buffer = (char *) malloc(source_length + 1);
    assert(new_buffer);

    /* Copy the source and put the terminating zero. */
    memcpy(new_buffer, source, source_length);
    new_buffer[source_length] = '\0';

    /* Free the old buffer and replace it with the new buffer. */
    sstring_clear(sstr);
    sstr->buffer = new_buffer;
    sstr->length = source_length;
}

/*-------------------------------------------------------------------------*/

sstring * sstring_printf(sstring * target, const char * format, ...)
{
    va_list args;
    char * buffer = NULL;
    size_t buffer_size = 256;   /* Could be enough for most strings. */
    size_t bytes_written;

    /* The formatting string must be valid. */
    assert(format);

    /* Create a new structured string if none is specified. */
    if (target == NULL) {
        target = sstring_create();
    }

    /* Build the resulting string in a dynamically allocated buffer. */
    va_start(args, format);
    for (;;) {

        /* (Re)allocate the buffer. */
        buffer = (char *) realloc(buffer, buffer_size);
        assert(buffer);

        /* Attempt to print the result into the buffer. */
        bytes_written = vsnprintf(buffer, buffer_size, format, args);
        if (bytes_written >= buffer_size) {

            /* In new glibc, the vsnprintf() call gives us the exact number
             * of bytes that would be needed; allocate this much and retry.
             */
            buffer_size = bytes_written + 1;
        }
        else if (bytes_written < 0) {

            /* In older glibc, the call just fails returning negative number;
             * we must iterate, so double the buffer size and retry.
             */
            buffer_size *= 2;
        }
        else {

            /* Success. */
            break;
        }
    }
    va_end(args);

    /* Replace the original contents of the sstring with the buffer. */
    sstring_clear(target);
    target->buffer = buffer;
    target->length = bytes_written;

    /* Return the modified sstring. */
    return target;
}

/*-------------------------------------------------------------------------*/

void sstring_copy(sstring * target, const sstring * source)
{
    assert(target);
    assert(source);
    assert(target != source);

    /* Remove the previous contents of the target. */
    sstring_clear(target);

    /* Allocate buffer and copy the source including the terminating zero. */
    target->length = source->length;
    target->buffer = (char *) malloc(target->length + 1);
    memcpy(target->buffer, source->buffer, target->length + 1);

}

/*-------------------------------------------------------------------------*/

void sstring_filename_to_identifier(sstring * target)
{
    unsigned int i;

    assert(target);

    /* Do nothing if the target is empty. */
    if (target->length == 0) {
        return;
    }

    /* Remove the leading underscore if present. */
    if (target->buffer[0] == '/') {
        sstring_remove_first_character(target);
    }

    /* Replace slashes by underscores. */
    for (i = 0; i < target->length; i++) {
        if (target->buffer[i] == '/') {
            target->buffer[i] = '_';
        }
    }
}

/*-------------------------------------------------------------------------*/

void sstring_remove_first_character(sstring * target)
{
    assert(target);

    /* Do nothing if the target is empty. */
    if (target->length == 0) {
        return;
    }

    /* Shift characters (including the terminating zero) by one. */
    memmove(target->buffer, target->buffer + 1, target->length + 1);

    /* Reduce the length. */
    target->length--;
}

/*-------------------------------------------------------------------------*/

void sstring_filename_to_basename(sstring * target)
{
    const char * last_slash;
    const char * filename_part;
    const char * extension_dot;

    /* Do nothing if the target is empty. */
    if (target->length == 0) {
        return;
    }

    /* Find where the filename starts. */
    last_slash = strrchr(target->buffer, '/');
    if (last_slash) {
        filename_part = last_slash + 1;
    }
    else {
        filename_part = target->buffer;
    }

    /* Find the (first) extension dot in the filename. */
    extension_dot = strchr(filename_part, '.');

    /* Return the filename up to the first dot. */
    if (extension_dot) {
        sstring_set_exact(target, filename_part,
            extension_dot - filename_part);
    }
    else {

        /* There is no extension dot, return the whole filename. */
        sstring_set(target, filename_part);
    }
}

/*-------------------------------------------------------------------------*/

void string_tr(char * str, char searchfor, char replacewith)
{
    while ( str[0] )
    {
        if (str[0] == searchfor)
            str[0] = replacewith;
        str++;
    }
}

/*-------------------------------------------------------------------------*/
