/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#ifndef STRING_UTILS_H
#define STRING_UTILS_H

/*-------------------------------------------------------------------------*/

/**
 * Formats the message in printf-style to an automatically sized block
 * of memory allocated by malloc().
 */
char * format_malloc(const char * format, ...);

/**
 * Allocates a copy of the specified string via malloc().
 */
char * string_duplicate(const char * str);

/**
 * Allocates a copy of the specified string via malloc(),
 * copying at most the specified number of characters (the terminating
 * zero is added automatically and does not count into the length).
 * If the specified length is 0, the resulting string contains only
 * the terminating zero.
 */
char * string_duplicate_max(const char * str, unsigned int length);

/**
 * Duplicates the string up to the first occurrence of the specified
 * terminator string. If the terminator string is not present,
 * the whole string is duplicated.
 */
char * string_duplicate_up_to(const char * str, const char * terminator);

/**
 * Returns true if the string exactly equals to the specified pattern,
 * false otherwise.
 */
int string_equals(const char * str, const char * pattern);

/**
 * Returns true if the string starts with the specified pattern,
 * false otherwise.
 */
int string_starts_with(const char * str, const char * pattern);

/**
 * Returns true if the string ends with the specified pattern,
 * false otherwise.
 */
int string_ends_with(const char * str, const char * pattern);

/**
 * Returns true if the string is empty, false otherwise.
 */
int string_empty(const char * str);

/**
 * Frees the old string (if any) and stores the new pointer
 * instead.
 */
void string_replace(char ** old_string, char * new_string);

/**
 * Frees the old string and stores a newly allocated copy of the new string
 * instead.
 */
void string_replace_with_copy(char ** old_string, const char * new_string);

/**
 * Returns true if the string is equal to one of the specified options,
 * otherwise returns false.
 * The options must be a list of strings terminated by a NULL.
 */
int string_one_of(const char * str, ...);

/**
 * Transforms the given string to a safe identifier
 * by replacing all 'dangerous' characters by underscores.
 * The result is a newly allocated string.
 */
char * string_to_safe_identifier(const char * source);

/**
 * Copies a string from source to target, using no more than the specified
 * number of bytes including the terminating zero (therefore, at most
 * target_size - 1 characters are copied, then the zero is added).
 */
void string_copy_max(char * target, const char * source,
    unsigned int target_size);

/**
 * Appends the source to the target, reallocating the target as needed.
 * The target must (obviously) be a dynamically allocated string.
 */
char * string_append_and_realloc(char * target, const char * source);

/**
 * Replaces all characters 'searchfor' in the 'str' with 'replacewith'
 */
void string_tr(char * str, char searchfor, char replacewith);
/*-------------------------------------------------------------------------*/

/**
 * A structured string.
 */
typedef struct
{
    /**
     * A dynamically allocated buffer holding the string.
     * The string is zero terminated.
     */
    char * buffer;

    /**
     * Length of the string in bytes, not including the terminating
     * zero byte.
     */
    unsigned int length;
}
sstring;

/*-------------------------------------------------------------------------*/

/**
 * Returns a pointer to the text content of the structured string
 * as a C-style string.
 */
#define CSTRING(sstr) (sstr->buffer)

/**
 * Creates an empty structured string.
 */
sstring * sstring_create(void);

/**
 * Creates a structured string containing the copy of the specified
 * C-style string.
 */
sstring * sstring_create_from_cstring(const char * source);

/**
 * Deletes a structured string.
 */
void sstring_delete(sstring * sstr);

/**
 * Clears the structured string, making it empty again.
 */
void sstring_clear(sstring * sstr);

/**
 * Stores a copy of the source string to the structured string.
 * Old contents of the structured string are removed.
 *
 * @note It is safe to use a part of the structured string itself
 *       as the source.
 */
void sstring_set(sstring * sstr, const char * source);

/**
 * Stores an exact number of characters from the source string
 * to the structured string.
 * Old contents of the structured string are removed.
 *
 * @note It is safe to use a part of the structured string itself
 *       as the source.
 */
void sstring_set_exact(sstring * sstr, const char * source,
    unsigned int source_length);

/**
 * Prints a formatted string in printf-style and stores it into
 * the specified structured string.
 * If the target pointer is NULL, a new structured string is allocated.
 */
sstring * sstring_printf(sstring * target, const char * format, ...);

/**
 * Copies the source string to the target, removing the old
 * contents of the target.
 */
void sstring_copy(sstring * target, const sstring * source);

/**
 * Produces an identifier from a filename by replacing all filesystem-special
 * characters by underscores. The result can safely be used as a filename
 * as well.
 */
void sstring_filename_to_identifier(sstring * target);

/**
 * Removes the first character from the string.
 */
void sstring_remove_first_character(sstring * target);

void sstring_filename_to_basename(sstring * target);

/*-------------------------------------------------------------------------*/

#endif
