/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#define _XOPEN_SOURCE 600
#define _POSIX_C_SOURCE 200112L
#include "subshell.h"
#include "lsb-autotest.h"
#include "string_utils.h"
#include "useful_macros.h"
#include "command_line.h"
#include "user_utils.h"
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <sys/wait.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <pty.h>
#include <grp.h>

/*-------------------------------------------------------------------------*/

pid_t subshell_pid = (pid_t) -1;
int pty_master_fd = -1;
FILE * pty_master_file = NULL;
FILE * subshell_log_file = NULL;

/*
 * Number of characters logged in this session (between opening
 * and closing of the log).
 */
static int characters_logged = 0;

/*
 * Last time when a character was logged.
 */
static time_t last_logging_time;

/*
 * Last time when a subshell status report was printed.
 */
static time_t last_reporting_time;

/*
 * How much seconds to wait before a notification about a stalling subshell
 * is printed.
 */
#define READ_NOTIFICATION_TIMEOUT 60

/*
 * How much seconds to wait when the shell stops producing any output
 * before we decide it is dead. Default: 1 hour.
 */
#define READ_PATIENCE 3600

/*-------------------------------------------------------------------------*/

void init_utmp_entry(
    struct utmp * utmp_entry,
    const char * user_name,
    const char * tty_name,
    pid_t shell_pid,
    int login
)
{
    time_t current_time;

    assert(utmp_entry);
    assert(tty_name);
    assert(strlen(tty_name) > 8);

    /* Clear the entry to zeros first. */
    memset(utmp_entry, 0, sizeof(utmp_entry));

    /* If we want a login entry, simulate a shell process running
     * on the terminal; if we want a logout, simulate an already
     * finished shell.
     */
    utmp_entry->ut_type = login ? USER_PROCESS : DEAD_PROCESS;

    /* For a login entry, a terminal line must be specified;
     * for a logout entry, it stays filled with zeros.
     */
    if (login) {
        strcpy(utmp_entry->ut_line, tty_name + strlen("/dev/"));
    }

    /* Fill in the utmp id from the terminal name. */
    strcpy(utmp_entry->ut_id, tty_name + strlen("/dev/tty"));

    /* For a login entry, a user name must be specified;
     * for a logout entry, it stays filled with zeros.
     */
    if (login) {
        assert(user_name);
        strcpy(utmp_entry->ut_user, user_name);
    }

    /* Store the specified PID as the shell PID. */
    utmp_entry->ut_pid = shell_pid;

    /* For a login entry, a login time must be specified;
     * for a logout entry, it stays zero.
     */
    if (login) {
        time(&current_time);
        utmp_entry->ut_time = current_time;
    }

    /* The 'ut_host' and 'ut_addr' fields are left empty because we simulate
     * a local connection. Also the 'ut_session' field is left empty
     * as it is not used for terminal sessions.
     */
}

/*-------------------------------------------------------------------------*/

void write_utmp_entry(struct utmp * utmp_entry)
{
    assert(utmp_entry);

    /* Open the utmp file. NOTE: This function does not return any status
     * information, and it seems to sometimes set errno even if it worked.
     * Therefore, no error checking here, we just hope it worked.
     */
    setutent();

    /* Write the entry. */
    if (pututline(utmp_entry) == NULL) {
        fail("Could not write the utmp entry: %m");
    }

    /* Close the utmp file. */
    endutent();
}

/*-------------------------------------------------------------------------*/

/*
 * Opens the log file where the copy of the subshell conversation
 * will be stored. Also resets logging timers and logged character counter.
 */
static void open_subshell_log(const char * file_name)
{
    /* Open the file. */
    subshell_log_file = fopen(file_name, "a");
    if (subshell_log_file == NULL) {
        fail(
            "Could not open subshell log for writing: %s: %m",
            file_name
        );
    }

    /* Reset the logged character counter. */
    characters_logged = 0;

    /* Reset the timers. */
    last_logging_time = time(NULL);
    last_reporting_time = last_logging_time;
}

/*-------------------------------------------------------------------------*/

/*
 * Closes the log file.
 */
static void close_subshell_log(void)
{
    if (subshell_log_file) {
        fclose(subshell_log_file);
        subshell_log_file = NULL;
    }
}

/*-------------------------------------------------------------------------*/

/*
 * Echoes a character to the given file. Unprintable characters are
 * replaced by dots.
 */
static void echo_character(int c, FILE * f)
{
    if ((c >= 0x20) && (c != 0x7f)) {
        fputc(c, f);
    }
    else if ((c == '\n') || (c == '\r') || (c == '\t')) {
        fputc(c, f);
    }
    else {
        fputc('.', f);
    }
}

/*-------------------------------------------------------------------------*/

/*
 * Appends a character to the log.
 * If the verbose mode is enabled, the character is also echoed
 * on the console. In non-verbose mode, only short messages every
 * 5 sec. is printed, informing about the amount of data received.
 */
static void log_character(int c)
{
    time_t now;

    if (subshell_log_file) {
        echo_character(c, subshell_log_file);
    }

    /* Update the character counter. */
    characters_logged++;

    if ( verbose ) {
        echo_character(c, stdout);
        fflush(stdout);
    }

        /* Print status message at most every 5 seconds. */
        now = time(NULL);
        if (last_reporting_time + 5 < now) {
        if (!verbose) {
            inform(
                "%u bytes received from the subshell so far",
                characters_logged
            );
        }
            last_reporting_time = now;
        }

        /* Update the last logging time. */
        last_logging_time = now;
    }

/*-------------------------------------------------------------------------*/

/*
 * Prints an information about how long the subshell output stalled.
 */
static void print_output_stall_duration(time_t stall)
{
    if (stall > 3600) {
        inform(
            "Subshell output stalled for %u hours, %u minutes, %u seconds",
            stall / 3600,
            stall / 60,
            stall % 60
        );
    }
    else if (stall > 60) {
        inform(
            "Subshell output stalled for %u minutes, %u seconds",
            stall / 60,
            stall % 60
        );
    }
    else {
        inform(
            "Subshell output stalled for %u seconds",
            stall
        );
    }
}

/*-------------------------------------------------------------------------*/

/*
 * Reads a single character from the subshell.
 */
static int read_character()
{
    int c;
    time_t read_start_time = time(NULL);
    time_t now;
    time_t stall_duration;

    for (;;) {

        /* Set up the timeout. */
        alarm(READ_NOTIFICATION_TIMEOUT);

        /* Read the character. */
        c = fgetc(pty_master_file);
        alarm(0);

        /* Normal characters are simply returned. */
        if (c != EOF) { return c; }

        /* Check for end-of-file, timeouts, and errors. */
        if (feof(pty_master_file)) {

            /* A normal end-of-file. */
            return EOF;
        }

        if (errno == EINTR) {

            /* Timeout. */

            /* How long the stall is. */
            now = time(NULL);
            stall_duration = now - read_start_time;

            if (!verbose) {
                print_output_stall_duration(stall_duration);
            }

            /* Check for probable deadlocks. */
            if (stall_duration > READ_PATIENCE) {
                fail("Subshell stalled for too long, terminating.");
            }

            /* Retry. */
            continue;
        }
        else if (errno == EIO) {

            /*
             * This seems to be also a correct closing of the other side.
             * Bash does this; probably it does not send EOF explicitly or
             * something like this. Ignore and return EOF.
             */
            return EOF;
        }
        else {

            /* Another error. */
            fail("I/O error reading from subshell: %m");
        }
    }
}

/*-------------------------------------------------------------------------*/

void open_subshell(
    const char * user_name,
    const char * start_dir,
    const char * log_file_name
)
{
    uid_t user_id;
    gid_t user_gid, tty_gid;
    char * user_home_dir;
    char * subshell_env[3];
    struct utmp utmp_entry;
    const char * tty_name;

    assert(user_name);
    assert(start_dir);

    /* Determine the user ID and primary group ID so that we can
     * run the subprocess as that user.
     */
    user_id = get_user_id(user_name);
    user_gid = get_user_primary_group_id(user_name);

    /* Get the user's home directory. */
    user_home_dir = get_user_home_dir(user_name);

    /* Determine the group ID used for terminal devices; we need it
     * for correct initialization of the terminal access rights.
     */
    tty_gid = get_group_id("tty");

    /* Obtain a pseudo-terminal and run a sub-process as the slave. */
    subshell_pid = forkpty(&pty_master_fd, NULL, NULL, NULL);
    if (subshell_pid < 0) {
        fail("Could not create a sub-process: %m");
    }
    if (subshell_pid == 0) {

        /* We are the subprocess. */

        /* Get our PID and controlling terminal. */
        subshell_pid = getpid();
        tty_name = ttyname(0);

        /* Make the test user the owner of the slave part of the terminal;
         * this is required for many tests that read or modify terminal
         * settings.
         */
        EXPECT_ZERO(
            chown(tty_name, user_id, tty_gid),
            "Could not change the owner of a tty"
        );

        /* Allow non-owner to read from the terminal; this is needed
         * for the shell to work correctly.
         */
        EXPECT_ZERO(
            chmod(tty_name, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH),
            "Could not change the access rights of a tty"
        );

        /* Create an utmp entry to simulate a real login on the terminal;
         * some tests need this.
         */
        init_utmp_entry(&utmp_entry, user_name, tty_name, subshell_pid, 1);
        write_utmp_entry(&utmp_entry);

        /* Become the testing user. */
        EXPECT_ZERO(
            setgid(user_gid),
            "Could not set process group ID"
        );
        EXPECT_ZERO(
            initgroups(user_name, user_gid),
            "Could not init process groups"
        );
        EXPECT_ZERO(
            setuid(user_id),
            "Could not set process user ID"
        );

        /* Prepare the command line and environment for the shell. */
        subshell_env[0] = format_malloc("TERM=%s", getenv("TERM"));
        subshell_env[1] = format_malloc("HOME=%s", user_home_dir);
        subshell_env[2] = NULL;

        /* Switch to the starting directory. */
        EXPECT_ZERO(
            chdir(start_dir),
            "Could not switch to directory"
        );

        /* Run the shell. */
        execle("/bin/bash", "/bin/bash", "-l", "-i", (char *) NULL,
            subshell_env);

        /* We can get here only if the execl() call fails. */
        fail("Could not exec the login shell: %m");
    }

    /* Open the master side of the terminal to communicate
     * with the subshell.
     */
    pty_master_file = fdopen(pty_master_fd, "a+");
    assert(pty_master_file);

    inform("Started a subshell (pid = %u)", subshell_pid);

    /* Open the log. */
    if (log_file_name) {
        open_subshell_log(log_file_name);
    }

    /* Clean up. */
    free(user_home_dir);
}

/*-------------------------------------------------------------------------*/

void close_subshell()
{
    struct utmp utmp_entry;

    /* Write the logout entry to the utmp file. */
    init_utmp_entry(&utmp_entry, NULL, ptsname(pty_master_fd), subshell_pid, 0);
    write_utmp_entry(&utmp_entry);

    /* Close the connection to the master side of the terminal. */
    assert(pty_master_file);
    fclose(pty_master_file);

    /* Close the log file. */
    close_subshell_log();

    /* Reinitialize state variables. */
    pty_master_file = NULL;
    pty_master_fd = -1;
    subshell_pid = -1;

    inform("Subshell closed");
}

/*-------------------------------------------------------------------------*/

void subshell_send(const char * message)
{
    assert(pty_master_file != NULL);
    assert(message);

    /* Send the message to the subshell. */
    if (fprintf(pty_master_file, "%s\r", message) < 0) {
        fail("I/O error while writing to subshell: %m");
    }

    /* Wait until the output is really sent. */
    fflush(pty_master_file);
}

/*-------------------------------------------------------------------------*/

void subshell_expect(const char * pattern)
{
    int match_mark = 0;
    int c;

    assert(pattern != NULL);
    assert(pty_master_file != NULL);

    /* Wait for the pattern, echoing everything to the terminal. */
    for (;;) {

        /* Get the next character. */
        c = read_character();

        /* Check for EOF and I/O errors. */
        if (c < 0) {
            fail(
                "Unexpected EOF or I/O error from the subshell "
                "(expecting '%s')",
                pattern
            );
        }

        /* Log the character. */
        log_character(c);

        /* Check if the character matches the pattern. */
        if (c == pattern[match_mark]) {
            match_mark++;
            if (pattern[match_mark] == '\0') {

                /* Complete match found. */
                break;
            }
        }
        else {

            /* This is not the pattern. */
            match_mark = 0;

            /* Check if it could be the start of another pattern. */
            if (c == pattern[0]) {
                match_mark = 1;
            }
        }
    }
    usleep(200000);     /* sleep 0.2 second */
}

/*-------------------------------------------------------------------------*/

void subshell_expect_and_reply(const char * pattern, const char * message)
{
    subshell_expect(pattern);
    subshell_send(message);
}

/*-------------------------------------------------------------------------*/

void wait_for_subshell()
{
    int subshell_status;
    int c;

    inform("Waiting for subshell (pid = %u) to complete", subshell_pid);

    /*
     * Read everything printed by the subshell (this is necessary
     * even if we are not interested in the data because the pipe
     * would otherwise overflow and deadlock) and wait until it closes
     * the pipe, which is a signal it is done.
     */
    for (;;) {

        /* Read the next character. */
        c = read_character();

        /* Stop if the pipe was closed. */
        if (c == EOF) { break; }

        /* Log the character. */
        log_character(c);
    }

    /* Ensure that all output is committed to the log file. */
    fflush(subshell_log_file);

    /* Wait for the subshell to really exit. */
    assert(subshell_pid > 0);
    waitpid(subshell_pid, &subshell_status, 0);

    /* Check whether the subshell exited cleanly. */
    if (!WIFEXITED(subshell_status)) {
        fail("The subshell has finished abnormally");
    }
}

/*-------------------------------------------------------------------------*/
