/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "tests.h"
#include "lsb-autotest.h"
#include "run_command.h"
#include "inform.h"
#include "package.h"
#include "backdoor.h"
#include "subshell.h"
#include "command_line.h"
#include "string_utils.h"
#include "user_utils.h"
#include "file_utils.h"
#include "run_command.h"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/*-------------------------------------------------------------------------*/

/* Passwords for the test users in the runtime test.
 * Can be completely arbitrary.
 */
#define RUNTIME_TEST_USER1_PASSWORD "gu112jj6"
#define RUNTIME_TEST_USER2_PASSWORD "gu112jj4"
#define RUNTIME_TEST_USER3_PASSWORD "gu112jj2"

/*-------------------------------------------------------------------------*/

int tests_to_run = 0;

/*-------------------------------------------------------------------------*/

/* Directory that contains X11 fonts. */
static const char * x_font_dir;

/* Font paths for use with X tests. */
static char * xt_fontpath;
static char * xt_fontpath_good;

/*-------------------------------------------------------------------------*/

void print_selected_tests()
{
    inform(
        "Tests to be run:%s%s%s%s%s%s%s%s%s%s",
        (tests_to_run & RUNTIME_TEST_BIT) ? " runtime" : "",
        (tests_to_run & LIBSTDCPP_TEST_BIT) ? " libstdc++" : "",
        (tests_to_run & CMDCHK_TEST_BIT) ? " cmdchk" : "",
        (tests_to_run & LIBCHK_TEST_BIT) ? " libchk" : "",
        (tests_to_run & VSW4_TEST_BIT) ? " vsw4" : "",
        (tests_to_run & XTS5_TEST_BIT) ? " xts5" : "",
        (tests_to_run & DESKTOP_TEST_BIT) ? " desktop" : "",
        (tests_to_run & TCL_TEST_BIT) ? " tcl" : "",
        (tests_to_run & EXPECT_TEST_BIT) ? " expect" : "",
        (tests_to_run & PYTHON_TEST_BIT) ? " python" : ""
    );
}

/*-------------------------------------------------------------------------*/

/*
 * Runs the specified test in a separate process. This is useful because
 * if the test fails, we still can run the others. If the test fails,
 * it is written to the list of failed tests in the output directory.
 */
static void run_test(const char * test_name, void (*test_fn)())
{
    pid_t test_pid;
    int test_exit_status;
    char * file_name;

    test_pid = fork();
    if (test_pid == 0) {

        /* We are the test process. */

        /* Run the test and exit. */
        test_fn();
        exit(0);
    }
    else if (test_pid < 0) {

        /* Error in fork(). */
        fail("Could not spawn a sub-process: %m");
    }
    else {

        /* We are the parent process. */

        /* Print information about the test. */
        inform("Executing test: %s (pid = %u)", test_name, test_pid);

        /* Wait for the test process to finish. */
        waitpid(test_pid, &test_exit_status, 0);

        /* Did the test process finish successfully? */
        if (
            WIFEXITED(test_exit_status) && WEXITSTATUS(test_exit_status) == 0
        ) {
            /* The test was successful. */
            return;
        }

        /* The test has failed; write its name to the list of failed tests. */
        file_name = format_malloc(
            "%s/autotest_info/failed_tests",
            output_dir
        );
        append_string_to_file(file_name, "%s\n", test_name);
        free(file_name);
    }
}

/*-------------------------------------------------------------------------*/

void run_selected_tests()
{
    /* Prepare all the selected tests first. */
    if (tests_to_run & RUNTIME_TEST_BIT) {
        inform("Preparing test: %s", "runtime");
        prepare_runtime_test();
    }
    if (tests_to_run & LIBSTDCPP_TEST_BIT) {
        inform("Preparing test: %s", "libstdcpp");
        prepare_libstdcpp_test();
    }
    if (tests_to_run & CMDCHK_TEST_BIT) {
        inform("Preparing test: %s", "cmdchk");
        prepare_cmdchk_test();
    }
    if (tests_to_run & LIBCHK_TEST_BIT) {
        inform("Preparing test: %s", "libchk");
        prepare_libchk_test();
    }
    if (tests_to_run & VSW4_TEST_BIT) {
        inform("Preparing test: %s", "vsw4");
        prepare_vsw4_test();
    }
    if (tests_to_run & XTS5_TEST_BIT) {
        inform("Preparing test: %s", "xts5");
        prepare_xts5_test();
    }
    if (tests_to_run & DESKTOP_TEST_BIT) {
        inform("Preparing test: %s", "desktop");
        prepare_desktop_test();
    }
    if (tests_to_run & TCL_TEST_BIT) {
        inform("Preparing test: %s", "tcl");
        prepare_tcl_test();
    }
    if (tests_to_run & EXPECT_TEST_BIT) {
        inform("Preparing test: %s", "expect");
        prepare_expect_test();
    }
    if (tests_to_run & PYTHON_TEST_BIT) {
        inform("Preparing test: %s", "python");
        prepare_python_test();
    }

    /* Now run them. */
    if (tests_to_run & RUNTIME_TEST_BIT) {
        run_test("runtime", run_runtime_test);
    }
    if (tests_to_run & LIBSTDCPP_TEST_BIT) {
        run_test("libstdcpp", run_libstdcpp_test);
    }
    if (tests_to_run & CMDCHK_TEST_BIT) {
        run_test("cmdchk", run_cmdchk_test);
    }
    if (tests_to_run & LIBCHK_TEST_BIT) {
        run_test("libchk", run_libchk_test);
    }
    if (tests_to_run & VSW4_TEST_BIT) {
        run_test("vsw4", run_vsw4_test);
    }
    if (tests_to_run & XTS5_TEST_BIT) {
        run_test("xts5", run_xts5_test);
    }
    if (tests_to_run & DESKTOP_TEST_BIT) {
        run_test("desktop", run_desktop_test);
    }
    if (tests_to_run & TCL_TEST_BIT) {
        run_test("tcl", run_tcl_test);
    }
    if (tests_to_run & EXPECT_TEST_BIT) {
        run_test("expect", run_expect_test);
    }
    if (tests_to_run & PYTHON_TEST_BIT) {
        run_test("python", run_python_test);
    }
}

/*-------------------------------------------------------------------------*/

static void determine_x_font_paths()
{
    /* Look where the basic fonts could be. */
    x_font_dir = "/usr/X11R6/lib/X11/fonts";
    if (!file_exists(x_font_dir)) {
        x_font_dir = "/usr/share/fonts";
    }
    if (!file_exists(x_font_dir)) {
        fail("Could not find the 'misc' font directory");
    }

    /* Build paths to fonts. */
    xt_fontpath = format_malloc(
        "/opt/lsb/test/vsw4/xtest/fonts/,%s/misc/",
        x_font_dir
    );
    xt_fontpath_good = format_malloc(
        "%s/100dpi/,%s/75dpi/",
        x_font_dir,
        x_font_dir
    );
}

/*-------------------------------------------------------------------------*/

static void cleanup_x_font_paths()
{
    free(xt_fontpath);
    free(xt_fontpath_good);
    xt_fontpath = NULL;
    xt_fontpath_good = NULL;
}

/*-------------------------------------------------------------------------*/

void prepare_runtime_test()
{
    /* Check whether the test package is installed, install it if needed. */
    check_package("lsb-tet3-lite", "runtime");
    check_package("lsb-runtime-test", "runtime");

    /* Remove stale config files. */
    unlink("/home/tet/test_sets/.configured");
    unlink("/home/tet/test_sets/loopback_disk.ext2");
    unlink("/home/tet/test_sets/TESTROOT/tetexec.cfg.old");

    /* Purge stale locks (can be left behind if the previous run
     * was stopped or crashed).
     */
    run_command("rm -f `find /home/tet/test_sets/TESTROOT/tset -name tet_lock`");

    /* Set passwords for testing users. */
    open_subshell("root", "/home/tet/test_sets", NULL);

    subshell_send("tty");
    
    subshell_send("passwd vsx0");
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER1_PASSWORD);
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER1_PASSWORD);
    
    subshell_send("passwd vsx1");
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER2_PASSWORD);
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER2_PASSWORD);

    subshell_send("passwd vsx2");
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER3_PASSWORD);
    subshell_expect_and_reply("assword:", RUNTIME_TEST_USER3_PASSWORD);
    
    close_subshell();
}

/*-------------------------------------------------------------------------*/

void run_runtime_test()
{
    sstring * log_file_name = sstring_create();
    sstring_printf(log_file_name, "%s/test_logs/runtime", output_dir);

    /* Install the backdoor needed for running 'su' without password
     * in the test setup.
     */
    install_backdoor();

    /* Open the subshell the test will run in. */
    open_subshell("vsx0", "/home/tet/test_sets", CSTRING(log_file_name));

    /* Prepare the environment. */
    //subshell_send("export HOME=/home/tet/test_sets");
    subshell_send(". profile");

    /* Run the tests. */
    subshell_send("./run_tests ; logout");

    /* Answer the questions. */
    subshell_expect("Name of person running tests");
    subshell_expect_and_reply("?", "Automated");
    subshell_expect("Organisation");
    subshell_expect_and_reply("?", "N/A");
    subshell_expect("Test System");
    subshell_expect_and_reply("?", "LSB Autotest");
    subshell_expect("Block special filename");
    subshell_expect_and_reply("?", "/home/tet/test_sets/nonexistb");
    subshell_expect("Does the implementation provide bash as /bin/sh");
    subshell_expect_and_reply("?", "y");
    subshell_expect("Does the implementation provide a C shell");
    subshell_expect_and_reply("?", "y");
    subshell_expect("Enter the name of the Kernel");
    subshell_expect_and_reply("?", "vmlinuz");
    subshell_expect("Does the implementation allow users to create devices using MAKEDEV");
    subshell_expect_and_reply("?", "n");
    subshell_expect("Does the implementation support the file command");
    subshell_expect_and_reply("?", "y");
    subshell_expect("Does the implementation support process accounting");
    subshell_expect_and_reply("?", "n");
    subshell_expect("Does the implementation support NIS");
    subshell_expect_and_reply("?", "n");
    subshell_expect("name of the user for PAM tests");
    subshell_expect_and_reply(":", "vsx0");
    subshell_expect("password of the user for PAM tests");
    subshell_expect_and_reply(":", RUNTIME_TEST_USER1_PASSWORD);
    subshell_expect("password of the test user vsx1");
    subshell_expect_and_reply(":", RUNTIME_TEST_USER2_PASSWORD);
    subshell_expect("password of the test user vsx2");
    subshell_expect_and_reply(":", RUNTIME_TEST_USER3_PASSWORD);
    subshell_expect("Installing locales for the i18n test suite");
    subshell_expect("Install /etc/pam.d/lsbpam_conf");
    subshell_expect_and_reply("?", "y");
    subshell_expect("Install nonexistent devices");
    subshell_expect_and_reply("?", "y");

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Remove the backdoor. */
    remove_backdoor();

    /* Close the subshell connection. */
    close_subshell();

    /* Copy the results to the output directory. */
    run_command(
        "cp `ls -1vd /home/tet/test_sets/results/0* | "
        "tail -1`/journal %s/test_journals/runtime_test.journal",
        output_dir
    );

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_libstdcpp_test()
{
    /* Check whether the test package is installed, install it if needed. */
    check_package("lsb-python", "appbat");
    check_package("qm", "runtime");
    check_package("qmtest_libstdcpp", "runtime");

    /* Clear the old test database (if any). */
    run_command("rm -rf /tmp/qmtest_libstdcpp_3.4.3");
}

/*-------------------------------------------------------------------------*/

void run_libstdcpp_test()
{
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/libstdcpp", output_dir);

    /* Open the subshell the test will run in. */
    open_subshell("root", output_dir, CSTRING(log_file_name));

    /* Start the test suite. */
    subshell_send("/opt/lsb/test/share/qmtest_libstdcpp_3.4.3/run_tests ; logout");

    /* Answer the questions. */
    subshell_expect("Where do you want to create the test database");
    subshell_expect_and_reply("?", "/tmp/qmtest_libstdcpp_3.4.3");
    subshell_expect("What is the GNU triplet for your operating system");
    subshell_expect_and_reply("?", "");
    subshell_expect("Name of person running tests");
    subshell_expect_and_reply("?", "Automated");
    subshell_expect("Organisation");
    subshell_expect_and_reply("?", "N/A");
    subshell_expect("Test System");
    subshell_expect_and_reply("?", "LSB Autotest");

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    /* Copy the result file to the result directory. */
    run_command(
        "cp `ls -v1 /tmp/qmtest_libstdcpp_3.4.3/v3db/journal* | "
        "tail -1` %s/test_journals/libstdcpp_test.journal",
        output_dir
    );

    /* Clean up. */
    run_command("rm -rf /tmp/qmtest_libstdcpp_3.4.3");

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_cmdchk_test()
{
    check_package("lsb-cmdchk", "runtime");
}

/*-------------------------------------------------------------------------*/

void run_cmdchk_test()
{
    sstring * starting_command = sstring_create();
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/cmdchk", output_dir);

    /* Open the subshell the test will run in. */
    open_subshell("root", output_dir, CSTRING(log_file_name));

    /* Start the test. */
    sstring_printf(
        starting_command,
        "/opt/lsb/bin/lsbcmdchk "
        "-j %s/test_journals/command_check.journal ; logout",
        output_dir
    );
    subshell_send(CSTRING(starting_command));

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    sstring_delete(starting_command);
    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_libchk_test()
{
    check_package("lsb-libchk", "runtime");
}

/*-------------------------------------------------------------------------*/

void run_libchk_test()
{
    sstring * starting_command = sstring_create();
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/libchk", output_dir);
    sstring_printf(
        starting_command,
        "/opt/lsb/bin/lsblibchk "
        " -j %s/test_journals/library_check.journal ; logout",
        output_dir
    );

    /* Open the subshell the test will run in. */
    open_subshell("root", output_dir, CSTRING(log_file_name));

    /* Start the test. */
    subshell_send(CSTRING(starting_command));

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    sstring_delete(log_file_name);
    sstring_delete(starting_command);
}

/*-------------------------------------------------------------------------*/

void prepare_vsw4_test()
{
    check_package("lsb-tet3-lite", "runtime");
    check_package("lsb-test-vsw4", "runtime");
    check_for_tool("Xvfb");

    /* Remove old config files. */
    run_command("rm -f /opt/lsb/test/vsw4/.configured");
}

/*-------------------------------------------------------------------------*/

void run_vsw4_test()
{
    sstring * log_file_name = sstring_create();

    determine_x_font_paths();

    sstring_printf(log_file_name, "%s/test_logs/vsw4", output_dir);

    /* Open the subshell the test will run in. */
    open_subshell("root", "/opt/lsb/test/vsw4", CSTRING(log_file_name));

    /* Start the test. */
    subshell_send("./run_vsw4.sh ; logout");

    /* Answer the questions. */
    subshell_expect("Name of person running tests");
    subshell_expect_and_reply("?", "Automated");
    subshell_expect("Organisation");
    subshell_expect_and_reply("?", "N/A");
    subshell_expect("Test System");
    subshell_expect_and_reply("?", "LSB Autotest");
    subshell_expect("Hostname of client for X tests");
    subshell_expect_and_reply("?", "");
    subshell_expect("DISPLAY of the Xvfb");
    subshell_expect_and_reply("?", "");
    subshell_expect("XT_FONTPATH");
    subshell_expect_and_reply("?", xt_fontpath);
    subshell_expect("XT_FONTPATH_GOOD");
    subshell_expect_and_reply("?", xt_fontpath_good);

    /* NOTE: The test runs a virtual framebuffer X server with
     * roughly this command line:
     *
     * Xvfb -render -screen 0 1024x768x16 :1.0 -ac +bs -kb
     *      -fp /opt/lsb/test/vsw4/xtest/fonts/,
     *      /usr/X11R6/lib/X11/fonts/misc/
     */

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    /* Free the previously created paths to X11 fonts. */
    cleanup_x_font_paths();

    /* Copy the result file to the output directory. */
    run_command(
        "cp `ls -1vd /opt/lsb/test/vsw4/xtest/results/* "
        "| tail -1`/journal %s/test_journals/x11_test.journal",
        output_dir
    );

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_xts5_test()
{
    check_package("lsb-tet3-lite", "runtime");
    check_package("lsb-test-xts5", "runtime");

    /* Remove old config files. */
    run_command("rm -f /opt/lsb/test/xts5/.configured");
}

/*-------------------------------------------------------------------------*/

void run_xts5_test()
{
    sstring * log_file_name = sstring_create();

    determine_x_font_paths();

    sstring_printf(log_file_name, "%s/test_logs/xts5", output_dir);

    /* Open the subshell the test will run in. */
    open_subshell("root", "/opt/lsb/test/lsb-xts5", CSTRING(log_file_name));

    /* Start the test. */
    subshell_send("./run_xts5.sh ; logout");

    /* Answer the questions. */
    subshell_expect("Hostname of client for X tests");
    subshell_expect_and_reply("?", "");
    subshell_expect("DISPLAY of the Xvfb");
    subshell_expect_and_reply("?", "");
    subshell_expect("XT_FONTPATH");
    subshell_expect_and_reply("?", xt_fontpath);
    subshell_expect("XT_FONTPATH_GOOD");
    subshell_expect_and_reply("?", xt_fontpath_good);

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    /* Free the previously created paths to X11 fonts. */
    cleanup_x_font_paths();

    /* Copy the result file to the output directory. */
    run_command(
        "cp `ls -1vd /opt/lsb/test/lsb-xts5/xts5/results/* "
        "| tail -1`/journal %s/test_journals/xts5_test.journal",
        output_dir
    );

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_desktop_test()
{
    check_package ("lsb-tet3-lite", "runtime");
    check_package ("lsb-test-desktop", "runtime");

    /* Remove old config files. */
    run_command("rm -f /opt/lsb/test/desktop/.configured");
}

/*-------------------------------------------------------------------------*/

void run_desktop_test()
{
    sstring * log_file_name = sstring_create();

    determine_x_font_paths();

    sstring_printf(log_file_name, "%s/test_logs/desktop", output_dir);

    /* Open the subshell the test will run in. */
    open_subshell("root", "/opt/lsb/test/desktop", CSTRING(log_file_name));

    /* Start the test. */
    subshell_send("./run_tests ; logout");

    /* Answer the questions. */
    subshell_expect("Name of person running tests");
    subshell_expect_and_reply("?", "Automated");
    subshell_expect("Organisation");
    subshell_expect_and_reply("?", "N/A");
    subshell_expect("Test System");
    subshell_expect_and_reply("?", "LSB Autotest");
    subshell_expect("Hostname of client for X tests");
    subshell_expect_and_reply("?", "");
    subshell_expect("DISPLAY of the Xvfb");
    subshell_expect_and_reply("?", "");
    subshell_expect("X11 font path");
    subshell_expect_and_reply("?", xt_fontpath);

    /* Wait until the subshell finishes. */
    wait_for_subshell();

    /* Close the subshell connection. */
    close_subshell();

    /* Free the previously created paths to X11 fonts. */
    cleanup_x_font_paths();

    /* Copy result files to the output directory. */
    run_command(
        "cp `ls -1vd /opt/lsb/test/desktop/gtkvts/results/* "
        "| tail -1`/journal %s/test_journals/gtkvts_test.journal",
        output_dir
    );
    run_command(
        "cp `ls -1vd /opt/lsb/test/desktop/qt3/results/* "
        "| tail -1`/journal %s/test_journals/qt3_test.journal",
        output_dir
    );
    run_command(
        "cp `ls -1vd /opt/lsb/test/desktop/fontconfig/results/* "
        "| tail -1`/journal %s/test_journals/fontconfig_test.journal",
        output_dir
    );
    run_command(
        "cp /opt/lsb/test/desktop/xml/journal.libxml2 "
        "%s/test_journals/libxml_test.journal",
        output_dir
    );
    run_command(
        "cp /opt/lsb/test/desktop/libpng/journal.pngtest "
        "%s/test_journals/libpng_test.journal",
        output_dir
    );

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_tcl_test()
{
    check_package("lsb-tcl", "appbat");
    check_data_file ("tcl8.4.11-tests.tar");

    /* Add the testing user. */
    add_user("tcl4lsb");

    /* Create the testing directory. */
    run_command("rm -rf /tmp/lsb-tcl-tests");
    run_command_as_user("tcl4lsb", "mkdir -p /tmp/lsb-tcl-tests");

    /* Populate it with files. */
    run_command_as_user(
        "tcl4lsb",
        "cp %s/tcl8.4.11-tests.tar /tmp/lsb-tcl-tests",
        package_dir
    );
    run_command_as_user(
        "tcl4lsb",
        "cd /tmp/lsb-tcl-tests ; tar xvf tcl8.4.11-tests.tar >/dev/null"
    );
}

/*-------------------------------------------------------------------------*/

void run_tcl_test()
{
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/tcl", output_dir);

    /* Run the test. */
    open_subshell("tcl4lsb", "/tmp/lsb-tcl-tests", CSTRING(log_file_name));
    subshell_send(
        "/opt/lsb/appbat/bin/tclsh8.4 all.tcl | tee testoutput ; logout"
    );
    wait_for_subshell();
    close_subshell();

    /* Copy result files to the output directory. */
    run_command(
        "cp /tmp/lsb-tcl-tests/testoutput %s/test_journals/tcl.tcljournal",
        output_dir
    );

    /* Clean up. */
    run_command("rm -rf /tmp/lsb-tcl-tests");
    run_command("userdel tcl4lsb");
    run_command("rm -rf /home/tcl4lsb");

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_expect_test()
{
    check_package("lsb-tcl", "appbat");
    check_package("lsb-expect", "appbat");
    check_data_file("expect-tests.tar");

    /* Add the testing user. */
    add_user("expect4lsb");

    /* Create a testing directory. */
    run_command("rm -rf /tmp/lsb-expect-test");
    run_command_as_user("expect4lsb", "mkdir -p /tmp/lsb-expect-test");

    /* Copy and unpack testing files. */
    run_command_as_user(
        "expect4lsb",
        "cp %s/expect-tests.tar /tmp/lsb-expect-test",
        package_dir
    );
    run_command_as_user(
        "expect4lsb",
        "cd /tmp/lsb-expect-test ; tar xvf expect-tests.tar >/dev/null"
    );
}

/*-------------------------------------------------------------------------*/

void run_expect_test()
{
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/expect", output_dir);

    /* Run the test. */
    open_subshell(
        "expect4lsb",
        "/tmp/lsb-expect-test",
        CSTRING(log_file_name)
    );
    subshell_send(
        "/opt/lsb/appbat/bin/expect all.tcl | tee testoutput ; logout"
    );
    wait_for_subshell();
    close_subshell();

    /* Copy result files to the output directory. */
    run_command(
        "cp /tmp/lsb-expect-test/testoutput %s/test_journals/expect.tcljournal",
        output_dir
    );

    /* Clean up. */
    run_command("rm -rf /tmp/lsb-expect-test");
    run_command("userdel expect4lsb");
    run_command("rm -rf /home/expect4lsb");

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/

void prepare_python_test()
{
    check_package("lsb-python", "appbat");

    /* Create a testing user. */
    add_user("python4lsb");

    /* Create a testing directory. */
    run_command("rm -rf /tmp/lsb-python-test");
    run_command_as_user("python4lsb", "mkdir /tmp/lsb-python-test");

    /* Prepare testing files. */
    run_command_as_user(
        "python4lsb",
        "cp /opt/lsb/appbat/lib/python2.4 /tmp/lsb-python-test -R"
    );
    run_command_as_user(
        "python4lsb",
        "chmod 755 /tmp/lsb-python-test/python2.4/test"
    );

    /*
     * The tar test insists on writing temporary files to the original
     * test directory so we have to make it writable. *sigh*
     */
    run_command("chmod a+w /opt/lsb/appbat/lib/python2.4/test");

    /* Purge precompiled files as the compiler will be tested as well. */
    run_command_as_user(
        "python4lsb",
        "rm -f /tmp/lsb-python-test/python2.4/test/*.pyc"
    );
    run_command_as_user(
        "python4lsb",
        "rm -f /tmp/lsb-python-test/python2.4/test/*.pyo"
    );
}

/*-------------------------------------------------------------------------*/

void run_python_test()
{
    sstring * log_file_name = sstring_create();

    sstring_printf(log_file_name, "%s/test_logs/python", output_dir);

    /* Run the test. */
    open_subshell(
        "python4lsb",
        "/tmp/lsb-python-test/python2.4/test",
        CSTRING(log_file_name)
    );
    subshell_send(
        "/opt/lsb/appbat/bin/python -E regrtest.py -u curses "
        "| tee /tmp/lsb-python-test/testoutput ; logout"
    );
    wait_for_subshell();
    close_subshell();

    /* Copy result files to the output directory. */
    run_command(
        "cp /tmp/lsb-python-test/testoutput %s/test_journals/"
        "python.pythonjournal",
        output_dir
    );

    /* Clean up. */
    run_command("rm -rf /tmp/lsb-python-test");
    run_command("userdel python4lsb");
    run_command("rm -rf /home/python4lsb");

    sstring_delete(log_file_name);
}

/*-------------------------------------------------------------------------*/
