/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "tet_journal.h"
#include "inform.h"
#include "string_utils.h"
#include "string_array.h"
#include "file_utils.h"
#include "useful_macros.h"
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

/* Various line codes in TET journal. */
#define TET_LINE_CODE_CONTROLLER_START              0
#define TET_LINE_CODE_SYSTEM_INFO                   5
#define TET_LINE_CODE_TEST_CASE_START               10
#define TET_LINE_CODE_TEST_CASE_MANAGER_START       15
#define TET_LINE_CODE_CONFIG_START                  20
#define TET_LINE_CODE_CONFIG_VARIABLE               30
#define TET_LINE_CODE_CONFIG_END                    40
#define TET_LINE_CODE_CONTROLLER_MESSAGE            50
#define TET_LINE_CODE_SCENARIO_INFO                 70
#define TET_LINE_CODE_TEST_CASE_END                 80
#define TET_LINE_CODE_USER_ABORT                    90
#define TET_LINE_CODE_CAPTURED_OUTPUT               100
#define TET_LINE_CODE_BUILD_START                   110
#define TET_LINE_CODE_BUILD_END                     130
#define TET_LINE_CODE_TEST_POINT_START              200
#define TET_LINE_CODE_TEST_POINT_RESULT             220
#define TET_LINE_CODE_CLEAN_START                   310
#define TET_LINE_CODE_CLEAN_END                     320
#define TET_LINE_CODE_COMPONENT_START               400
#define TET_LINE_CODE_COMPONENT_END                 410
#define TET_LINE_CODE_TEST_SUITE_ERROR              510
#define TET_LINE_CODE_TEST_SUITE_MESSAGE            520
#define TET_LINE_CODE_CONTROLLER_END                900

/*-------------------------------------------------------------------------*/

/* A structured form of a single line of a TET journal. */
typedef struct {
    int code;
    int line_length;
    char detail[TET_LINE_SIZE];
    char detail_part1[TET_LINE_SIZE];
    char detail_part2[TET_LINE_SIZE];
    char test_case_name[TET_LINE_SIZE];
    int activity_number;
    int test_point;
    int result;
    char time[TET_LINE_SIZE];
}
TetJournalLine;

/*-------------------------------------------------------------------------*/

/*
 * Parses a single line of the TET journal and stores the data
 * to the TetJournalLine structure.
 * Returns nonzero on success, zero on failure (if the call fails,
 * the contents of the TetJournalLine structure are undefined).
 */
static int parse_tet_journal_line(
    const char * line,
    TetJournalLine * line_struct
)
{
    int fields_read;

    assert(line);
    assert(line_struct);

    /* Clear the output structure first. */
    memset(line_struct, 0, sizeof(TetJournalLine));

    /* Store the total line length. */
    line_struct->line_length = strlen(line);
    assert(line_struct->line_length < LINE_BUFFER_SIZE);

    /* Read the numeric code at the start of the line; the rest
     * is the detail.
     */
    fields_read = sscanf(
        line,
        "%d|%[^\n]",
        &line_struct->code,
        line_struct->detail
    );
    if (fields_read != 2) {

        /* The line does not have the prescriped format. */
        return 0;
    }

    /* If the detail has two parts, separate them; the first part contains
     * more factual info, the rest is typically some comment or message.
     * If the detail is not structured in this way, ignore it and store
     * two empty strings.
     */
    fields_read = sscanf(
        line_struct->detail,
        "%[^|]|%[^\n]",
        line_struct->detail_part1,
        line_struct->detail_part2
    );
    if (fields_read != 2) {
        line_struct->detail_part1[0] = '\0';
        line_struct->detail_part2[0] = '\0';
    }

    /* Look for additional info for some line types. */

    if (line_struct->code == TET_LINE_CODE_TEST_CASE_START) {

        fields_read = sscanf(
            line_struct->detail,
            "%d %s %[0-9:]",
            &line_struct->activity_number,
            line_struct->test_case_name,
            line_struct->time
        );
        if (fields_read != 3) {

            /* Bad syntax. */
            return 0;
        }
    }
    else if (line_struct->code == TET_LINE_CODE_TEST_CASE_END) {

        fields_read = sscanf(
            line_struct->detail,
            "%d %d %[0-9:]",
            &line_struct->activity_number,
            &line_struct->result,
            line_struct->time
        );
        if (fields_read != 3) {

            /* Bad syntax. */
            return 0;
        }
    }
    else if (line_struct->code == TET_LINE_CODE_TEST_POINT_START) {

        fields_read = sscanf(
            line_struct->detail,
            "%d %d %[0-9:]",
            &line_struct->activity_number,
            &line_struct->test_point,
            line_struct->time
        );
        if (fields_read != 3) {

            /* Bad syntax. */
            return 0;
        }
    }
    else if (line_struct->code == TET_LINE_CODE_TEST_POINT_RESULT) {

        /* The detail should contain three numbers:
         * the activity number, the point number,
         * and the result code.
         */
        fields_read = sscanf(
            line_struct->detail,
            "%d %d %d %[0-9:]",
            &line_struct->activity_number,
            &line_struct->test_point,
            &line_struct->result,
            line_struct->time
        );
        if (fields_read != 4) {

            /* Bad syntax. */
            return 0;
        }
    }

    /* The line was parsed successfully. */
    return 1;
}

/*-------------------------------------------------------------------------*/

/*
 * Builds an indexed array of test points for the given test case.
 * Called from tet_journal_finish().
 */
static void build_test_point_index(TetTestCase * test_case)
{
    TetTestPoint * test_point;
    unsigned int max_test_point_number = 0;

    assert(test_case);

    /* Find the largest test point number. */
    max_test_point_number = 0;
    FOREACH(test_point, test_case->first_test_point) {
        if (test_point->number > max_test_point_number) {
            max_test_point_number = test_point->number;
        }
    }

    /* Do not attempt to build the index twice. */
    assert(test_case->test_points == NULL);

    /* Allocate an array for test points and fill it with NULLs. */
    test_case->test_points = (TetTestPoint **)(
        calloc(max_test_point_number, sizeof(TetTestPoint *))
    );

    /*
     * Populate the array with test points according to their numbers.
     * Each test point is stored at the index equal to its number minus 1
     * (because test point numbers start at 1). If some tests are missing,
     * NULLs will remain at appropriate indices.
     */
    FOREACH(test_point, test_case->first_test_point) {
        assert(test_point->number > 0);
        assert(test_case->test_points[test_point->number - 1] == NULL);
        test_case->test_points[test_point->number - 1] = test_point;
    }

    /* Store the maximum test point number. */
    test_case->max_test_point_number = max_test_point_number;

    /* Build per-test-case statistics of failures, warnings and FIPs. */
    test_case->failed_test_point_count = 0;
    test_case->fipped_test_point_count = 0;
    test_case->warning_test_point_count = 0;
    FOREACH(test_point, test_case->first_test_point) {
        if (tet_result_is_failure(test_point->result)) {
            test_case->failed_test_point_count++;
        }
        else if (test_point->result == TET_RESULT_FIP) {
            test_case->fipped_test_point_count++;
        }
        else if (test_point->result == TET_RESULT_WARNING) {
            test_case->warning_test_point_count++;
        }
    }
}

/*-------------------------------------------------------------------------*/

/*
 * Builds statistics for the journal.
 * Called from tet_journal_finish().
 */
static void build_stats(TetJournal * journal)
{
    TetTestCase * test_case;
    TetTestPoint * test_point;
    char * test_point_name;

    assert(journal);

    /* Prepare the stats structure. */
    journal->stats.failed_test_list = string_array_create();
    journal->stats.fip_list = string_array_create();
    journal->stats.warning_list = string_array_create();

    /*
     * Scan all test points in all test cases and count
     * passes, fails, skips and FIPs.
     */
    FOREACH(test_case, journal->first_test_case) {
        FOREACH(test_point, test_case->first_test_point) {

            /* Build the full name of the test point. */
            test_point_name = format_malloc(
                "%s %d",
                test_case->name,
                test_point->number
            );

            /* Update test point statistics. */
            switch (test_point->result) {
                case TET_RESULT_PASSED:
                    journal->stats.test_points_passed_cleanly++;
                    break;

                case TET_RESULT_WARNING:
                    string_array_append(
                        journal->stats.warning_list,
                        test_point_name
                    );
                    journal->stats.test_points_passed_with_warning++;
                    break;

                case TET_RESULT_FAILED:
                case TET_RESULT_UNRESOLVED:
                case TET_RESULT_UNINITIATED:
                case TET_RESULT_UNREPORTED:
                    string_array_append(
                        journal->stats.failed_test_list,
                        test_point_name
                    );
                    journal->stats.test_points_failed++;
                    break;

                case TET_RESULT_UNUSED:
                case TET_RESULT_UNSUPPORTED:
                case TET_RESULT_UNTESTED:
                case TET_RESULT_UNIMPLEMENTED:
                case TET_RESULT_UNAPPROVED:
                    journal->stats.test_points_skipped++;
                    break;

                case TET_RESULT_FIP:
                    string_array_append(
                        journal->stats.fip_list,
                        test_point_name
                    );
                    journal->stats.test_points_fipped++;
                    break;
            }

            /* Clean up. */
            free(test_point_name);
        }
    }
}

/*-------------------------------------------------------------------------*/

TetJournal * tet_journal_load(const char * filename)
{
    TetJournal * journal = NULL;
    TetTestCase * test_case = NULL;
    TetTestPoint * test_point = NULL;
    TetJournalLine line;
    int line_number = 0;
    char * line_buffer = NULL;
    int buffer_size = 0;
    StringArray * test_point_messages;

    assert(filename);

    /* Open the journal file. */
    FILE * f = fopen(filename, "r");
    if (f == NULL) {
        return NULL;
    }

    /* Create and initialize the journal structure. */
    journal = tet_journal_start();

    /* Create an array to hold messages captured from the test. */
    test_point_messages = string_array_create();

    /* Process the journal line by line. */
    for (;;) {

        /* Read the next line from the journal. */
        line_number++;
        if (!read_line(f, &line_buffer, &buffer_size)) {
            break;
        }
        assert(line_buffer);

        /* Skip empty lines. */
        if (string_empty(line_buffer)) {
            continue;
        }

        /* Analyze the line structure. */
        if (!parse_tet_journal_line(line_buffer, &line)) {

            /* Bad syntax; print a warning and skip the line. */
            warn(
                "%s: %d: unrecognized journal line \"%s\"",
                filename,
                line_number,
                line_buffer
            );
            journal->bad_journal_flag = 1;
            continue;
        }

        /* The line code determines what the line means. */

        if (line.code == TET_LINE_CODE_SYSTEM_INFO) {

            /* Copy the info to the journal. */
            strcpy(journal->system_info.info, line.detail_part1);
        }
        else if (line.code == TET_LINE_CODE_CONTROLLER_START) {

            /* Copy the info to the journal. */
            strcpy(journal->run_info.info, line.detail);
        }
        else if (line.code == TET_LINE_CODE_TEST_CASE_START) {
            if (test_case) {
                warn(
                    "%s: %d: previous test case was not properly finished",
                    filename,
                    line_number
                );
            }

            /* Add a new test case structure to the journal-> */
            test_case = tet_journal_add_test_case(
                journal,
                line.test_case_name,
                line.time,
                line.activity_number
            );
        }
        else if (line.code == TET_LINE_CODE_TEST_CASE_END) {
            if (test_case == NULL) {
                warn(
                    "%s: %d: test case ends without proper start",
                    filename,
                    line_number
                );
                continue;
            }

            /* Store the finishing info of the test case. */
            string_copy_max(test_case->end_time, line.time, TET_TIME_SIZE);
            test_case->end_status = line.result;

            /* Leave this test case. */
            test_case = NULL;
        }
        else if (line.code == TET_LINE_CODE_TEST_SUITE_MESSAGE) {

            /* Capture the mesage. */
            string_array_append(test_point_messages, line.detail_part2);
        }
        else if (line.code == TET_LINE_CODE_TEST_POINT_RESULT) {

            /* Complain if the test point is outside any test case. */
            if (test_case == NULL) {
                warn("%s: %d: test point outside test case");
                journal->bad_journal_flag = 1;
                continue;
            }

            /* Add the test point to the journal-> */
            test_point = tet_journal_add_test_point(
                journal,
                test_case,
                line.test_point,
                line.result,
                NULL,
                "",             /* FIXME: need a start time */
                line.time
            );

            /* Store the messages captured during the test point (if any). */
            if (test_point_messages->length > 0) {
                test_point->messages = string_array_implode(
                    test_point_messages, "\n"
                );
            }

            /* Free the captured test messages. */
            string_array_clear(test_point_messages);

            /* Leave this test point. */
            test_point = NULL;
        }
    }

    /* Close the journal file. */
    fclose(f);

    /* Free the message array. */
    string_array_delete(test_point_messages);

    /* Finish the journal and return it. */
    tet_journal_finish(journal);
    return journal;
}

/*-------------------------------------------------------------------------*/

const char * tet_result_to_string(int tet_result)
{
    switch (tet_result) {
        case TET_RESULT_PASSED:         return "Passed";
        case TET_RESULT_UNUSED:         return "Unused";
        case TET_RESULT_UNSUPPORTED:    return "Unsupported";
        case TET_RESULT_UNTESTED:       return "Untested";
        case TET_RESULT_UNIMPLEMENTED:  return "Unimplemented";
        case TET_RESULT_UNAPPROVED:     return "Unapproved";
        case TET_RESULT_FAILED:         return "Failed";
        case TET_RESULT_UNRESOLVED:     return "Unresolved";
        case TET_RESULT_UNINITIATED:    return "Uninitiated";
        case TET_RESULT_UNREPORTED:     return "Unreported";
        case TET_RESULT_WARNING:        return "Warning";
        case TET_RESULT_FIP:            return "FIP";
        default:                        return "Unknown result";
    }
}

/*-------------------------------------------------------------------------*/

int tet_result_is_failure(int tet_result)
{
    switch (tet_result) {
        case TET_RESULT_FAILED:
        case TET_RESULT_UNRESOLVED:
        case TET_RESULT_UNINITIATED:
        case TET_RESULT_UNREPORTED:
            return 1;

        default:
            return 0;
    }
}

/*-------------------------------------------------------------------------*/

int tet_result_is_skip(int tet_result)
{
    switch (tet_result) {
        case TET_RESULT_UNUSED:
        case TET_RESULT_UNSUPPORTED:
        case TET_RESULT_UNTESTED:
        case TET_RESULT_UNIMPLEMENTED:
        case TET_RESULT_UNAPPROVED:
            return 1;

        default:
            return 0;
    }
}

/*-------------------------------------------------------------------------*/

void tet_journal_delete(TetJournal * journal)
{
    TetTestCase * test_case, * next_test_case;
    TetTestPoint * test_point, * next_test_point;

    assert(journal);

    /* Iterate over test cases and free them. */
    test_case = journal->first_test_case;
    while (test_case) {

        /* Iterate over test points and free them. */
        test_point = test_case->first_test_point;
        while (test_point) {

            /* Free the contents of the structure. */
            free(test_point->messages);

            /* Free the structure. */
            next_test_point = test_point->next;
            free(test_point);
            test_point = next_test_point;
        }

        /* Free the contents of the structure. */
        free(test_case->name);

        /* Free the structure. */
        next_test_case = test_case->next;
        free(test_case);
        test_case = next_test_case;
    }

    /* Free the journal-> */
    free(journal);
}


/*-------------------------------------------------------------------------*/

TetJournal * tet_journal_start(void)
{
    TetJournal * journal;

    /* Allocate the structure and initialize it with zeros and NULLs. */
    journal = (TetJournal *) (calloc(1, sizeof(TetJournal)));
    assert(journal);

    return journal;
}

/*-------------------------------------------------------------------------*/

void tet_journal_finish(TetJournal * journal)
{
    TetTestCase * test_case;

    assert(journal);
    assert(!journal->finished);

    /* Scan all test cases, building an index of test points for each. */
    FOREACH(test_case, journal->first_test_case) {
        build_test_point_index(test_case);
    }

    /* Build statistics. */
    build_stats(journal);

    /* Mark the journal as finished. */
    journal->finished = 1;
}

/*-------------------------------------------------------------------------*/

TetTestCase * tet_journal_add_test_case(
    TetJournal * journal,
    const char * name,
    const char * start_time,
    unsigned int activity_number
)
{
    TetTestCase * test_case;

    assert(journal);

    /* Create and initialize the test case structure. */
    test_case = (TetTestCase *)(calloc(1, sizeof(TetTestCase)));
    assert(test_case);
    test_case->name = string_duplicate(name);
    string_copy_max(test_case->start_time, start_time, TET_TIME_SIZE);
    test_case->activity_number = activity_number;

    /* Queue it to the end of the journal. */
    if (journal->first_test_case == NULL) {
        assert(journal->last_test_case == NULL);
        journal->first_test_case = test_case;
        journal->last_test_case = test_case;
    }
    else {
        assert(journal->last_test_case);
        journal->last_test_case->next = test_case;
        journal->last_test_case = test_case;
    }

    return test_case;
}

/*-------------------------------------------------------------------------*/

TetTestPoint * tet_journal_add_test_point(
    TetJournal * journal,
    TetTestCase * test_case,
    int number,
    int result,
    const char * messages,
    const char * start_time,
    const char * end_time
)
{
    TetTestPoint * test_point;

    assert(journal);
    assert(test_case);

    /* Create and initialize the test point structure. */
    test_point = (TetTestPoint *)(calloc(1, sizeof(TetTestPoint)));
    assert(test_point);
    test_point->number = number;
    test_point->result = result;
    test_point->messages = messages ? string_duplicate(messages) : NULL;
    string_copy_max(test_point->start_time, start_time, TET_TIME_SIZE);
    string_copy_max(test_point->end_time, end_time, TET_TIME_SIZE);

    /* Queue it to the end of the test case. */
    if (test_case->first_test_point == NULL) {
        assert(test_case->last_test_point == NULL);
        test_case->first_test_point = test_point;
        test_case->last_test_point = test_point;
    }
    else {
        assert(test_case->last_test_point);
        test_case->last_test_point->next = test_point;
        test_case->last_test_point = test_point;
    }

    return test_point;
}

/*-------------------------------------------------------------------------*/

TetTestCase * tet_journal_find_test_case(
    TetJournal * journal,
    const char * name
)
{
    TetTestCase * test_case;

    assert(journal);
    assert(name);

    /* Iterate over test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* Check if the name of the test case matches the specified. */
        if (strcasecmp(test_case->name, name) == 0) {
            return test_case;
        }
    }

    /* No match. */
    return NULL;
}

/*-------------------------------------------------------------------------*/

TetTestPoint * tet_journal_get_test_point(
    TetTestCase * test_case,
    unsigned int number
)
{
    assert(test_case);
    assert(number > 0);
    assert(number <= test_case->max_test_point_number);
    return (test_case->test_points[number - 1]);
}

/*-------------------------------------------------------------------------*/

void tet_journal_print_stats(TetJournal * journal, FILE * f)
{
    int i;

    assert(journal);
    assert(f);

    /* Print test counts. */
    fprintf(
        f,
        "%6d tests passed\n",
        journal->stats.test_points_passed_cleanly
    );
    fprintf(
        f,
        "%6d tests passed with warning\n",
        journal->stats.test_points_passed_with_warning
    );
    fprintf(
        f,
        "%6d tests with FIP status\n",
        journal->stats.test_points_fipped
    );
    fprintf(
        f,
        "%6d tests failed\n",
        journal->stats.test_points_failed
    );
    fprintf(
        f,
        "%6d tests skipped\n",
        journal->stats.test_points_skipped
    );

    /* Print lists of failures, warnings, and FIPs. */
    if (journal->stats.failed_test_list->length > 0) {
        fprintf(f, "\nFailures:\n");
        for (i = 0; i < journal->stats.failed_test_list->length; i++) {
            fprintf(
                f,
                "    %s\n",
                string_array_get(journal->stats.failed_test_list, i)
            );
        }
    }
    if (journal->stats.fip_list->length > 0) {
        fprintf(f, "\nFIPs:\n");
        for (i = 0; i < journal->stats.fip_list->length; i++) {
            fprintf(
                f,
                "    %s\n",
                string_array_get(journal->stats.fip_list, i)
            );
        }
    }
    if (journal->stats.warning_list->length > 0) {
        fprintf(f, "\nWarnings:\n");
        for (i = 0; i < journal->stats.warning_list->length; i++) {
            fprintf(
                f,
                "    %s\n",
                string_array_get(journal->stats.warning_list, i)
            );
        }
    }

    /* Print error flags. */
    fputc('\n', f);
    if (journal->test_suite_error_flag) {
        fprintf(f, "An error has occurred in the test suite.\n");
    }
    if (journal->user_abort_flag) {
        fprintf(f, "The test suite run was aborted by the user.\n");
    }
    if (journal->bad_journal_flag) {
        fprintf(f, "There were problems with parsing the test journal.\n");
    }
}

/*-------------------------------------------------------------------------*/

void tet_journal_print_results(TetJournal * journal, FILE * f)
{
    TetTestCase * test_case;
    TetTestPoint * test_point;

    assert(journal);
    assert(f);

    /* Iterate over test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* Iterate over test points in the test case. */
        FOREACH(test_point, test_case->first_test_point) {
            fprintf(
                f,
                "%s %d %s\n",
                test_case->name,
                test_point->number,
                tet_result_to_string(test_point->result)
            );
        }
    }
}

/*-------------------------------------------------------------------------*/

sstring * tet_journal_summarize(TetJournal * journal, sstring * result)
{
    assert(journal);

    if (journal->stats.test_points_failed > 0) {
        return sstring_printf(result, "%u failures",
            journal->stats.test_points_failed);
    }

    if (journal->stats.test_points_passed_with_warning > 0) {
        return sstring_printf(result, "%u warnings",
            journal->stats.test_points_passed_with_warning);
    }

    if (journal->stats.test_points_fipped > 0) {
        return sstring_printf(result, "success (%u passed, %u skips, "
            "%u FIPs)", journal->stats.test_points_passed_cleanly,
            journal->stats.test_points_skipped,
            journal->stats.test_points_fipped);
    }

    return sstring_printf(result, "success (%u passed, %u skips)",
        journal->stats.test_points_passed_cleanly,
        journal->stats.test_points_skipped);
}

/*-------------------------------------------------------------------------*/
