/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#ifndef TET_JOURNAL_H
#define TET_JOURNAL_H

#include "string_utils.h"
#include "string_array.h"
#include <stdio.h>
#include <stdint.h>

/*-------------------------------------------------------------------------*/

/* Maximum size of a single line of the TET journal. */
#define TET_LINE_SIZE 1024

/* Maximum buffer size needed for storing time value in TET journal. */
#define TET_TIME_SIZE 64

/* Size of the buffer used for reading journal lines. */
#define LINE_BUFFER_SIZE 1024

/*-------------------------------------------------------------------------*/

/* Result codes used in TET journals. */
#define TET_RESULT_PASSED           0
#define TET_RESULT_FAILED           1
#define TET_RESULT_UNRESOLVED       2
#define TET_RESULT_UNUSED           3
#define TET_RESULT_UNSUPPORTED      4
#define TET_RESULT_UNTESTED         5
#define TET_RESULT_UNINITIATED      6
#define TET_RESULT_UNREPORTED       7
#define TET_RESULT_WARNING          101
#define TET_RESULT_FIP              102
#define TET_RESULT_UNIMPLEMENTED    103
#define TET_RESULT_UNAPPROVED       104

/*-------------------------------------------------------------------------*/

/** A test point in the TET journal. */
typedef struct tagTetTestPoint {

    /** Number of the test point within the test case. */
    unsigned int number;

    /** Result of the test point. */
    int result;

    /** Messages produced during this test point (NULL if none). */
    char * messages;

    /** Time when the test point started. */
    char start_time[TET_TIME_SIZE];

    /** Time when the test point finished. */
    char end_time[TET_TIME_SIZE];

    /** Pointer to the next test point (NULL if none). */
    struct tagTetTestPoint * next;
}
TetTestPoint;

/*-------------------------------------------------------------------------*/

/** A test case in the TET journal. */
typedef struct tagTetTestCase {

    /** Name of the test case. */
    char * name;

    /** Time when the test case started. */
    char start_time[TET_TIME_SIZE];

    /** Time when the test case finished. */
    char end_time[TET_TIME_SIZE];

    /** Activity number assigned to this test case. */
    unsigned int activity_number;

    /** Status when finished. */
    int end_status;

    /** Pointer to the first test point (NULL if the test case is empty). */
    TetTestPoint * first_test_point;

    /** Pointer to the last test point (NULL if the test case is empty). */
    TetTestPoint * last_test_point;

    /**
     * An array containing test points indexed by their number
     * (minus 1 because test points are numbered starting from 1).
     * The array may contain NULLs if some test points were silently skipped.
     */
    TetTestPoint ** test_points;

    /**
     * Maximum number of a test point; this is also the length
     * of the test_points array.
     */
    unsigned int max_test_point_number;

    /**
     * Number of failed test points in this test case.
     */
    unsigned int failed_test_point_count;

    /**
     * Number of test points with FIP result in this test case.
     */
    unsigned int fipped_test_point_count;

    /**
     * Number of test points that passed with warning in this test case.
     */
    unsigned int warning_test_point_count;

    /** Pointer to the next test case (NULL if this is the last one). */
    struct tagTetTestCase * next;
}
TetTestCase;

/*-------------------------------------------------------------------------*/

/** System information in the TET journal. */
typedef struct {

    /** The info text. */
    char info[TET_LINE_SIZE];
}
TetSystemInfo;

/*-------------------------------------------------------------------------*/

/** Test run information in the TET journal. */
typedef struct {

    /** The info text. */
    char info[TET_LINE_SIZE];
}
TetRunInfo;

/*-------------------------------------------------------------------------*/

/**
 * Statistical information about a TET journal.
 */
typedef struct {

    /** Total number of test points. */
    int test_points_total;

    /** Number of test points that passed cleanly. */
    int test_points_passed_cleanly;

    /** Number of test points that passed with warning. */
    int test_points_passed_with_warning;

    /** Number of failed test points. */
    int test_points_failed;

    /** Number of skipped test points. */
    int test_points_skipped;

    /** Number of FIPped test points. */
    int test_points_fipped;

    /** List of full names of failed test points. */
    StringArray * failed_test_list;

    /** List of full names of FIPped test points. */
    StringArray * fip_list;

    /** List of full names of test points which passed with warnings. */
    StringArray * warning_list;
}
TetJournalStats;

/*-------------------------------------------------------------------------*/

/** A parsed TET journal. */
typedef struct {

    /** The system info. */
    TetSystemInfo system_info;

    /** The test run info. */
    TetRunInfo run_info;

    /** Pointer to the first test case. */
    TetTestCase * first_test_case;

    /** Pointer to the last test case. */
    TetTestCase * last_test_case;

    /** If nonzero, a test suite error was detected. */
    int test_suite_error_flag;

    /** If nonzero, a test or the test suite was aborted by user. */
    int user_abort_flag;

    /** If nonzero, an error in the journal was detected. */
    int bad_journal_flag;

    /** Statistics of passed and failed test points. */
    TetJournalStats stats;

    /** This flag is set when the journal is finished and ready to use. */
    int finished;
}
TetJournal;

/*-------------------------------------------------------------------------*/

/**
 * Returns a textual representation of the given TET result
 * as a static string.
 */
const char * tet_result_to_string(int tet_result);

/**
 * Returns true if the specified TET result means some kind of a failure,
 * false otherwise.
 */
int tet_result_is_failure(int tet_result);

/**
 * Returns true if the specified TET result means the test was skipped
 * harmlessly, false otherwise.
 */
int tet_result_is_skip(int tet_result);

/**
 * Loads a TET journal and returns it in a newly allocated TetJournal
 * structure. Returns NULL on failure.
 */
TetJournal * tet_journal_load(const char * filename);

/**
 * Deletes the parsed journal from memory.
 */
void tet_journal_delete(TetJournal * journal);

/**
 * Starts creation of a new journal.
 * Returns a pointer to a newly allocated TetJournal structure.
 * This structure should later be populated with test cases
 * and test points using tet_journal_add_test_case() and
 * test_journal_add_test_point().
 * After all test points and test cases are added,
 * tet_journal_finish() must be called to produce a complete journal.
 */
TetJournal * tet_journal_start(void);

/**
 * The last step in creation of a journal.
 * This involves indexing test points and building statistics.
 * Note that the journal is not usable until it is properly finished!
 */
void tet_journal_finish(TetJournal * journal);

/**
 * Adds a test case to the journal.
 * Returns a pointer to the new TetTestCase structure.
 */
TetTestCase * tet_journal_add_test_case(
    TetJournal * journal,
    const char * name,
    const char * start_time,
    unsigned int activity_number
);

/**
 * Adds a test point to the journal.
 * Returns a pointer to the new TetTestPoint structure.
 */
TetTestPoint * tet_journal_add_test_point(
    TetJournal * journal,
    TetTestCase * test_case,
    int number,
    int result,
    const char * messages,
    const char * start_time,
    const char * end_time
);

/**
 * Finds a test case by its name.
 * Returns a pointer to the TetTestCase structure, or NULL
 * if not found.
 */
TetTestCase * tet_journal_find_test_case(
    TetJournal * journal,
    const char * name
);

/**
 * Finds a test point within a test case by its number.
 * Returns a pointer to the TetTestPoint structure, or NULL
 * if not found.
 */
TetTestPoint * tet_journal_get_test_point(
    TetTestCase * test_case,
    unsigned int number
);

/**
 * Prints the journal statistics in human-readable text format.
 */
void tet_journal_print_stats(TetJournal * journal, FILE * f);

/**
 * Prints the results of all test points in the journal in human-readable
 * text format.
 * Same results are not coalesced together; there is strictly one line
 * per test point.
 */
void tet_journal_print_results(TetJournal * journal, FILE * f);

/**
 * Summarizes the journal into a single-line, short result.
 */
sstring * tet_journal_summarize(TetJournal * journal, sstring * result);

/*-------------------------------------------------------------------------*/

#endif
