/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "tet_journal_compat.h"
#include "file_utils.h"
#include "string_utils.h"
#include "inform.h"
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

TetJournal * tcl_to_tet_journal(const char * filename)
{
    TetJournal * journal;
    TetTestCase * test_case = NULL;
    TetTestPoint * test_point = NULL;
    FILE * f;
    char * line_buffer = NULL;
    int buffer_size = 0;
    StringArray * messages;
    int line_number = 0;

    assert(filename);

    /* Open the journal file. */
    f = fopen(filename, "r");
    if (f == NULL) {
        return NULL;
    }

    /* Create an empty journal. */
    journal = tet_journal_start();

    messages = string_array_create();

    /* Read the Tcl journal line by line. */
    for (;;) {

        /* Read the next line. */
        line_number++;
        if (!read_line(f, &line_buffer, &buffer_size)) {
            break;
        }

        /* A single word ending with '.test' marks the start
         * of a test case of that name, and the end of the previous one.
         */
        if (
            string_ends_with(line_buffer, ".test") &&
            (strchr(line_buffer, ' ') == NULL)
        ) {
            /* Finish the previous test case and test point. */
            if (test_point) {

                /* Write all accumulated messages to the test point. */
                assert(test_point->messages == NULL);
                test_point->messages = string_array_implode(
                    messages, "\n");
                string_array_clear(messages);
            }

            test_case = tet_journal_add_test_case(journal, line_buffer,
                "", 0);

            /* In a Tcl journal, every test case has exactly one
             * test point. Add it, assuming for now that it will succeed.
             */
            test_point = tet_journal_add_test_point(
                journal,
                test_case,
                1,
                TET_RESULT_PASSED,
                NULL,
                "",
                ""
            );
        }
        else if (string_ends_with(line_buffer, "FAILED")) {

            /* The current test point has failed. */
            if (test_point) {
                assert(test_case);
                test_point->result = TET_RESULT_FAILED;
            }
            else {
                warn(
                    "%s: %d: A failure line without previous test name",
                    filename,
                    line_number
                );
                journal->bad_journal_flag = 1;
            }
        }
        else {

            /* Any other line is considered informative and stored. */
            string_array_append(messages, line_buffer);
        }
    }

    /* Return the resulting journal. */
    tet_journal_finish(journal);
    return journal;
}

/*-------------------------------------------------------------------------*/

TetJournal * python_to_tet_journal(const char * filename)
{
    TetJournal * journal;
    TetTestCase * test_case = NULL;
    TetTestPoint * test_point = NULL;
    FILE * f;
    char * line_buffer = NULL;
    int buffer_size = 0;
    StringArray * messages;
    int line_number = 0;
    char * test_case_name;

    assert(filename);

    /* Open the journal file. */
    f = fopen(filename, "r");
    if (f == NULL) {
        return NULL;
    }

    /* Create an empty journal. */
    journal = tet_journal_start();

    messages = string_array_create();

    /* Read the Python journal line by line. */
    for (;;) {

        /* Read the next line. */
        line_number++;
        if (!read_line(f, &line_buffer, &buffer_size)) {
            break;
        }
        assert(line_buffer);

        /* Lines starting with "test_" either start a new test,
         * or report the result of the previous one.
         */
        if (string_starts_with(line_buffer, "test_")) {

            /* Extract the name of the test case. */
            test_case_name = string_duplicate_up_to(line_buffer, " ");

            /* A line containing only the test name means a start
             * of a new test. If the test succeeds, it is the only line
             * which is printed.
             */
            if (strchr(line_buffer, ' ') == NULL) {

                /* Finish the previous test case and test point. */
                if (test_point) {

                    /* Write all accumulated messages to the test point. */
                    assert(test_point->messages == NULL);
                    test_point->messages = string_array_implode(
                        messages, "\n");
                    string_array_clear(messages);
                }

                /* Add the new test case. */
                test_case = tet_journal_add_test_case(
                    journal,
                    test_case_name,
                    "",
                    0
                );

                /* Every test case has exactly one test point.
                 * Add it, assuming for now that it will succeed.
                 */
                test_point = tet_journal_add_test_point(
                    journal,
                    test_case,
                    1,
                    TET_RESULT_PASSED,
                    NULL,
                    "",
                    ""
                );
            }
            else if (strstr(line_buffer, "skipped --") != NULL) {
                test_point->result = TET_RESULT_UNSUPPORTED;
            }
        }
        else {

            /* Any other line is considered informative and stored. */
            string_array_append(messages, line_buffer);
        }
    }

    tet_journal_finish(journal);
    return journal;
}

/*-------------------------------------------------------------------------*/

TetJournal * load_any_journal(const char * filename)
{
    TetJournal * journal = NULL;

    /* Identify the journal type and load it appropriately. */
    if (string_ends_with(filename, ".tcljournal")) {

        /* Looks like a Tcl journal. */
        journal = tcl_to_tet_journal(filename);
    }
    else if (string_ends_with(filename, ".pythonjournal")) {

        /* Looks like a Python journal. */
        journal = python_to_tet_journal(filename);
    }
    else {

        /* A normal journal (hopefully). */
        journal = tet_journal_load(filename);
    }

    /* Return the result. */
    return journal;
}

/*-------------------------------------------------------------------------*/
