/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "file_utils.h"
#include "string_utils.h"
#include "tet_journal_html.h"
#include "useful_macros.h"
#include "html_output.h"
#include <assert.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

static void print_test_point_range(
    html_file * f,
    TetTestCase * test_case,
    TetTestPoint * first_test_point,
    TetTestPoint * last_test_point
)
{
    sstring * title = sstring_create();
    sstring * buffer = sstring_create();
    sstring * anchor = sstring_create();

    /* True if this result needs a HTML anchor, i.e. we want to add it
     * to a clickable summary. This is true for interesting results,
     * which are failures, warnings and FIPs.
     */
    int need_anchor = 0;

    /* True if we want to print messages associated with this test point.
     * This is true for failures, warnings, and FIPs.
     *
     * NOTE: These messages would be quite interesting for all test points
     * but we cannot print all of them. There is simply too much of them,
     * and the resulting HTML file would be *huge*. What can we do...
     */
    int need_messages = 0;

    /* The result of the test range. All tests in the range should have
     * the same result (otherwise they would not be grouped into a range).
     */
    int tet_result = first_test_point->result;

    /* Choose title and highlighting according to the result. */
    if (tet_result_is_skip(tet_result)) {
        sstring_printf(
            title,
            "<span style=\"color:gray;\">%s</span>",
            tet_result_to_string(tet_result)
        );
    }
    else if (tet_result_is_failure(tet_result)) {
        sstring_printf(
            title,
            "<span style=\"color:red; font-weight:bold;\">%s</span>",
            tet_result_to_string(tet_result)
        );
        need_anchor = 1;
        need_messages = 1;
    }
    else if (tet_result == TET_RESULT_WARNING) {
        sstring_printf(
            title,
            "<span style=\"color:orange; font-weight:bold;\">%s</span>",
            tet_result_to_string(tet_result)
        );
        need_anchor = 1;
        need_messages = 1;
    }
    else if (tet_result == TET_RESULT_FIP) {
        sstring_printf(
            title,
            "<span style=\"color:green;\">%s</span>",
            tet_result_to_string(tet_result)
        );
        need_anchor = 1;
        need_messages = 1;
    }
    else {
        sstring_printf(title, "%s", tet_result_to_string(tet_result));
    }

    /* Print the numbers of the test points in the range. */
    html_start_table_line(f);
    if (first_test_point == last_test_point) {

        /* A single test point. */
        sstring_printf(buffer, "%d", first_test_point->number);
    }
    else {

        /* A whole range of test points. */
        sstring_printf(buffer, "%d-%d", first_test_point->number,
            last_test_point->number);
    }
    html_print_table_field(f, CSTRING(buffer));

    /* Print the title, optionally with an anchor. */
    if (need_anchor) {
        sstring_printf(anchor, "%s_%d", test_case->name,
            first_test_point->number);
        sstring_filename_to_identifier(anchor);
        sstring_printf(buffer, "<a name=\"%s\">%s</a>", CSTRING(anchor),
            CSTRING(title));
        html_print_table_field(f, CSTRING(buffer));
    }
    else {
        html_print_table_field(f, CSTRING(title));
    }

    /* Print messages associated with this test point(s). */
    if (need_messages) {
        if (first_test_point->messages) {
            sstring_printf(buffer, "<pre>%s</pre>", first_test_point->messages);
            html_print_table_field(f, CSTRING(buffer));
        }
    }

    /* Finish the line. */
    html_end_table_line(f);

    sstring_delete(title);
    sstring_delete(anchor);
    sstring_delete(buffer);
}

/*-------------------------------------------------------------------------*/

#define MERGE_IGNORE_MESSAGES      1
#define MERGE_ALL_FAILURES         2

/*
 * Checks whether the test point can be merged with the following one
 * into a single range in the output.
 *
 * Flags:
 *
 * - MERGE_IGNORE_MESSAGES:
 *      Merge points even if their messages differ.
 *
 * - MERGE_ALL_FAILURES:
 *      Merge failed points even if the exact cause of the failure differs.
 */
static int tet_test_point_can_merge(
    TetTestPoint * point,
    int flags
)
{
    assert(point);

    if (point->next == NULL) {

        /* No next point to merge with. */
        return 0;
    }

    /* Look if the two consecutive test points have the same results.
     * If the MERGE_ALL_FAILURES flag is used, we consider all failure
     * types as the same result.
     */
    if (
        (point->result == point->next->result) ||
        (
            (flags & MERGE_ALL_FAILURES) &&
            (tet_result_is_failure(point->result) && tet_result_is_failure(point->next->result))
        )
    ) {

        /* Only points with sequential numbers can be merged. */
        if (point->next->number != point->number + 1) {
            return 0;
        }

        /* If the MERGE_IGNORE_MESSAGES flag is used, we can merge
         * the points regardless of whether they have messages or not.
         */
        if (flags & MERGE_IGNORE_MESSAGES) {
            return 1;
        }

        /* If both points have no messages, they can be merged. */
        if (
            (point->messages == NULL) &&
            (point->next->messages == NULL)
        ) {
            return 1;
        }

        /* If both points have exactly the same messages, they also
         * can be merged.
         */
        else if (point->messages && point->next->messages &&
            string_equals(point->messages, point->next->messages)) {

            return 1;
        }
    }

    /* Does not seem to be the case. */
    return 0;
}

/*-------------------------------------------------------------------------*/

/*
 * Determines a range of test points that can be coalesced into a single
 * table line in the output.
 * Returns the length of the range (always at least 1).
 * Also stores the last test point of the range in the 'last' variable.
 */
static unsigned int make_test_point_range(
    TetTestPoint * first,
    TetTestPoint ** last,
    int flags
)
{
    TetTestPoint * test_point;
    unsigned int range_length = 0;

    assert(first);
    assert(last);

    /* Look if we can merge two or more points together. */
    test_point = first;
    for (;;) {
        if (tet_test_point_can_merge(test_point, flags)) {
            test_point = test_point->next;
            range_length++;
            continue;
        }
        break;
    }

    /* Store the last mergeable point and return nonzero
     * if at least two can be coalesced, zero if none.
     */
    *last = test_point;
    return range_length;
}

/*-------------------------------------------------------------------------*/

/*
 * Builds a HTML link to the given test case.
 */
static void test_case_to_html_link(
    sstring * buffer,
    const char * test_case_name,
    const char * path_to_html_journal
)
{
    /* Create an anchor from the test case name. */
    sstring * anchor = sstring_create();
    sstring_set(anchor, test_case_name);
    sstring_filename_to_identifier(anchor);

    sstring_printf(
        buffer,
        "<a href=\"%s#%s\">%s</a> ",
        path_to_html_journal,
        CSTRING(anchor),
        test_case_name
    );
    sstring_delete(anchor);
}

/*-------------------------------------------------------------------------*/

/*
 * Builds a HTML anchor that refers to the specified test point
 * of the specified test case in the HTML-ized journal.
 */
static void build_test_point_anchor(
    sstring * buffer,
    const char * test_case_name,
    unsigned int test_point_number
)
{
    sstring_printf(buffer, "%s %u", test_case_name, test_point_number);
    sstring_filename_to_identifier(buffer);
}

/*-------------------------------------------------------------------------*/

/*
 * Builds a HTML link to the given test point or a range of test points.
 */
static void test_point_to_html_link(
    sstring * buffer,
    const char * test_case_name,
    unsigned int test_point_range_start,
    unsigned int test_point_range_end,
    const char * path_to_html_journal
)
{
    sstring * anchor = sstring_create();

    /* Build a HTML anchor to the start of the test point range. */
    build_test_point_anchor(anchor, test_case_name,
        test_point_range_start);

    /* Print the test point number(s) with the anchor. */
    if (test_point_range_end != test_point_range_start) {

        /* A whole range of test points. */
        sstring_printf(
            buffer,
            "<a href=\"%s#%s\">%u-%u</a>",
            path_to_html_journal,
            CSTRING(anchor),
            test_point_range_start,
            test_point_range_end
        );
    }
    else {

        /* Only one test point. */
        sstring_printf(
            buffer,
            "<a href=\"%s#%s\">%u</a>",
            path_to_html_journal,
            CSTRING(anchor),
            test_point_range_start
        );
    }

    sstring_delete(anchor);
}

/*-------------------------------------------------------------------------*/

#define FILTER_FAILURES         1
#define FILTER_WARNINGS         2
#define FILTER_FIPS             3

static void print_test_point_list(
    html_file * f,
    TetJournal * journal,
    const char * path_to_html_journal,
    unsigned int filter
)
{
    sstring * buffer = sstring_create();
    TetTestCase * test_case;
    TetTestPoint * test_point;
    TetTestPoint * next_test_point;
    unsigned int merge_rule = 0;

    assert(journal);

    merge_rule = MERGE_IGNORE_MESSAGES;
    if (filter == FILTER_FAILURES) {
        merge_rule |= MERGE_ALL_FAILURES;
    }

    /* Scan all test points in all test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* Skip test cases which have no test points of appropriate type. */
        if (filter == FILTER_FAILURES) {
            if (test_case->failed_test_point_count == 0) { continue; }
        }
        else if (filter == FILTER_WARNINGS) {
            if (test_case->warning_test_point_count == 0) { continue; }
        }
        else if (filter == FILTER_FIPS) {
            if (test_case->fipped_test_point_count == 0) { continue; }
        }

        /* Print the name of the test case and a link to it. */
        test_case_to_html_link(buffer, test_case->name, path_to_html_journal);
        html_println(f, CSTRING(buffer));

        /* Print the numbers of failed test points. */
        FOREACH(test_point, test_case->first_test_point) {

            /* Skip test points which are not of appropriate type.
             * If the filter is FILTER_FAILURES, we skip all non-failures;
             * if the filter is FILTER_WARNINGS, we skip all non-warnings;
             * if the filter is FILTER_FIPS, we skip all non-FIPs.
             */
            if (filter == FILTER_FAILURES) {
                if (!tet_result_is_failure(test_point->result)) { continue; }
            }
            else if (filter == FILTER_WARNINGS) {
                if (test_point->result != TET_RESULT_WARNING) { continue; }
            }
            else if (filter == FILTER_FIPS) {
                if (test_point->result != TET_RESULT_FIP) { continue; }
            }

            /* Sequences of test points with the same result are merged
             * into a single range for readability.
             */
            make_test_point_range(
                test_point,
                &next_test_point,
                merge_rule
            );

            /* Print the test point number (or range). */
            test_point_to_html_link(
                buffer,
                test_case->name,
                test_point->number,
                next_test_point->number,
                path_to_html_journal
            );
            html_println(f, CSTRING(buffer));

            /* Move to the next unprocessed test point. */
            test_point = next_test_point;
        }

        /* Finish the line. */
        html_println(f, "<br>");
    }
}

/*-------------------------------------------------------------------------*/

void tet_journal_to_html(html_file * f, TetJournal * journal)
{
    TetTestCase * test_case;
    TetTestPoint * test_point;
    TetTestPoint * last_in_range;
    sstring * buffer = sstring_create();
    sstring * anchor = sstring_create();
    unsigned int merge_rule;

    assert(f);
    assert(journal);

    /* Print the host system and test run info. */
    html_print_head(f, 3, "Test Summary");
    html_start_paragraph(f);
    html_start_table(f);
    html_start_table_body(f);
    html_start_table_line(f);
    html_print_table_field(f, "<strong>System info: </strong>");
    html_print_table_field(f, journal->system_info.info);
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Test run info: </strong>");
    html_print_table_field(f, journal->run_info.info);
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Test suite error flag: </strong>");
    html_print_table_field(f, journal->test_suite_error_flag ?
        "<span style=\"color: red; font-weight: bold;\">set</span>"
        : "OK");
    html_next_table_line(f);
    html_print_table_field(f, "<strong>User abort flag: </strong>");
    html_print_table_field(f, journal->user_abort_flag ?
        "<span style=\"color: red; font-weight: bold;\">set</span>"
        : "OK");
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Journal error flag: </strong>");
    html_print_table_field(f, journal->bad_journal_flag ?
        "<span style=\"color: red; font-weight: bold;\">set</span>"
        : "OK");
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Tests passed:</strong>");
    sstring_printf(buffer, "%d", journal->stats.test_points_passed_cleanly);
    html_print_table_field(f, CSTRING(buffer));
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Tests skipped:</strong>");
    sstring_printf(buffer, "%d", journal->stats.test_points_skipped);
    html_print_table_field(f, CSTRING(buffer));
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Tests passed with warning:</strong>");
    sstring_printf(buffer, "%d", journal->stats.test_points_passed_with_warning);
    html_print_table_field(f, CSTRING(buffer));
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Tests that failed:</strong>");
    if (journal->stats.test_points_failed > 0) {
        sstring_printf(buffer, "<span style=\"color: red; font-weight: bold;\">"
            "%d</span>", journal->stats.test_points_failed);
    }
    else {
        sstring_printf(buffer, "0");
    }
    html_print_table_field(f, CSTRING(buffer));
    html_next_table_line(f);
    html_print_table_field(f, "<strong>Tests with FIP result:</strong>");
    sstring_printf(buffer, "%d", journal->stats.test_points_fipped);
    html_print_table_field(f, CSTRING(buffer));
    html_end_table_line(f);
    html_end_table_body(f);
    html_end_table(f);
    html_end_paragraph(f);

    /* Print lists of warnings, failures and FIPs. */
    if (journal->stats.test_points_passed_with_warning > 0) {
        html_print_head(f, 3, "Warnings");
        html_start_paragraph(f);
        print_test_point_list(f, journal, "", FILTER_WARNINGS);
        html_end_paragraph(f);
    }
    if (journal->stats.test_points_failed > 0) {
        html_print_head(f, 3, "Failures");
        html_start_paragraph(f);
        print_test_point_list(f, journal, "", FILTER_FAILURES);
        html_end_paragraph(f);
    }
    if (journal->stats.test_points_fipped > 0) {
        html_print_head(f, 3, "FIPs");
        html_start_paragraph(f);
        print_test_point_list(f, journal, "", FILTER_FIPS);
        html_end_paragraph(f);
    }

    /* Start the HTML table to contain the journal lines. */
    html_print_head(f, 3, "Test details");
    html_start_table(f);
    html_start_table_body(f);

    /* Iterate over test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* Print the test case header with a HTML anchor that allows
         * linking from other documents.
         */
        sstring_set(anchor, test_case->name);
        sstring_filename_to_identifier(anchor);
        html_start_table_line(f);
        sstring_printf(buffer,
            "<a name=\"%s\"><strong>%s</strong></a></td></tr>\n",
            CSTRING(anchor),
            test_case->name
        );
        html_print_multicolumn_field(f, CSTRING(buffer), 3);
        html_end_table_line(f);

        /* Iterate over test points in the test case. */
        FOREACH(test_point, test_case->first_test_point) {

            /* If the result of this test point is not much interesting,
             * like a pass or a skip, we can merge a range of them regardless
             * of their messages because the messages are not printed anyway.
             */
            if (tet_result_is_skip(test_point->result) ||
                test_point->result == TET_RESULT_PASSED) {
                merge_rule = MERGE_IGNORE_MESSAGES;
            }
            else {
                merge_rule = 0;
            }

            /* Try if the point can be merged with the following ones
             * to save lines in the HTML output.
             */
            if (make_test_point_range(test_point, &last_in_range,
                merge_rule)) {

                print_test_point_range(
                    f,
                    test_case,
                    test_point,
                    last_in_range
                );

                /* Skip directly to the end of the range. */
                test_point = last_in_range;
            }
            else {

                /* Print the test point line. */
                print_test_point_range(f, test_case, test_point, test_point);
            }
        }
    }

    sstring_delete(buffer);
    sstring_delete(anchor);
}

/*-------------------------------------------------------------------------*/

void tet_journal_to_html_file(const char * filename, TetJournal * journal)
{
    html_file * f;

    assert(filename);
    assert(journal);

    /* Open the HTML file. */
    f = html_open(filename);
    if (f == NULL) {
        fail("Could not open output file: %s", filename);
    }

    /* Print the journal as HTML. */
    tet_journal_to_html(f, journal);

    /* Close the output file. */
    html_close(f);
}

/*-------------------------------------------------------------------------*/

void tet_journal_summarize_html(TetJournal * journal, sstring * result)
{
    assert(journal);
    assert(result);

    if (journal->stats.test_points_failed > 0) {
        sstring_printf(result, "<span style=\"color:red;\">"
            "%u fails</span>", journal->stats.test_points_failed);
    }
    else if (journal->stats.test_points_passed_with_warning > 0) {
        sstring_printf(result, "<span style=\"color: orange;\">"
            "%u warnings</span>",
            journal->stats.test_points_passed_with_warning);
    }
    else if (journal->stats.test_points_fipped > 0) {
        sstring_printf(result, "<span style=\"color: green;\">"
            "success (pass/skip %u/%u, %u FIPs)</span>",
            journal->stats.test_points_passed_cleanly,
            journal->stats.test_points_skipped,
            journal->stats.test_points_fipped);
    }
    else {
        sstring_printf(result, "<span style=\"color: green;\">"
            "success (pass/skip %u/%u)</span>",
            journal->stats.test_points_passed_cleanly,
            journal->stats.test_points_skipped);
    }
}

/*-------------------------------------------------------------------------*/
