/*-------------------------------------------------------------------------*/
/* LSB Autotest, the automatic LSB test running tool                       */
/*                                                                         */
/* Copyright (c) 2006 SUSE LINUX Products GmbH                             */
/*                                                                         */
/* Author: Jiri Dluhos <jdluhos@suse.cz>                                   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License version 2 as       */
/* published by the Free Software Foundation.                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */
/*                                                                         */
/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              */
/* MA 02110-1301, USA.                                                     */
/*-------------------------------------------------------------------------*/

#include "file_utils.h"
#include "string_utils.h"
#include "tet_journal_indexed.h"
#include "useful_macros.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------*/

void tet_journal_print_indexable_results(
    TetJournal * journal,
    FILE * results_file,
    FILE * index_file
)
{
    TetTestCase * test_case;
    TetTestPoint * test_point;
    unsigned long results_file_index;
    unsigned int i;

    assert(journal);
    assert(results_file);
    assert(index_file);

    /* Print the beginning-of-file marker to the results file. */
    fputc('!', results_file);

    /* Iterate over test cases. */
    FOREACH(test_case, journal->first_test_case) {

        /* For each test case, write the offset in the results file
         * to the index file.
         */
        results_file_index = ftell(results_file);
        fprintf(
            index_file,
            "%s %lu %u\n",
            test_case->name,
            results_file_index,
            test_case->max_test_point_number
        );

        /* Iterate over test points in the test case. */
        for (i = 0; i < test_case->max_test_point_number; i++) {
            test_point = test_case->test_points[i];
            if (test_point == NULL) {
                fputc('.', results_file);
            }
            else {
                switch (test_point->result) {
                    case TET_RESULT_PASSED:
                        fputc('p', results_file);
                        break;

                    case TET_RESULT_FAILED:
                        fputc('F', results_file);
                        break;

                    case TET_RESULT_UNRESOLVED:
                        fputc('R', results_file);
                        break;

                    case TET_RESULT_UNUSED:
                        fputc('u', results_file);
                        break;

                    case TET_RESULT_UNSUPPORTED:
                        fputc('s', results_file);
                        break;

                    case TET_RESULT_UNTESTED:
                        fputc('t', results_file);
                        break;

                    case TET_RESULT_UNINITIATED:
                        fputc('I', results_file);
                        break;

                    case TET_RESULT_UNREPORTED:
                        fputc('E', results_file);
                        break;

                    case TET_RESULT_WARNING:
                        fputc('w', results_file);
                        break;

                    case TET_RESULT_FIP:
                        fputc('f', results_file);
                        break;

                    case TET_RESULT_UNIMPLEMENTED:
                        fputc('i', results_file);
                        break;

                    case TET_RESULT_UNAPPROVED:
                        fputc('a', results_file);
                        break;

                    default:
                        fputc('?', results_file);
                        break;
                }
            }
        }
        fputc(' ', results_file);
    }

    /* Print the end-of-file marker to the results file. */
    fputc('@', results_file);
}

/*-------------------------------------------------------------------------*/

void tet_journal_to_res_idx(TetJournal * journal,
    const char * base_file_name)
{
    FILE * res_file = NULL;
    FILE * idx_file = NULL;
    sstring * file_name;

    file_name = sstring_create();

    /* Open the .res and .idx file. */
    sstring_printf(file_name, "%s.res", base_file_name);
    res_file = open_file_for_writing(file_name->buffer);
    sstring_printf(file_name, "%s.idx", base_file_name);
    idx_file = open_file_for_writing(file_name->buffer);

    /* Write contents. */
    tet_journal_print_indexable_results(journal, res_file, idx_file);

    /* Clean up. */
    sstring_delete(file_name);
}

/*-------------------------------------------------------------------------*/

int tet_journal_read_result_from_indexable_file(
    const char * results_file_name,
    const char * index_file_name,
    const char * test_case_name,
    unsigned int test_point_number
)
{
    unsigned long offset;
    unsigned long test_point_count;
    FILE * results_file;
    FILE * index_file;
    int result;

    /* Find and read the test case line in the .idx file. */
    index_file = fopen(index_file_name, "r");
    if (index_file == NULL) {
        return -1;
    }
    if (
        !idx_find_test_case(
            index_file,
            test_case_name,
            &offset,
            &test_point_count
        )
    ) {

        /* Not found. */
        fclose(index_file);
        return -1;
    }
    fclose(index_file);

    /* Open the results file. */
    results_file = fopen(results_file_name, "r");
    if (results_file == NULL) {
        return -1;
    }

    /* Seek to the offset in the results file
     * where the results for this case start, plus the byte offset
     * of the test point result itself, which is equal to its number
     * minus 1 (because test points are numbered starting from 1).
     */
    if (
        !res_seek_to_test_point_result(
            results_file,
            offset,
            test_point_number
        )
    ) {

        /* The index was wrong, or an I/O error has occurred. */
        fclose(results_file);
        fclose(index_file);
        return -1;
    }

    /* Read the result. */
    result = res_read_test_point_result(results_file);

    /* Done. */
    fclose(results_file);
    return result;
}

/*-------------------------------------------------------------------------*/

int idx_read_line(
    FILE * idx_file,
    char ** test_case_name,
    unsigned long * offset,
    unsigned long * result_count
)
{
    char * buffer = NULL;
    int buffer_size = 0;

    assert(idx_file);
    assert(test_case_name);
    assert(offset);
    assert(result_count);

    /* Read the whole line. */
    if (!read_line(idx_file, &buffer, &buffer_size)) {

        /* I/O error or end of file. */
        return 0;
    }

    /* Separate the test case name, which is at the start. */
    *test_case_name = string_duplicate_up_to(buffer, " ");
    if (sscanf(buffer, "%*s %lu %lu", offset, result_count) != 2) {

        /* The index line does not have the correct structure. */
        free(buffer);
        return 0;
    }

    /* Done. */
    free(buffer);
    return 1;
}

/*-------------------------------------------------------------------------*/

int idx_find_test_case(
    FILE * idx_file,
    const char * test_case_name,
    unsigned long * offset,
    unsigned long * result_count
)
{
    char * idx_test_case_name = NULL;

    /* Start reading the .idx file from the beginning. */
    rewind(idx_file);

    /* Scan the .idx file, looking for the test case name. */
    for (;;) {
        if (
            !idx_read_line(
                idx_file,
                &idx_test_case_name,
                offset,
                result_count
            )
        ) {

            /* End of file (or an I/O error). */
            return 0;
        }

        /* Is this the test case we look for? */
        if (string_equals(idx_test_case_name, test_case_name)) {
            return 1;
        }
    }
}

/*-------------------------------------------------------------------------*/

int res_read_test_point_result(
    FILE * res_file
)
{
    int result_char;

    assert(res_file);

    /* Read the single-byte result and decode it. */
    result_char = fgetc(res_file);
    switch (result_char) {
        case 'p':       return TET_RESULT_PASSED;
        case 'F':       return TET_RESULT_FAILED;
        case 'R':       return TET_RESULT_UNRESOLVED;
        case 'u':       return TET_RESULT_UNUSED;
        case 's':       return TET_RESULT_UNSUPPORTED;
        case 't':       return TET_RESULT_UNTESTED;
        case 'I':       return TET_RESULT_UNINITIATED;
        case 'E':       return TET_RESULT_UNREPORTED;
        case 'w':       return TET_RESULT_WARNING;
        case 'f':       return TET_RESULT_FIP;
        case 'i':       return TET_RESULT_UNIMPLEMENTED;
        case 'a':       return TET_RESULT_UNAPPROVED;

        /* A dot means that the test result was missing completely;
         * assume it is the same as UNUSED.
         */
        case '.':       return TET_RESULT_UNUSED;

        /* Anything other is an error. */
        default:        return -1;
    }
}

/*-------------------------------------------------------------------------*/

int res_seek_to_test_point_result(
    FILE * res_file,
    unsigned long test_case_offset,
    unsigned int test_point_number
)
{
    assert(res_file);
    assert(test_point_number > 0);  /* test point number is never 0 */

    /* Seek to the test case offset, plus the byte offset
     * of the test point result, which is equal to its number
     * minus 1 (because test points are numbered starting from 1).
     */
    if (
        fseek(
            res_file,
            test_case_offset + test_point_number - 1,
            SEEK_SET
        ) != 0
    ) {

        /* Could not seek(). */
        return 0;
    }

    return 1;
}

/*-------------------------------------------------------------------------*/
